/*   Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
//**************************************************************************//
//                                                                          //
//                        Linux ELF64 Symbol Harvester                      //
//                        ----------------------------                      //
//                                                                          //
//**************************************************************************//
//
#include "a2nhdr.h"

#define SEGMENT_TYPE_EXECUTE(flag)    ((flag) & PF_X)
#define SEGMENT_TYPE_WRITE(flag)      ((flag) & PF_W)
#define SEGMENT_TYPE_READ(flag)       ((flag) & PF_R)


//
// Externs 
//
extern gv_t a2n_gv;                         // from a2n.c


//
// Demangler interface (see cplus-dem.c)
//
char * cplus_demangle(const char * mangled, int options);

#define DMGL_PARAMS     (1 << 0)            // Include function args
#define DMGL_ANSI       (1 << 1)            // Include const, volatile, etc
#define DMGL_GNU        (1 << 9)            // gnu-style demangling


#if SYNTHESIZE_SYMBOLS

   /**
	* Under certain conditions on Linux/PPC64, synthetic symbols must
	* be generated using bfd_get_synthetic_symtab().  The global
	* variable, synth_syms, points to an array of synthetic asymbol
	* structs returned from bfd_get_synthetic_symtab.  (See linuxelf64.c.)
	* Upon returning from bfd_get_synthetic_symtab, pointers into this array
	* are stored in a member of a MOD_NODE struct declared below.  The
	* synth_syms variable is declared global to keep it in scope for the
	* entire life of the MOD_NODE struct that references it.
   **/
   asymbol * synth_syms;


/* The following stuff was liberally "borrowed" from binutils. */

/* Iterate through BFD's sections, harvest code, and add section node. */

void processSections(bfd * abfd, asection * section, void *context_rec)
{
	bfd_size_type datasize = 0;
	if ((section->flags & (SEC_CODE | SEC_HAS_CONTENTS))
	    != (SEC_CODE | SEC_HAS_CONTENTS))
		return;

	CONTEXT_REC *cr = (CONTEXT_REC *) context_rec;
	datasize = bfd_get_section_size(section);
	if (datasize == 0)
		return;
	section->userdata = malloc(datasize);
	int rc = bfd_get_section_contents(abfd, section, section->userdata, 0,
				 datasize);
	if (!rc) {
		dbghv(("*E* processBFDSymbols: ** INTERNAL ERROR: bfd_get_section_contents returned 0 for %s; section %s\n",
			abfd->filename, section->name));
		return;
	}
	// Call the callback function with each executable section.
	// Don't care whether user accepts of not.
	//
	SendBackSection(section->name,	// Section name
			section->index,	// Section number
			bfd_get_section_by_name(abfd, section->name),	// BFD asection pointer
			section->vma,	// Section start address/offset
			section->filepos, // Section offset
			datasize,	// Section size
			section->flags,	// Section flags
			section->userdata,	// Where I have it loaded
			1,	// Executable section
			cr);	// Context record

	return;

}


/* Filter out certain types of symbols;
   compact them downward to get rid of the rest.
   Return the number of symbols to be printed.  */

static long
filter_symbols(bfd * abfd, bfd_boolean dynamic, void *minisyms,
	       long symcount, unsigned int size)
{
	bfd_byte *from, *fromend, *to;
	asymbol *store;

	store = bfd_make_empty_symbol(abfd);
	if (store == NULL) {
		errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_make_empty_symbol returned NULL"));
		return -1;
	}

	from = (bfd_byte *) minisyms;
	fromend = from + symcount * size;
	to = (bfd_byte *) minisyms;

	for (; from < fromend; from += size) {
		int keep = 1;
		asymbol *sym;

		sym =
		    bfd_minisymbol_to_symbol(abfd, dynamic,
					     (const void *) from, store);
		if (sym == NULL) {
			errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_minisymbol_to_symbol returned NULL"));
			return -1;
		}

		if (keep && (sym->flags & BSF_DEBUGGING) != 0)
			keep = 0;

		if (keep && (bfd_is_abs_section(sym->section)
			     || bfd_is_und_section(sym->section)))
			keep = 0;

		if (keep && bfd_is_target_special_symbol(abfd, sym))
			keep = 0;

		if (keep) {
			memcpy(to, from, size);
			to += size;
		}
	}

	return (to - (bfd_byte *) minisyms) / size;
}

static bfd *sort_bfd;
static asymbol *sort_x;
static asymbol *sort_y;
#define valueof(x) ((x)->section->vma + (x)->value)

/* Numeric sorts.  Undefined symbols are always considered "less than"
   defined symbols with zero values.  Common symbols are not treated
   specially -- i.e., their sizes are used as their "values".  */

static int non_numeric_forward(const void *P_x, const void *P_y)
{
	asymbol *x, *y;
	const char *xn, *yn;

	x = bfd_minisymbol_to_symbol(sort_bfd, 0, P_x, sort_x);
	y = bfd_minisymbol_to_symbol(sort_bfd, 0, P_y, sort_y);
	if (x == NULL || y == NULL) {
		errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_minisymbol_to_symbol returned NULL"));
		return 0;
	}

	xn = bfd_asymbol_name(x);
	yn = bfd_asymbol_name(y);

	if (yn == NULL)
		return xn != NULL;
	if (xn == NULL)
		return -1;

#ifdef HAVE_STRCOLL
	/* Solaris 2.5 has a bug in strcoll.
	   strcoll returns invalid values when confronted with empty strings.  */
	if (*yn == '\0')
		return *xn != '\0';
	if (*xn == '\0')
		return -1;

	return strcoll(xn, yn);
#else
	return strcmp(xn, yn);
#endif
}

/* Intended for use as the comparison function argument to
 * qsort().  If value of symbol is the same,
 * we try to sort by section vma.
 */
static int size_forward1(const void *P_x, const void *P_y)
{
	asymbol *x, *y;
	asection *xs, *ys;
	const char *xn, *yn;
	size_t xnl, ynl;
	int xf, yf;

	x = bfd_minisymbol_to_symbol(sort_bfd, 0, P_x, sort_x);
	y = bfd_minisymbol_to_symbol(sort_bfd, 0, P_y, sort_y);
	if (x == NULL || y == NULL) {
		errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_minisymbol_to_symbol returned NULL"));
		return 0;
	}

	xs = bfd_get_section(x);
	ys = bfd_get_section(y);

	if (bfd_is_und_section(xs))
		abort();
	if (bfd_is_und_section(ys))
		abort();

	if (valueof(x) != valueof(y))
		return valueof(x) < valueof(y) ? -1 : 1;

	if (xs->vma != ys->vma)
		return xs->vma < ys->vma ? -1 : 1;

	xn = bfd_asymbol_name(x);
	yn = bfd_asymbol_name(y);
	xnl = strlen(xn);
	ynl = strlen(yn);

	/* The symbols gnu_compiled and gcc2_compiled convey even less
	   information than the file name, so sort them out first.  */

	xf = (strstr(xn, "gnu_compiled") != NULL
	      || strstr(xn, "gcc2_compiled") != NULL);
	yf = (strstr(yn, "gnu_compiled") != NULL
	      || strstr(yn, "gcc2_compiled") != NULL);

	if (xf && !yf)
		return -1;
	if (!xf && yf)
		return 1;

	/* We use a heuristic for the file name.  It may not work on non
	   Unix systems, but it doesn't really matter; the only difference
	   is precisely which symbol names get printed.  */

#define file_symbol(s, sn, snl)			\
  (((s)->flags & BSF_FILE) != 0			\
   || ((sn)[(snl) - 2] == '.'			\
       && ((sn)[(snl) - 1] == 'o'		\
	   || (sn)[(snl) - 1] == 'a')))

	xf = file_symbol(x, xn, xnl);
	yf = file_symbol(y, yn, ynl);

	if (xf && !yf)
		return -1;
	if (!xf && yf)
		return 1;

	return non_numeric_forward(P_x, P_y);
}


struct elf_internal_sym {
	bfd_vma st_value;	/* Value of the symbol */
	bfd_vma st_size;	/* Associated symbol size */
	unsigned long st_name;	/* Symbol name, index in string tbl */
	unsigned char st_info;	/* Type and binding attributes */
	unsigned char st_other;	/* Visibilty, and target specific */
	unsigned int st_shndx;	/* Associated section index */
};

typedef struct elf_internal_sym Elf_Internal_Sym;

/* Information held for an ELF symbol.  The first field is the
   corresponding asymbol.  Every symbol is an ELF file is actually a
   pointer to this structure, although it is often handled as a
   pointer to an asymbol.  */

typedef struct {
	/* The BFD symbol.  */
	asymbol symbol;
	/* ELF symbol information.  */
	Elf_Internal_Sym internal_elf_sym;
	/* Backend specific information.  */
	union {
		unsigned int hppa_arg_reloc;
		void *mips_extr;
		void *any;
	} tc_data;

	/* Version information.  This is from an Elf_Internal_Versym
	   structure in a SHT_GNU_versym section.  It is zero if there is no
	   version information.  */
	unsigned short version;

} elf_symbol_type;

/* End of code borrowed from binutils . */

// keepThisSymbol() returns 1, indicating the symbol is worthwhile keeping;
// otherwise returns 0;
int static keepThisSymbol(asymbol * sym, unsigned int symtype,
			  int it_is_the_kernel)
{
	if (symtype == BSF_NO_FLAGS)
		return 0;

	if (!(sym->section->flags & SEC_CODE))
		return 0;

	if (!(sym->flags & (BSF_LOCAL | BSF_GLOBAL | BSF_WEAK)))
		return 0;
	if (!sym->name || sym->name[0] == '\0')
		return 0;
	// C++ exception stuff
	if (sym->name[0] == '.' && sym->name[1] == 'L')
		return 0;

	if (!it_is_the_kernel  && symtype == SYMBOL_TYPE_OBJECT) {
		/* keep functions and labels only */
		return 0;
	}
	/* It's a keeper */
	return 1;
}

int static processBFDSymbols(CONTEXT_REC * cr, bfd * abfd,
			     asymbol ** asyms, int symcount,
			     uint size, int it_is_the_kernel)
{
	int numberOfSymbolsKept = 0;

	bfd_byte *from, *fromend;
	asymbol *sym = NULL;
	asymbol *store_sym, *store_next;

	unsigned int symtype;

	from = (bfd_byte *) asyms;
	fromend = from + (symcount * size);

	store_sym = sort_x;
	store_next = sort_y;

	if (from < fromend) {
		sym =
		    bfd_minisymbol_to_symbol(abfd, 0,
					     (const void *) from,
					     store_sym);
		if (sym == NULL) {
			errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_minisymbol_to_symbol returned null\n"));
			return 0;
		}
	}
	dbghv(("name \t type \t symoffset \t vma \t codeoffset \t filepos\n"));
	for (; from < fromend; from += size) {
		asymbol *next;
		asection *sec;
		bfd_vma sz = 0;
		asymbol *temp;

		if (from + size < fromend) {
			next = bfd_minisymbol_to_symbol(abfd,
							0,
							(const void
							 *) (from + size),
							store_next);
			if (next == NULL) {
				errmsg(("*E* processBFDSymbols: ***** INTERNAL ERROR: bfd_minisymbol_to_symbol returned null\n"));
				return 0;
			}
		} else
			next = NULL;

		dbghv(("%s \t\t\t", sym->name));
		symtype = BSF_NO_FLAGS;
		if (sym->flags & BSF_WEAK & BSF_FUNCTION) {
			symtype = SYMBOL_TYPE_FUNCTION_WEAK;
			dbghv(("FUNCTION_WEAK \t"));
		} else if (sym->flags & BSF_OBJECT) {
			dbghv(("OBJECT \t"));
			symtype = SYMBOL_TYPE_OBJECT;
		} else if (sym->flags & BSF_FUNCTION) {
			symtype = SYMBOL_TYPE_FUNCTION;
			dbghv(("FUNCTION \t"));
		} else if ((sym->flags & BSF_LOCAL)
			   || (sym->flags & BSF_GLOBAL)) {
			symtype = SYMBOL_TYPE_LABEL;
			dbghv(("LABEL \t"));
		} else
			dbghv(("BSF_NO_FLAGS \t"));

		if (keepThisSymbol (sym, symtype, it_is_the_kernel)) {
			if (symtype == SYMBOL_TYPE_LABEL) {
				sz = 0;
			} else {

				sec = bfd_get_section(sym);

				if (bfd_is_com_section(sec))
					sz = sym->value;
				else {
					if (from + size < fromend
					    && sec ==
					    bfd_get_section(next))
						sz = valueof(next) -
						    valueof(sym);
					else
						sz = (bfd_get_section_vma
						      (abfd, sec)
						      +
						      bfd_section_size
						      (abfd, sec)
						      - valueof(sym));
				}
			}
			uint symoffset;
			/* Eventually, symoffset will be compared to offsets calculated
			 * from real physical addresses -- i.e, relative to the base address
			 * of segments that were memory-mapped and hung off the task_struct.
			 * 		vma = tp->mm->mmap
			 * where tp is of type 'task_struct'.
			 * So, for user tasks, we use section filepos in calculating symoffset;
			 * for the kernel then, we  use setion vma.
			 */
			if (it_is_the_kernel)
				symoffset = sym->value + sym->section->vma;
			else
				symoffset = sym->value + sym->section->filepos;
			uint code_offset = sym->value;
			dbghv(("%X \t %p \t %X \t %X\n", symoffset, sym->section->vma, code_offset, sym->section->filepos));
			char *code_ptr =  (char *)PtrAdd(((char *) Uint32ToPtr(code_offset)), sym->section->userdata);

			PrintBytes(code_ptr, sz);


			// Call the callback function with each symbol we've decided to keep.
			// Callback function returns 0 if it wants us to continue, non-zero
			// if it wants us to stop.
			int rc = SendBackSymbol (sym->name,	// Symbol name
					     symoffset,	// Symbol offset
						sz,	// Symbol size
					     symtype,	// Symbol type
					     sym->section->index,	// Section to which the symbol belongs
					     code_ptr,	// Code associated with this symbol
					     cr);	// Context record
			if (rc) {
				dbghv(("< processBFDSymbols: SendBackSymbol returned %d. Stopping\n", rc));
				return (numberOfSymbolsKept);	// Error. Stop enumerating and return
				// number of symbols found so far.
			}
			numberOfSymbolsKept++;

		} else
			dbghv((" \t Discard symbol\n"));


		sym = next;
		temp = store_sym;
		store_sym = store_next;
		store_next = temp;

		}

	//
	// Done enumerating symbols
	//
	dbghv(("< processBFDSymbols: Done enumerating. Found %d symbols.\n", numberOfSymbolsKept));
	return (numberOfSymbolsKept);
}

// GetSymbolsInternal() for BFD symbols, including synthesized symbols as needed.
int static GetSymbolsUsingBfdInternal(CONTEXT_REC * cr, char *modname,
				      int module_type, bfd * abfd)
{
	void *buf;
	uint size;
	MOD_NODE *mn = (MOD_NODE *) cr->handle;

	bfd_map_over_sections(abfd, processSections, cr);
	long nr_mini_syms = bfd_read_minisymbols(abfd, 0, &buf, &size);
	if (nr_mini_syms < 1)
		return 0;
	asymbol **mini_syms = (asymbol **) buf;
	buf = NULL;
	synth_syms = NULL;

	long nr_synth_syms = bfd_get_synthetic_symtab (abfd, nr_mini_syms,
						       mini_syms, 0, NULL,
						       &synth_syms);
	if (nr_synth_syms < 0) {
		free (mini_syms);
		return 0;
	}

	int asymcnt = nr_mini_syms + nr_synth_syms;
	/* Add '1' to actual count for NULL marker */
	asymbol **asym =
	    (asymbol **) malloc((asymcnt + 1) * sizeof(asymbol **));
	size_t i;
	for (i = 0; i < (size_t) nr_mini_syms; ++i)
		asym[i] = mini_syms[i];
	for (i = 0; i < (size_t) nr_synth_syms; ++i)
		asym[nr_mini_syms + i] = synth_syms + i;
	// bfd_canonicalize_symtab does this, so shall we
	asym[asymcnt] = NULL;
	mini_syms = asym;
	int it_is_the_kernel = 0;
	if (mn->flags & A2N_FLAGS_KERNEL)
		it_is_the_kernel = 1;


	/*It's OK to do this in place; we'll free the storage anyway
	   (after printing).  */

	asymcnt = filter_symbols(abfd, 0, mini_syms, asymcnt, size);

	sort_bfd = abfd;
	sort_x = bfd_make_empty_symbol(abfd);
	sort_y = bfd_make_empty_symbol(abfd);

	qsort(mini_syms, asymcnt, size, size_forward1);
	int numSymbolsKept =
	    processBFDSymbols(cr, abfd, mini_syms, asymcnt, size,
			      it_is_the_kernel);
	if (numSymbolsKept != 0)
		mn->flags &= ~A2N_FLAGS_PLT_SYMBOLS_ONLY;	// There are symbols besides the PLT
	free(mini_syms);
	return numSymbolsKept;
}
#endif // SYNTHESIZE_SYMBOLS


//
// GetSymbolsFromElf64ExecutableInternal()
// ***************************************
//
int static GetSymbolsFromElf64ExecutableInternal (CONTEXT_REC * cr, char *modname, int module_type, bfd * abfd)
{
   Elf64_Ehdr * eh;                         // Elf Header
   Elf64_Phdr * ph_table, * ph;             // Program Header
   Elf64_Shdr * sh_table, * sh, * sym_sh;   // Section Header
   Elf64_Sym  * sym_table, * sym;           // Symbol Table/Section

   MOD_NODE * mn;

   int i, rc, symtype, symbind;
   int NumberOfSymbols = 0;                 // Number of symbols in executable
   int NumberOfSymbolsKept = 0;             // Number of symbols we actually kept
   int it_is_the_kernel = 0;                // Just what it says

   char * shstrtab;                         // String table base for Sections
   char * strtab;                           // String section base for SYMTAB section
   char * symname, * demangled_name = NULL; // Symbol name
   uint symoffset;                          // Symbol offset
   unsigned char * code_ptr;                // Code associated with symbol

   char * secname;
   int have_debug_line = 0, have_debug_info = 0;

   uintptr base_addr = UINTPTR_MAX;

   SECTION_INFO e_sec[MAX_EXEC_SECTIONS];   // Executable section information
   int escnt = 0;                           // Number of executable sections
   int plt_index = -1;                      // Index to .plt section

   SECTION_INFO a_sec[MAX_ALLOC_SECTIONS];  // Non-executable section information
   int ascnt = 0;                           // Number of non-executable sections

   int symtab_section_cnt = 0;              // Number of SYMTAB sections

#if defined(A2N_USE_BFD)
   long storage;
#endif


   dbghv(("> GetSymbolsFromElf64ExecutableInternal: fn='%s', cr=%p, module_type=%d, abfd=%p\n", modname, cr, module_type, abfd));
   //
   // Set beginning of the Program Header table and Section Header table
   //
   mn = (MOD_NODE *)cr->handle;
   eh = (Elf64_Ehdr *)cr->mod_addr;                       // Elf header is beginning of file
   ph_table = (Elf64_Phdr *)PtrAdd(eh, eh->e_phoff);      // Program headers table
   sh_table = (Elf64_Shdr *)PtrAdd(eh, eh->e_shoff);      // Section headers table

   mn->li.abfd = abfd;
   mn->li.asym = NULL;
   mn->li.asymcnt = 0;
   mn->flags |= A2N_FLAGS_64BIT_MODULE;                   // ELF (64-bit) module

   if (mn->flags & A2N_FLAGS_KERNEL)
      it_is_the_kernel = 1;

   //
   // Check if this is a relocatable (REL or DYN) and it's already fixed up
   // (kind of like an EXEC which looks more like an image).  I've seen this
   // in RH8.0's /lib/i686/libc-2.2.93.so.
   //
	if ((module_type == MODULE_TYPE_ELF64_REL)
	    && eh->e_entry >= cr->load_addr) {
      // Entry point >= loaded_addr. Assume e_entry is a virtual addr not an offset,
      // Change to this pseudo type so we can handle offsets/addrs correctly.
      module_type = MODULE_TYPE_ELF64_REL_FIXED;
      dbghv(("- GetSymbolsFromElfExecutableInternal: changing module_type to MODULE_TYPE_ELF64_REL_FIXED\n"));
   }
   //
   // Enumerate the program headers.
   // Remember the lowest virtual address of all the PT_LOAD headers.
   // That will be used when calculating symbol offsets later on ...
   //
   dbghv(("\n  ****************  Program Headers [%d]  **************\n", eh->e_phnum));
   dbghv(("##    Type          Offset           VirtAddr            PhysAdd            FileSize            MemSize         Flags      Align\n"));
   dbghv(("-- ---------- ------------------ ------------------ ------------------ ------------------ ------------------ ---------- ----------\n"));

   ph = ph_table;
   for (i = 0; i < eh->e_phnum; i++) {
      dbghv (("%02d 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x 0x%0x\n",
              i, ph->p_type, ph->p_offset, ph->p_vaddr, ph->p_paddr,
              ph->p_filesz, ph->p_memsz, ph->p_flags, ph->p_align));

      if (ph->p_type == PT_LOAD && ph->p_vaddr < base_addr) {
         base_addr = ph->p_vaddr;
      }
      ph++;
   }

   dbghv(("*** Lowest PT_LOAD virtual = %p\n", base_addr));
   if (base_addr == UINTPTR_MAX) {
      base_addr = (uintptr)cr->load_addr;
      dbghv(("*** Lowest PT_LOAD virtual changed to %p\n", cr->load_addr));
   }
   dbghv(("  *****************************************************\n\n"));

   //
   // Enumerate all section (headers) looking for executable sections
   // (sections that contain code) and for *THE* symbol table (the section
   // that contains debug symbols).
   //
   sh = sh_table + eh->e_shstrndx;                   // Section header for SHSTRTAB (pointer addition!)
   shstrtab = (char *)PtrAdd(eh, sh->sh_offset);     // Start of string table

   dbghv(("  ******************* Sections ********************\n"));
   dbghv(("      type           flags          link       info           addr             f_offset       m_offset        size              addralign          entsize\n"));
   for (i = 0, sh = sh_table;  i < eh->e_shnum;  i++, sh++) {
      secname = (char *)PtrAdd(shstrtab, sh->sh_name);
      //
      // Executable section?
      // If so remember it.  Only symbols in these sections will be harvested.
      // Check for SHF_EXECINSTR (executable machine instructions) in sh_flags.
      //
      if (sh->sh_flags & SHF_EXECINSTR) {
         if (escnt == MAX_EXEC_SECTIONS) {
            errmsg(("*E* GetSymbolsFromElf64ExecutableInternal: ***** INTERNAL ERROR: more than %d executable sections!\n", MAX_EXEC_SECTIONS));
            return (0);                           // No symbols
         }
         if (sh->sh_size > 0xFFFFFFFF || sh->sh_offset > 0xFFFFFFFF) {
            errmsg(("*E* GetSymbolsFromElf64ExecutableInternal: ***** INTERNAL ERROR: sh_size or sh_offset > 0xFFFFFFFF!\n"));
            errmsg(("                                   sh_size = 0x%X  sh_offset = 0x%X\n", sh->sh_size, sh->sh_offset));
            return (0);                           // No symbols
         }
         e_sec[escnt].number = i;
         e_sec[escnt].name = secname;

         if (module_type == MODULE_TYPE_ELF64_EXEC || module_type == MODULE_TYPE_ELF64_REL_FIXED) {
            e_sec[escnt].start_addr = sh->sh_addr;
            e_sec[escnt].offset = PtrToUint32(PtrSub(sh->sh_addr, base_addr));
         }
         else { // MODULE_TYPE_ELF64_REL
            e_sec[escnt].start_addr = cr->load_addr + sh->sh_addr;
            e_sec[escnt].offset = PtrToUint32(sh->sh_addr);
         }

         e_sec[escnt].size = PtrToUint32(sh->sh_size);
         e_sec[escnt].end_addr    = e_sec[escnt].start_addr + sh->sh_size - 1;
         e_sec[escnt].file_offset = PtrToUint32(sh->sh_offset);
         e_sec[escnt].flags = PtrToUint32(sh->sh_flags);
         e_sec[escnt].type = sh->sh_type;

         e_sec[escnt].loc_addr = NULL;      // Until we decide if we need code

         dbghv(("%2x 0x%08x 0x%016llx 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x"  *X* '%s'\n",
                i, sh->sh_type, sh->sh_flags, sh->sh_link, sh->sh_info,
                sh->sh_addr, sh->sh_offset, e_sec[escnt].offset,
                sh->sh_size, sh->sh_addralign, sh->sh_entsize, secname));

         //
         // Remember index to .plt section
         //
         if (strcmp(e_sec[escnt].name, ".plt") == 0) {
            plt_index = escnt;
            dbghv((".plt section is number %d, e_sec_index = %d\n", i, plt_index));
         }

         escnt++;
         continue;
      }

      //
      // Non-executable and allocated section?
      // If so remember it.  These are sections that contain no executable
      // code but are part of the loaded image (thus the allocated name).
      // Since they're not executable there should not be symbols for them
      // and, in fact, we won't gather symbols for them even if they're
      // present.  We only gather symbols for executable sections.
      //
      if (sh->sh_flags & SHF_ALLOC) {
         if (ascnt == MAX_ALLOC_SECTIONS) {
            errmsg(("*E* GetSymbolsFromElf64ExecutableInternal: ***** INTERNAL ERROR: more than %d executable sections!\n", MAX_ALLOC_SECTIONS));
            return (0);                           // No symbols
         }
         a_sec[ascnt].number = i;
         a_sec[ascnt].name = secname;

         if (module_type == MODULE_TYPE_ELF64_EXEC || module_type == MODULE_TYPE_ELF64_REL_FIXED) {
            a_sec[ascnt].start_addr = sh->sh_addr;
            a_sec[ascnt].offset = PtrToUint32(PtrSub(sh->sh_addr, base_addr));
         }
         else { // MODULE_TYPE_ELF64_REL
            a_sec[ascnt].start_addr = cr->load_addr + sh->sh_addr;
            a_sec[ascnt].offset = PtrToUint32(sh->sh_addr);
         }

         a_sec[ascnt].size = PtrToUint32(sh->sh_size);
         a_sec[ascnt].end_addr    = a_sec[ascnt].start_addr + sh->sh_size - 1;
         a_sec[ascnt].file_offset = PtrToUint32(sh->sh_offset);
         a_sec[ascnt].flags = PtrToUint32(sh->sh_flags);
         a_sec[ascnt].type = sh->sh_type;

         a_sec[ascnt].loc_addr = NULL;   // Never code for these!

         dbghv(("%2x 0x%08x 0x%016llx 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x"  *A* '%s'\n",
                i, sh->sh_type, sh->sh_flags, sh->sh_link, sh->sh_info,
                sh->sh_addr, sh->sh_offset, a_sec[ascnt].offset,
                sh->sh_size, sh->sh_addralign, sh->sh_entsize, secname));

         ascnt++;
         continue;
      }

      //
      // Symbol Table?
      // If we find it we have a non-stripped executable with symbols.
      // If we don't find it we have a stripped executable without symbols.
      // ***** Currently non-stripped executables have one and only one
      // ***** Symbol Table.
      //
      if (sh->sh_type == SHT_SYMTAB) {
         if (symtab_section_cnt > 0) {
            errmsg(("*E* GetSymbolsFromElf64ExecutableInternal: ***** INTERNAL ERROR: more than 1 SYMTAB sections!\n"));
            return (0);                           // No symbols
         }
         NumberOfSymbols = sh->sh_size / sizeof(Elf64_Sym);
         sym_sh = sh;                       // Remember section header
         dbghv(("%2x 0x%08x 0x%016llx 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x"  *S* '%s' symcnt=%d\n",
                i, sh->sh_type, sh->sh_flags, sh->sh_link, sh->sh_info,
                sh->sh_addr, sh->sh_offset, 0xffffffff, sh->sh_size,
                sh->sh_addralign, sh->sh_entsize, secname, NumberOfSymbols));
         continue;                          // Assumes there is only 1 symbol table in executable!
      }

      //
      // Some other section type
      //
      if (stricmp(secname, ".debug_line") == 0) {
         have_debug_line = 1;
      }
      if (stricmp(secname, ".debug_info") == 0) {
         have_debug_info = 1;
      }

      dbghv(("%2x 0x%08x 0x%016llx 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%"_LZ64x"      '%s'\n",
             i, sh->sh_type, sh->sh_flags, sh->sh_link, sh->sh_info,
             sh->sh_addr, sh->sh_offset, 0xffffffff, sh->sh_size,
             sh->sh_addralign, sh->sh_entsize, secname));
   }
   dbghv(("  *************************************************\n\n"));

   //
   // If we're being asked to keep the code around then copy the
   // executable sections so we can keep them around after we free the module.
   // If not, then just indicate we don't have any code.
   // As we do this send back each section.
   //
   dbghv(("  ************** Executable Sections **************\n"));
   dbghv(("      type       flags     s_offset       s_addr             e_addr          size          l_addr        name\n"));
   for (i = 0; i < escnt; i++) {
      if (G(gather_code) && e_sec[i].type != SHT_NOBITS) {
         e_sec[i].loc_addr = malloc(e_sec[i].size);
         if (e_sec[i].loc_addr == NULL) {
            errmsg(("*E* GetSymbolsFromElf64ExecutableInternal: Unable to malloc(%d) for code for section %s\n", e_sec[i].size, e_sec[i].name));
         }
      }

      dbghv(("%2x 0x%08x 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x %"_PZp"  %s\n",
             e_sec[i].number, e_sec[i].type, e_sec[i].flags,
             e_sec[i].offset, e_sec[i].start_addr, e_sec[i].end_addr,
             e_sec[i].size, e_sec[i].loc_addr, e_sec[i].name));

      if (e_sec[i].loc_addr != NULL) {
         dbghv(("- GetSymbolsFromElf64ExecutableInternal: Saving %d bytes of code for section %s\n", e_sec[i].size, e_sec[i].name));
         memcpy(e_sec[i].loc_addr,
                (void *)PtrAdd(eh, e_sec[i].file_offset),
                e_sec[i].size);
      }

#if defined(A2N_USE_BFD)
      e_sec[i].sec = bfd_get_section_by_name(abfd, e_sec[i].name);
#else
      e_sec[i].sec = NULL;
#endif
      dbghv(("- GetSymbolsFromElf64ExecutableInternal: asction for %s = %p\n", e_sec[i].name, e_sec[i].sec));

      //
      // Call the callback function with each executable section.
      // Don't care whether user accepts of not.
      //
      SendBackSection(e_sec[i].name,        // Section name
                      e_sec[i].number,      // Section number
                      e_sec[i].sec,         // BFD asection pointer
                      e_sec[i].start_addr,  // Section start address/offset
                      e_sec[i].offset,      // Section offset
                      e_sec[i].size,        // Section size
                      e_sec[i].flags,       // Section flags
                      e_sec[i].loc_addr,    // Where I have it loaded
                      1,                    // Executable section
                      cr);                  // Context record

      //
      // If there is a .plt section then create a symbol for it.
      // Don't care whether it works or not.
      //
      if (i == plt_index) {
         dbghv(("- GetSymbolsFromElf64ExecutableInternal: Adding pseudo symbol for .plt\n"));
         mn->flags |= A2N_FLAGS_PLT_SYMBOLS_ONLY;  // Only symbol is PLT (so far)
         SendBackSymbol(A2N_SYMBOL_PLT_STR,   // Symbol name
                        e_sec[i].offset,      // Symbol offset
                        e_sec[i].size,        // Symbol size
                        SYMBOL_TYPE_ELF_PLT,  // Symbol type
                        e_sec[i].number,      // Section to which the symbol belongs
                        e_sec[i].loc_addr,    // Code associated with this symbol
                        cr);                  // Context record
      }
   }

   dbghv(("  ************ Non-executable Sections ************\n"));
   dbghv(("      type       flags     s_offset       s_addr             e_addr          size          l_addr        name\n"));
   for (i = 0; i < ascnt; i++) {
      dbghv (("%2x 0x%08x 0x%08x 0x%08x 0x%"_LZ64x" 0x%"_LZ64x" 0x%08x %"_PZp"  %s\n",
              a_sec[i].number, a_sec[i].type, a_sec[i].flags, a_sec[i].offset,
              a_sec[i].start_addr, a_sec[i].end_addr, a_sec[i].size,
              a_sec[i].loc_addr, a_sec[i].name));

#if defined(A2N_USE_BFD)
      a_sec[i].sec = bfd_get_section_by_name(abfd, a_sec[i].name);
#else
      a_sec[i].sec = NULL;
#endif
      dbghv(("- GetSymbolsFromElf64ExecutableInternal: asction for %s = %p\n", a_sec[i].name, a_sec[i].sec));

      //
      // Call the callback function with each executable section.
      // Don't care whether user accepts of not.
      //
      SendBackSection(a_sec[i].name,        // Section name
                      a_sec[i].number,      // Section number
                      a_sec[i].sec,         // BFD asection pointer
                      a_sec[i].start_addr,  // Section start address/offset
                      a_sec[i].offset,      // Section offset
                      a_sec[i].size,        // Section size
                      a_sec[i].flags,       // Section flags
                      a_sec[i].loc_addr,    // Where I have it loaded
                      0,                    // Non-executable section
                      cr);                  // Context record
   }
   dbghv(("  *************************************************\n\n"));

   //
   // If there are no symbols (whether because there was no Symbol Table
   // section or because there were 0 symbols) then we're done.
   //
   if (NumberOfSymbols == 0) {
      dbghv(("< GetSymbolsFromElf64ExecutableInternal: file '%s' contains no symbols.\n", modname));
      return (0);                           // No symbols
   }
   //
   // Enumerate symbols using BFD so we can get line numbers later on ...
   // We know the module has symbols.
   //
   // ######################################################################
   //   This is a hack for now. The symbols returned by BFD don't have a
   //   length which means I can't collapse labels inside functions to the
   //   function name. So for now I'm reading the symbol table myself.
   //   I do need to canonicalize symbols because the symbol table pointer
   //   is required later on the call to get the nearest line number.
   //   I need to figure out how to use the BFD symbols and add all the
   //   other stuff I need ...
   // ######################################################################
   //
   if (G(lineno)) {
#if defined(A2N_USE_BFD)
      storage = bfd_get_symtab_upper_bound(abfd);
      if (storage <= 0) {
         warnmsg(("*W* GetSymbolsFromElf64ExecutableInternal: bfd_get_symtab_upper_bound() returned %d. No line numbers.\n", storage));
      }
      else {
         mn->li.asym = (asymbol **)malloc(storage);
         if (mn->li.asym == NULL) {
            warnmsg(("*W* GetSymbolsFromElf64ExecutableInternal: unable to malloc() for bfd symbols. No line numbers.\n"));
         }
         else {
            mn->li.asymcnt = bfd_canonicalize_symtab(abfd, mn->li.asym);
            if (mn->li.asymcnt < 0) {
               warnmsg(("*W* GetSymbolsFromElf64ExecutableInternal: bfd_canonicalize_symtab() returned %d. No line numbers.\n", NumberOfSymbols));
               mn->li.asymcnt = 0;
            }
         }
      }
	   if (mn->li.asymcnt > 0) {
         NumberOfSymbols = mn->li.asymcnt;
         if (have_debug_line && have_debug_info) {
            dbghv (("***** Looks like there is line number debug information *****\n"));
            mn->flags &= ~A2N_FLAGS_NO_LINENO;
         }
      }
#endif
   }

   //
   // Now enumerate the symbols ...
   // sym_sh is pointing to the section header of the SYMTAB section ...
   //
   sym_table = (Elf64_Sym *)PtrAdd(eh, sym_sh->sh_offset);  // Beginning of symbol table
   sh = sh_table + sym_sh->sh_link;                         // SYMTAB string section (pointer addition!)
   strtab = (char *)PtrAdd(eh, sh->sh_offset);              // Start of string section

   dbghv(("*S*  st_name     st_value            st_size             st_info  st_other  st_shndx   name\n"));

   for (i = 0, sym = sym_table;  i < NumberOfSymbols;  i++, sym++) {

      // Invalid section index
      if (sym->st_shndx == 0) {
         DumpElf64Symbol("NULL", "***Section index == 0***", sym);
         continue;
      }

      symname = (char *)PtrAdd(strtab, sym->st_name);      // Grab symbol name (NULL str if st_name == 0)
      symtype = ELF32_ST_TYPE(sym->st_info);    // Isolate symbol type
      symbind = ELF32_ST_BIND(sym->st_info);    // Isolate symbol binding

      //
      // Labels:    (STT_NOTYPE && (STB_LOCAL || STB_GLOBAL))
      // Functions: (STT_FUNC && (STB_LOCAL || STB_GLOBAL || STB_WEAK))
      // Objects;   (STT_OBJECT && (STB_LOCAL || STB_GLOBAL))
      //
      // NOTE:
      //  If it is the kernel then collect objects as well as functions and labels.
      //  Need to do that because objects appear in KSYSMs and if we don't collect
      //  them now then kernel symbol validation will fail.
      //  All these kernel special cases are here only to make validation work.
      //
      //    BIND        TYPE      sym->st_info   What it means
      //  ----------  ----------  ------------   -------------------------------------
      //  STB_LOCAL   STT_NOTYPE      0x00       Local (static) Label
      //  STB_LOCAL   STT_OBJECT      0x01       Local (static) Object
      //  STB_LOCAL   STT_FUNC        0x02       Local (static) Function
      //  STB_GLOBAL  STT_NOTYPE      0x10       Global Label
      //  STB_GLOBAL  STT_OBJECT      0x11       Global Object
      //  STB_GLOBAL  STT_FUNC        0x12       Global Function
      //  STB_WEAK    STT_NOTYPE      0x20       Weak Label (?? does it make sense ??)
      //  STB_WEAK    STT_FUNC        0x22       Weak Function
      //
      if (symbind != STB_LOCAL  &&  symbind != STB_GLOBAL  &&  symbind != STB_WEAK) {
         DumpElf64Symbol(symname, "***Not function or label (BIND != LOCAL/GLOBAL/WEAK)***", sym);
         continue;                          // BIND attribute not what we want
      }

      if (it_is_the_kernel) {
         // Grab functions, labels and objects
         if (symtype != STT_FUNC  &&  symtype != STT_NOTYPE  &&  symtype != STT_OBJECT) {
            DumpElf64Symbol(symname, "***Not function, label, or object (TYPE != FUNC/NOTYPE/OBJECT)***", sym);
            continue;                          // TYPE attribute not what we want
         }

         // Probably absolute symbol but value too small
         if (sym->st_value < base_addr) {
            DumpElf64Symbol(symname, "***Kernel and absolute symbol < base_addr*****", sym);
            continue;
         }

         // Grab symbols from any section
      }
      else {
         // Grab functions and labels only
         if (symtype != STT_FUNC  &&  symtype != STT_NOTYPE) {
            DumpElf64Symbol(symname, "***Not function or label (TYPE != FUNC/NOTYPE)***", sym);
            continue;                          // TYPE attribute not what we want
         }

         // Symbol in undefined/reserved section
         if (sym->st_shndx == SHN_UNDEF  ||  sym->st_shndx > eh->e_shnum) {
            DumpElf64Symbol(symname, "***Undefined(0)/Reserved(>0xff00) Section***", sym);
            continue;                          // Undefined/Reserved section
         }

         // Throw away symbols for non-executable sections.
         // A section is executable if:
         //    ((sh->sh_type == SHT_PROGBITS) && (sh->sh_flags & SHF_EXECINSTR))
         sh = sh_table + sym->st_shndx;     // Section to which this symbol belongs (pointer addition!)
         if ((sh->sh_type != SHT_PROGBITS) || !(sh->sh_flags & SHF_EXECINSTR)) {
            DumpElf64Symbol(symname, "**Non-executable section**", sym);
            continue;                          // Don't want this one
         }
      }

#if (defined(_PPC) && defined(_32BIT)) || defined(_S390)
      //
      // These platforms have some symbols that are repeated over and over.
      // In general they are contained in other symbols (functions) so I'm
      // not filtering them out for now.
      //
      // * Letext
      //   This is usually a label to some number of NOPs.  It appears to be
      //   used to force alignment (32-bit) for function entry points.
      // * .LT##_0 (where ## are numbers)
      //   These seem to be some sort of branch or offset table, usually located
      //   immediately following the function prologue code.  Prologue code branches
      //   around it and code later on references it by offset.
      // * .L# (where # is a number)
      // * L## (where # is a number and thing between them is a smily face)
      //   These contain initialized data at some offset from the label.  I don't
      //   know what they are for.
#endif

      //
      // It's a keeper!
      //
      // Basically what we have left are either Labels or Functions which are
      // either Local, Global or Weak, and are not generated by gcc.
      // For the kernel they can be in any section (ie. executable or not).
      // For all other modules they can only be in executable sections.
      //
      // The symbol address we return is really an offset.
      // For executable (ie. non-relocatable) modules, st_value is given in
      // terms of a virtual address so we need to convert that to an offset.
      // For relocatable modules, st_value is given as an offset so we
      // just use it as-is.
      //
      NumberOfSymbolsKept++;                // Keping another one
      DumpElf64Symbol(symname, NULL, sym);
      if (module_type == MODULE_TYPE_ELF64_EXEC || module_type == MODULE_TYPE_ELF64_REL_FIXED)
         symoffset = PtrToUint32(PtrSub(sym->st_value, base_addr));
      else
         symoffset = PtrToUint32(sym->st_value);

      if (sym->st_size == 0)
         symtype = SYMBOL_TYPE_LABEL;
      else {
         if (symbind == STB_WEAK)
            symtype = SYMBOL_TYPE_FUNCTION_WEAK;
         else if (symtype == STT_OBJECT)
            symtype = SYMBOL_TYPE_OBJECT;
         else
            symtype = SYMBOL_TYPE_FUNCTION;
      }

      //
      // Calculate where the code for this symbol starts (if needed)
      //
      code_ptr = GetCodePtrForSymbol((char *)Uint32ToPtr(symoffset), &e_sec[0], escnt);
      PrintBytes(code_ptr, (int)sym->st_size);   // Only if debugging harvester

      //
      // Demangle symbol name if asked to do so.
      // ***** Doesn't seem to work on IA64 or PPC so don't do it there *****
      //
#if (!defined(_IA64) && !defined(_PPC))
      if (G(demangle_cpp_names)) {
         demangled_name = cplus_demangle(symname, (DMGL_PARAMS | DMGL_ANSI));
         dbghv(("DEMANGLE: %s -> %s\n", symname, demangled_name));
         if (demangled_name)
            symname = demangled_name;
      }
#endif

      //
      // Call the callback function with each symbol we've decided to keep.
      // Callback function returns 0 if it wants us to continue, non-zero
      // if it wants us to stop.
      //

      rc = SendBackSymbol(symname,          // Symbol name
                          symoffset,        // Symbol offset
                          sym->st_size,     // Symbol size
                          symtype,          // Symbol type
                          sym->st_shndx,    // Section to which the symbol belongs
                          code_ptr,         // Code associated with this symbol
                          cr);              // Context record

      if (demangled_name)
         free(demangled_name);

      if (rc) {
         dbghv(("< GetSymbolsFromElf64ExecutableInternal: SendBackSymbol returned %d. Stopping\n", rc));
         return (NumberOfSymbolsKept);      // Error. Stop enumerating and return
                                            // number of symbols found so far.
      }
   }

   //
   // Done enumerating symbols
   //
   dbghv(("< GetSymbolsFromElf64ExecutableInternal: Done enumerating. Found %d symbols.\n", NumberOfSymbolsKept));

   if (NumberOfSymbolsKept != 0)
      mn->flags &= ~A2N_FLAGS_PLT_SYMBOLS_ONLY;  // There are symbols besides the PLT

   return (NumberOfSymbolsKept);
}


#if SYNTHESIZE_SYMBOLS
int static GetSymbolsInternal(CONTEXT_REC * cr, char * modname, int module_type, bfd * abfd)
{
    int retval = 0;
    extern const bfd_target bfd_elf64_powerpc_vec;
    extern const bfd_target bfd_elf64_powerpcle_vec;
    bool is_elf64_powerpc_target = (abfd->xvec == &bfd_elf64_powerpc_vec)
        || (abfd->xvec == &bfd_elf64_powerpcle_vec);
    if (is_elf64_powerpc_target)
        retval = GetSymbolsUsingBfdInternal(cr, modname, module_type, abfd);
    else
        retval = GetSymbolsFromElf64ExecutableInternal(cr, modname, module_type, abfd);
    return retval;
}
#else
int static GetSymbolsInternal(CONTEXT_REC * cr, char * modname, int module_type, bfd * abfd)
{
    return (GetSymbolsFromElf64ExecutableInternal(cr, modname, module_type, abfd));
}
#endif


//
// GetSymbolsFromElf64Executable()
// *******************************
//
// Harvest symbols from non-stripped ELF64 executables.
// As symbols are found a user callback function is called, once per
// symbol, to allow the user to manipulate/catalog the symbols.
// GetSymbolsFromExecutable() returns when:
// - all symbols are enumerated, or
// - the callback function requests it to stop, or
// - an error is encountered
// For relocatable modules (ex. shared objects) the load address is not used -
// the symbols are in terms of an offset from the text section.
// For non-relocatable modules (ex. executables) the load address is used
// to calculate a symbol offset since the symbols are in terms of a virtual
// address.
//
//
int GetSymbolsFromElf64Executable (CONTEXT_REC * cr, char * modname, int module_type, bfd * abfd)
{
	MOD_NODE * mn;
	mn = (MOD_NODE *) cr->handle;

	mn->li.abfd = abfd;
	mn->li.asym = NULL;
	mn->li.asymcnt = 0;
	mn->flags |= A2N_FLAGS_64BIT_MODULE;	// ELF (64-bit) module
	return (GetSymbolsInternal (cr, modname, module_type, abfd));

}


//
// DumpElf64Symbol()
// *****************
//
void DumpElf64Symbol(char *name, char * comment, Elf64_Sym * sym)
{
   if (!G(debug))
      return;

   if (comment == NULL) {                   // Good symbols
      int  symtype = ELF32_ST_TYPE(sym->st_info);
      int  symbind = ELF32_ST_BIND(sym->st_info);

      if (symtype == STT_NOTYPE) {
         dbghv(("Lbl  "));
		}
		else if (symbind == STB_WEAK) {
         dbghv(("WFn  "));
		}
		else if (symtype == STT_FUNC) {
         dbghv(("Fnc  "));
		}
		else if (symtype == STT_OBJECT) {
         dbghv(("Obj  "));
		}
		else {
         dbghv(("???  "));
      }

		dbghv (("0x%08x  0x%" _LZ64x "  0x%" _LZ64x
			"  0x%02x     0x%02x      0x%04x     %s\n", sym->st_name,
			sym->st_value, sym->st_size, sym->st_info, sym->st_other, sym->st_shndx, name));
   }
   else {                                   // Bad symbols
		dbghv (("     0x%08x  0x%" _LZ64x "  0x%" _LZ64x
			"  0x%02x     0x%02x      0x%04x     %s   %s\n",
             sym->st_name, sym->st_value, sym->st_size, sym->st_info, sym->st_other, sym->st_shndx, name, comment));
   }
   return;
}
