/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2007
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

static unsigned int hash_thread(void * thr);
static unsigned int size_thread(void * thrd);
static int          compare_thread(void * thrd1, void * thrd2);

#include "post.h"
#include "a2n.h"
#include "hash.h"

void get_jtnm(int pid, int tid, thread_t * tp);
void * constNode(void);

// Generic Hash Routines

/******************************/
void * hash_calloc(unsigned int size)
{
   void * p = NULL;

   if (size == 0) return NULL;

   p = malloc(size);

   if ( NULL == p )
   {
      fprintf(stderr, " ** Out of memory in hash_calloc(unsigned int) ***\n");
      exit(1);
   }
   memset(p, 0, size);
   return p;
}

/******************************/
void hash_free(void * p)
{
   if (p) free(p);
   p = NULL;
}

/******************************/
void hash_init(hash_t *     table,
               int          size,
               unsigned int(* hash_f)(void *),
               unsigned int(* size_f)(void *),
               int          (* compare_f)(void *, void *))
{
   table->n_entries = 0;
   table->size      = size;
   table->entries   = hash_calloc(size * sizeof(bucket_t *));
   table->hash_f    = hash_f;
   table->size_f    = size_f;
   table->compare_f = compare_f;
}

/******************************/
void * hash_iterate(hash_t * table,void * (*f)(void *, void *), void * arg)
{
   unsigned int i;

   for (i = 0; i < table->size; i++)
   {
      bucket_t * bp = table->entries[i];

      while (bp)
      {
         arg = f(bp->content, arg);
         bp = bp->next;
      }
   }
   return arg;
}

/******************************/
/* remove entry from hash table */
// remove all objects that are true re f(*,arg)
void hash_remove2(hash_t * table,
                  int (* f)(void *, void *), void * arg)
{
   unsigned int i;

   for (i = 0; i < table->size; i++)
   {
      bucket_t ** bpp = &(table->entries[i]);
      bucket_t *  bp;

      while ((bp = *bpp) != 0)
      {
         if (f(bp->content, arg))
         {                              // remove func(f) true
            *bpp = bp->next;            // wire around bp
            hash_free(bp->content);     // free content
            hash_free(bp);              // free hash entry
            table->n_entries--;         // decrement element count
            continue;                   // bpp already setup
         }
         bpp = &(bp->next);             // next bpp
      }
   }
}

/******************************/
// lookup, remove from table, rtn content ptr(for moving)
void hash_remove(hash_t * table, void * cptr)
{
   int index       = table->hash_f(cptr) % table->size;
   bucket_t ** bpp = &(table->entries[index]);
   bucket_t *  bp;

   while ((bp = * bpp) != 0)
   {
      if (table->compare_f(cptr, bp->content) != 0)
      {
         *bpp = bp->next;               // bypass bp
         hash_free(bp);                 // free hash element (bucket_t)
         return;
      }
      bpp = &(bp->next);
   }
   return;
}

/******************************/
/* remove hash table entries */
void hash_removeall(hash_t *table)
{
   unsigned int i;

   for (i = 0; i < table->size; i++)
   {
      bucket_t ** p = &(table->entries[i]);
      bucket_t * bucket;

      while ((bucket = *p) != 0)
      {
         *p = bucket->next;
         hash_free(bucket->content);
         hash_free(bucket);
         table->n_entries--;
      }
   }
}

/******************************/
/* lookup: present => retn ptr,else NULL */
void * hash_lookup(hash_t * table, void * new)
{
   int index     = table->hash_f(new) % table->size;
   bucket_t ** p = &(table->entries[index]);

   while (*p)
   {
      if (table->compare_f(new, (* p)->content) != 0)
      {
         return(* p)->content;
      }
      p = &((* p)->next);
   }
   return NULL;
}

/******************************/
/* put entry in table. Used when hash_lookup rtns NULL */
void * hash_put(hash_t * table, void * new)
{
   int         ind  = table->hash_f(new) % table->size;
   int         sz   = table->size_f(new);
   bucket_t  * newb = hash_calloc(sizeof(bucket_t));
   void * ptr       = hash_calloc(sz);

   newb->content       = ptr;
   newb->next          = table->entries[ind];

   table->entries[ind] = newb;          // front of list
   table->n_entries++;

   return ptr;
}

// ************** Thread Section ***********************
static hash_t     thread_table;
static thread_t * thread_head = NULL;

/******************************/
static unsigned int hash_thread(void * _thread)
{
   thread_t * t = _thread;

   return(unsigned int)t->tid;
}

/******************************/
static unsigned int size_thread(void * _thread)
{
   return sizeof(thread_t);
}

/******************************/
static int compare_thread(void * _thread1, void * _thread2)
{
   thread_t * tp1 = _thread1;
   thread_t * tp2 = _thread2;
   int eq1, eq2, eq3;

   // equal tid's & valid
   eq1 = tp1->tid == tp2->tid;
   eq2 = tp1->pid == tp2->pid;
   eq3 = (tp1->valid == 1) && (tp2->valid == 1);

   return(eq1 && eq2 && eq3);
}

/******************************/
void thread_table_init(void)
{
   /* hash_init(&thread_table, 65, */
   hash_init(&thread_table, 1023,
             hash_thread, size_thread, compare_thread);
}

/******************************/
// need ADD pid to lookup_thread
thread_t * lookup_thread(int tid, int pid)
{
   thread_t   t;
   thread_t * tp;

   t.tid    = tid;
   t.pid    = pid;
   t.valid  = 1;

   tp = (thread_t *)hash_lookup(&thread_table, &t);
   return tp;
}

/******************************/
thread_t * get_thread_head()
{
   return thread_head;
}

/******************************/
thread_t * insert_thread(int tid, int pid)
{
   thread_t t;
   thread_t * tp;
   char     buf[32];

   t.tid   = tid;
   t.pid   = pid;
   t.valid = 1;

   tp = hash_lookup(&thread_table, &t);

   if (tp == NULL)
   {
      tp = hash_put(&thread_table, &t);
      tp->tid    = tid;
      tp->pid    = pid;
      tp->valid  = 1;

      tp->oc      = 0;
      tp->cfcnt   = 0;
      tp->ring    = 0;
      tp->to      = 0;
      tp->ci      = 0;
      tp->mnm[0]  = '\0';
      tp->tmend   = 0;

      tp->currm = tp->currT = (void *)constNode();
      tp->currT->par = NULL;

      tp->thread_name = strdup("");

      // thread name for tree
      sprintf(buf, "%x_%x", pid, tid);
      tp->currm->mod = strdup(buf);
      tp->currm->mnm = strdup("thrd");

      memset(&(tp->sd), 0, sizeof(SYMDATA) );
      tp->sd_rc = 3;                    // invalidate sym cache

      // thread linked list
      tp->next    = thread_head;
      thread_head = tp;

   }
   return tp;
}



// ************** Process Section ***********************
static hash_t     process_table;

/******************************/
static unsigned int hash_process(void * _process)
{
   process_t * t = _process;

   return(unsigned int)t->pid;
}

/******************************/
static unsigned int size_process(void * _process)
{
   return sizeof(process_t);
}

/******************************/
static int compare_process(void * _process1, void * _process2)
{
   process_t * pp1 = _process1;
   process_t * pp2 = _process2;

   return(pp1->pid == pp2->pid);
}

/******************************/
void process_table_init(void)
{
   hash_init(&process_table, 1023,
             hash_process, size_process, compare_process);
}

/******************************/
process_t * lookup_process(int pid)
{
   process_t   pt;
   process_t * ppt;

   pt.pid = pid;
   ppt    = (process_t *)hash_lookup(&process_table, &pt);

   return ppt;
}

/******************************/
process_t * insert_process(int pid)
{
   process_t   pt;
   process_t * ppt;

   pt.pid = pid;
   ppt    = hash_lookup(&process_table, &pt);

   if (ppt == NULL)
   {
      ppt = hash_put(&process_table, &pt);   // All fields zero

      ppt->pid = pid;
   }
   return ppt;
}
