/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2008
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*! \mainpage

<b>post</b> is the Itrace / Tprof post processor.

The Address-to-Name processor (a2n) is used for name resolution.
Input to <b>post</b> is by .nrm2 files which contain:
<ul>
<li>mte hooks for a2n
<li>from/to branch addresses for callflow
<li>tprof hooks
</ul>

 \verbatim

 0x20/0x21  call_flow(ker/app) : ft[c] fftt[c]
 0x22/0x23  call_flow(ker/app) : fft[c]
 0x24/0x25  call_flow(ker/app) : fb ffb (b=BBCC)
 0x28/0x29  Interrupt / Exception : vcff
 0x2c       cf_pidtid : LX  : tid c
 0x2c       cf_pidtid : Win : pid tid c
 0x2d       cf_pidtid : Win : pid tid c ts*
 0x2e       itracestt cpid ctid sttpid stttid mode
 0x2f       itracestp cpu cfhkcnt    (per processor)

 0x04       void handle_interrupt(trace_t * tr)
 0x10       void handle_tprof_hook(trace_t * tr)
 0x11       void handle_startup_hook(trace_t * tr)
 0x12       void handle_dispatch(trace_t * tr)
 0x19       0x01 void handle_19_1_hook(trace_t * tr)
            0x34 void handle_19_34_hook(trace_t * tr)
            0x63 void handle_pidmap(trace_t * tr)
 0xf9       void handle_f9(trace_t * tr)
 \endverbatim

*/

// DEPRECATED call flow MAJORS
// case 0xaa: // call flow - LX (implicit ring3 addr)
// case 0xab: // call flow - W2K
// case 0xac: // minor=pid int[0]=tid
// case 0xad: // call flow (minor = internum)

#include "post.h"
#include <math.h>
#include "bputil.h"
#include "a2n.h"
#include "bdate.h"
#include "sdate.h"
#include "hash.h"
#include "aixtrc.h"
#if defined(_LINUX)
   #include "piperf_hardware.h"
   #include "perfutil.h"
#endif

#if 0  // zero
   #ifdef _ZOS
      #include <dynit.h>
      #include "binder.h"
   #endif
#endif

typedef char * STRING;
typedef unsigned long Address;

// return from inst_type call
#define INTRPT 0
#define CALL   1
#define RETURN 2
#define JUMP   3
#define IRET   4
#define OTHER  5
#define UNDEFD 6
#define ALLOC  7

char oclab[8][8] =
{ "INTRPT", "CALL", "RETURN", "JUMP", "IRET", "OTHER", "ALLOC"};

char *ku[2] = { "K", "U"};
void noskew(void);

#if 0 // defined(CONFIG_PPC64) || defined(_AIX)
int decode_inst_type( char * iptr );    // return CALL RETURN JUMP IRET or OTHER
#else
// in i386dis[64].c
int inst_type(char * opcode);           // return CALL RETURN JUMP IRET or OTHER
#endif

char * i386_instruction(
                       char * data,     /* binary instruction (up to 16 bytes) */
                       uint64 addr,     /* address */   //STJ64
                       int  * size,     /* inst length */
                       char * buf,
                       int    mode);    /* 16(0), 64(1), 32(2) bit segment, see a2n.h */

void handle_generic_mte_hook(trace_t * tr);
void handle_pidmap(trace_t * tr);
void read_jita2n(process_t * ppt);

void panic(char * why);
#if defined (CONFIG_PPC64) && defined(BUILD_STAP) // Linux
extern int stap_to_nrm2(char *fnm, char *tmp_nrm2);

char 	tmp_nrm2[PATH_MAX];
#endif

GV gv = {0};                            // global variables
GD gd = {NULL};                         // global FILE descriptors

trace_t  * tr = NULL;                   // set by get_next_rec
UINT64 currtt = 0;                      // curr trace time

// output arc : int2f || (int3c == 1)
// int3 is deprecated ???
int int3c = 0;                          // no of int3 hooks
int int3f = 1;                          // output arc
int aout  = 1;                          // arc out: int3f || int3c == 1
int int3x = 0;                          // use to init arc after int3 stt

pTNode root;                            // tprof tick database
int   dlev[5] = {0};

int     hmaj[64 * 1024];
char  * hmajlab[64 * 1024];
int     ssring[4]          = {0};
void    output(void);

char ban[40] = {"================================"};

// default reports
char rlev[16][6] = {
   "1", "13","134","12","123","1234","3", "34", "345"};
int  ropt[16] = {1, 0, 1, 0, 0, 1, 1, 1, 0};   // nt report vector

char tlabx[8][10] = { "", "Process", "Thread", "Module", "Symbol", "Offset"};
char * tlab[8] = { "", "PID", "TID", "MOD", "SYM", "OFF"};

// global double time
double trstt, trstp;                    // trace time: sec = (stp - stt) / cycles/sec

char * kmap   = {""};                   // kernel map file def
char * knmptr = {"/usr/src/linux/vmlinux"};   // kernel file def
char * kas    = "/proc/kallsyms";
char * kmd    = "/proc/modules";
int knew = 0;                           // k => 0x1 kmap => 0x2   kas/kmod => 0x4

int64 kas_off = 0, kas_size = 0;
int64 kmd_off = 0, kmd_size = 0;

int a2n_nsf = 0;                        // Section name instead of NoSymbols/NoSymbolFound */
char * default_jita2n_file_prefix = "log";
char * jitpre = NULL;
char * mrgfnm = NULL;                   // merge  file def

// 0(first look) 1(finished) 2(ptr to unused JREC)
// replace with entry in process struct
//JTNM * pidtnm[64 * 1024];

// a2n rc histogram
int a2nrc[8] = {0};

// call flow stats
int cfbytes  = 0;
int cfbbs    = 0;

unsigned int cfinsts  = 0;
int hbys[66] = {0};                     // log-histogram of cf bytes executed
int hto[66]  = {0};                     // log-histogram of ct "to" distance(signed)

int cpu2pt[64][4]     = {{0}};          // pid tid opid otid
int cpu2stt[64]       = {0};            // dispatch occurred (ia64 cf only)
int cpu2hk[64]        = {0};
thread_t * cpu2tp[64] = {NULL};

int InterNo = -1;                       // interrupt vector ( different for IA64 vs other ??)

int a2ncalls = 0;

// cpu# ring lastopc instcnt @ offset method:module pid_tid_pidname_thrdname lastopcStr linenum
char * farcd = " %2d %s %d  %5d %c %4x %s:%s   %x_%x_%s_%s %s %d\n";

int necnt1 = 0;                         // prev cf hk
int necnt2 = 0;                         //

// aix temp hack
char * gv_tppnm;

#ifdef _AIX
int aix = 1;
#else
int aix = 0;
#endif

int a2ntestflag = 0;
int a2nanysym = 0;

double getTime(int cpu);

char * gf_nrm2;


#if 0 // defined(CONFIG_PPC64) || defined(_AIX)
/************************/
int decode_inst_type( char * iptr )
{
   int oc = OTHER;                      // Other

   unsigned long opcode = *( (unsigned long *)iptr );

   if ( 0x4C000020 == ( opcode & 0xFC0007FE ) )   // Includes bclrl (RETURN-CALL)
   {
      oc = RETURN;                      // RETURN instruction
   }
   else if ( 0x40000000 == ( opcode & 0xF400000 ) )
   {
      oc = ( opcode & 1 ) ? CALL : JUMP;   // LK selects CALL or JUMP instruction
   }
   else if ( 0x4C000420 == ( opcode & 0xFC0007FE ) )
   {
      oc = ( opcode & 1 ) ? CALL : JUMP;   // LK selects CALL or JUMP instruction
   }

   if ( gv.verbose )
   {
      fprintf( gd.msg, "DECODE_INST_TYPE: %08X = %d\n",
               *( (unsigned long *)iptr ),
               oc );
      fflush( gd.msg );
   }
   return(oc);
}
#endif

// ********************
// **** CODE START ****
// ********************

size_t OptVMsg( const char *fmt, ... )
{
   size_t  n = 0;
   va_list args;

   if ( gv.verbose )
   {
      va_start( args, fmt );

      if ( 0 != ( gv_verbose_logmsg & gv.verbose ) && 0 != gd.msg )
      {
         n = vfprintf( gd.msg, fmt, args );
         fflush(gd.msg);
      }
      else                              // if ( 0 != ( gv_verbose_stderr & gv.verbose ) )
      {
         n = vfprintf( stderr, fmt, args );
      }

      va_end( args );
   }
   return( n );
}

size_t OptVMsgLog( const char *fmt, ... )   // OptVMsg + write to log
{
   size_t  n = 0;
   va_list args;

   if ( 0 != gd.msg )
   {
      va_start( args, fmt );
      n = vfprintf( gd.msg, fmt, args );
      va_end( args );
      fflush(gd.msg);
   }

   if ( 0 == gd.msg || ( 0 != gv.verbose &&  0 == ( gv_verbose_logmsg & gv.verbose ) ) )
   {
      va_start( args, fmt );
      n = vfprintf( stderr, fmt, args );
      va_end( args );
   }

   return( n );
}

size_t ErrVMsgLog( const char *fmt, ... )   // Mandatory VMsg + write to log
{
   size_t  n = 0;
   va_list args;

   if ( 0 != gd.msg )
   {
      va_start( args, fmt );
      n = vfprintf( gd.msg, fmt, args );
      va_end( args );
      fflush(gd.msg);
   }

   va_start( args, fmt );
   n = vfprintf( stderr, fmt, args );
   va_end( args );

   return( n );
}

size_t MaskedVMsg( long mask, const char *fmt, ... )
{
   size_t  n = 0;
   va_list args;

   if ( 0 != ( gv.mask & mask ) )
   {
      va_start( args, fmt );

      if ( 0 != ( gv_verbose_logmsg & gv.verbose ) && 0 != gd.msg )
      {
         n = vfprintf( gd.msg, fmt, args );
         fflush(gd.msg);
      }
      else
      {
         n = vfprintf( stderr, fmt, args );
      }

      va_end( args );
   }
   return( n );
}

size_t WriteFlush( FILE *file, const char *fmt, ... )
{
   size_t  n = 0;
   va_list args;

   if ( file != NULL )
   {
      va_start( args, fmt );

      n = vfprintf( file, fmt, args );
      fflush(file);

      va_end( args );
   }
   return( n );
}

/**********************************/
FILE * wopen(char * fnm, char * mode)   // Weak Open of files, failure allowed
{
   FILE * fd = NULL;

   if ( 0 != ( fd = fopen(fnm, mode) ) )
   {
      if ( 'w' == *mode )
      {
         ErrVMsgLog( "Writing %s\n", fnm );
      }
      else
      {
         ErrVMsgLog( "Reading %s\n", fnm );
      }
   }
   else
   {
      MVMsg( gv_mask_Show_Files, ( "Unable to open optional file: %s\n", fnm ) );
   }
   return(fd);
}

/**********************************/
FILE * xopen(char * fnm, char * mode)   // Open with eXit on failure
{
   FILE * fd = NULL;

   if ( 'w' == *mode )
   {
      ErrVMsgLog( "Writing %s\n", fnm );
   }
   else
   {
      ErrVMsgLog( "Reading %s\n", fnm );
   }

   if ((fd = fopen(fnm, mode)) == NULL)
   {
      fprintf(stderr, "\nError opening required file: %s\n", fnm);
      exit(1);
   }
   return(fd);
}

#if defined(USE_LIBOPCODES)

   #include "dis-asm.h"
   #include <stdarg.h>

/* Pseudo FILE object for strings.  */
/**********************************/
typedef struct
{
   char * buffer;
   size_t size;
   char * current;
} SFILE;

/* sprintf to a "stream" */

/**********************************/
static int
stream_sprintf( SFILE *f, const char *format, ... )
{
   int     n;
   va_list args;

   va_start( args, format );

   n = vsprintf( f->current, format, args );

   f->current += n;

   va_end( args );
   return( n );
}

/**********************************/
// Disassemble an instruction
//   target = buffer to receive the output (512+ bytes)
//   insbuf = address of the instruction buffer
//   ip     = current instruction pointer

unsigned long disassemble( char *target, char *insbuf, bfd_vma ip, disassembler_ftype func )
{
   static struct disassemble_info info;
   static disassembler_ftype      disassemble_fn = 0;   // Function pointer
   static SFILE                   sf;

   sf.buffer  =
   sf.current = target;
   sf.size    = 512;

   if ( 0 == disassemble_fn )
   {
      disassemble_fn = func;

      INIT_DISASSEMBLE_INFO( info, &sf, stream_sprintf );

      info.buffer_length = 8;
   }

   info.mach          = bfd_mach_x86_64;
   info.buffer        = (unsigned char *)insbuf;
   info.buffer_vma    = (bfd_vma)ip;    // Must match first parm of call

   return( (*disassemble_fn)( ip, &info ) );
}

#endif

/**********/
/*
   0x20/0x21  call_flow(ker/app) : ft[c] fftt[c]
   0x22/0x23  call_flow(ker/app) : fft[c]
   0x24/0x25  call_flow(ker/app) : fb ffb (b=BBCC)
   0x28/0x29  Interrupt / Exception : vcff
   0x2c       cf_pidtid : LX  : tid c
   0x2c       cf_pidtid : Win : pid tid c
   0x2d       cf_pidtid : Win : pid tid c ts*
   0x2e       itracestt cpid ctid sttpid stttid mode
   0x2f       itracestp cpu cfhkcnt    (per processor)

   // other
   0x04  void handle_interrupt(trace_t * tr) {
   0x10  void handle_tprof_hook(trace_t * tr) {
   0x11  void handle_startup_hook(trace_t * tr) {
   0x12  void handle_dispatch(trace_t * tr) {
   0x19  0x01 void handle_19_1_hook(trace_t * tr) {
         0x34 void handle_19_34_hook(trace_t * tr) {
         0x63 void handle_pidmap(trace_t * tr) {   // 19/63
   0xf9  void handle_f9(trace_t * tr) {
 */

/**********************************/
void mlegend(void)
{
   FILE * fd = gd.micro;

   fprintf(fd, "\n");
   fprintf(fd, "\n Legend for tprof.micro\n");
   fprintf(fd,   " ======================\n\n");
   fprintf(fd, "   ### MOD  Ticks  Ticks%%  ModuleName\n");
   fprintf(fd, "   ### Sym  Ticks  Ticks%%  SymbolName  StartAddr  Size\n");
   fprintf(fd, "   F: SourceFileName\n");
   fprintf(fd, "   T: Offset(Hex)  LineNo  Ticks\n");
   fprintf(fd, "   C: CodeAddr Code\n");
   fprintf(fd, "\n\n");
}

/**********************************/
UINT64 getTraceTime(void)
{
   UINT64 tmtr = 0;                     // curr trace time

   if (tr)
   {
      tmtr = ((UINT64)(tr->thi)) << 32;
      tmtr += (UINT64)(tr->tlo);
   }
   else
   {                                    // @ trace end
      return(currtt);
   }

   return(tmtr);
}

/**********************************/
// A2nGetSymbol wrapper
int getSymbol(unsigned int pid, uint64 addr, SYMDATA * sdp)
{                                       //STJ64
   int         rc = 0;
   char      * cp;
   process_t * ppt;

   // handle pid = -1
   if (pid != 0xffffffff)
   {
      rc = A2nGetSymbol(pid, addr, sdp);

      if (sdp->code != NULL)
      {
         cp = (char *)sdp->code;        // code at addr given
         // move from addr to sym_addr
         sdp->code = cp + (int)(sdp->sym_addr - addr);

         if ( sdp->flags & SDFLAG_SYMBOL_JITTED )
         {
            ppt = insert_process( pid );

            sdp->flags |= ppt->flags;   // Set the flags 32/64-bit flag for JITTED methods for this PID
         }
      }
   }
   else
   {
      // fill in struct
      rc = 3;                           // NoModule
      sdp->owning_pid_name = "MissingPID";
      sdp->owning_pid = pid;
      sdp->mod_name   = "NoModule";
      sdp->mod_addr   = 0;
      sdp->mod_length = 0;
      sdp->sym_name   = "NoSymbol";
      sdp->code       = NULL;
      sdp->src_file   = "MissingSrcFile";
      sdp->src_lineno = -1;
      sdp->process    = NULL;
      sdp->module     = NULL;
      sdp->symbol     = NULL;
      sdp->flags      = 0;
   }
   return(rc);
}

/**********************************/
int getlineno(SYMDATA * s)
{
#ifndef NOLINENOS
   return(s->src_lineno);
#else
   return(-2);
#endif
}

/**********************************/
// ia64 (ring) ia32 (t->ring) ???
void arcflush(thread_t * tp)
{
   SYMDATA * sdp = &(tp->sd);

   cfinsts   += tp->ci;
   tp->cfcnt += tp->ci;

   if ( gv.verbose )
   {
      fprintf( gd.msg,
               farcd,                   // Format string
               tp->cpu,                 // CPU
               ku[tp->ring],            // Ring ( Kernel or User )
               tp->oc,                  // Opcode Type (call/ret/jump/iret/other)
               tp->ci,
               '@',
               tp->offset,
               sdp->sym_name,
               sdp->mod_name,
               tp->pid,
               tp->tid,
               sdp->owning_pid_name,
               tp->thread_name,
               oclab[tp->oc],
               getlineno(sdp) );
   }

   fprintf( gd.arc,
            farcd,                      // Format string
            tp->cpu,                    // CPU
            ku[tp->ring],               // Ring ( Kernel or User )
            tp->oc,                     // Opcode Type (call/ret/jump/iret/other)
            tp->ci,
            '@',
            tp->offset,
            sdp->sym_name,
            sdp->mod_name,
            tp->pid,
            tp->tid,
            sdp->owning_pid_name,
            tp->thread_name,
            oclab[tp->oc],
            getlineno(sdp) );

   fflush(gd.arc);

   tp->ci = 0;
}

/**********************************/
void arcbuf( thread_t * tp,
             SYMDATA * s,
             int tcnt,
             int offset,
             int ring,
             int oc,
             int cpu,
             int oc1)
{
   if (tp->ci >= 1)
   {
      if ( (tp->sd.sym_name != s->sym_name) ||
           (tp->sd.mod_name != s->mod_name) ||
           (tp->ci >= 1000) )
      {
         arcflush(tp);                  // delayed/compacted output must go now
      }
   }

   // create separate CALL event for 131 IA64 LX alloc issue
   if (tp->ci == 0)
   {                                    // only between
      if (oc1 == 7)
      {                                 // not already a CALL
         fprintf( gd.arc,
                  farcd,
                  cpu,
                  ku[ring],
                  0,
                  0,
                  '>',
                  0,
                  s->sym_name,
                  s->mod_name,
                  tp->pid,
                  tp->tid,
                  (char *)s->owning_pid_name,
                  tp->thread_name,
                  "CALL",
                  getlineno(s) );
      }
   }

   if (gv.compact == 0)
   {                                    // new
      cfinsts += tcnt;

      if ( gv.verbose )
      {
         fprintf( gd.msg,
                  farcd,
                  cpu,
                  ku[ring],
                  oc,
                  tcnt,
                  'X',
                  offset,
                  s->sym_name,
                  s->mod_name,
                  tp->pid,
                  tp->tid,
                  (char *)s->owning_pid_name,
                  tp->thread_name,
                  oclab[oc],
                  getlineno(s) );
      }

      // ia64 (ring) ia32 (t->ring) ???
      fprintf( gd.arc,
               farcd,
               cpu,
               ku[ring],
               oc,
               tcnt,
               '@',
               offset,
               s->sym_name,
               s->mod_name,
               tp->pid,
               tp->tid,
               (char *)s->owning_pid_name,
               tp->thread_name,
               oclab[oc],
               getlineno(s) );

      tp->ci = 0;
   }

   else
   {                                    // save
      if (tp->ci == 0)
      {                                 // first bb
         tp->offset = offset;           // offset from 1st bb
         tp->ring   = ring;
      }
      tp->ci  += tcnt;
      tp->oc   = oc;                    // last bb
      tp->cpu  = cpu;
   }
}

/**********************************/
// for writing breaks, vectors, etc.
// ia64 (ring) ia32 (t->ring) ???
void arcout( thread_t * tp,
             int        tcnt,
             char       dir,
             int        offset,
             char     * mnm,
             char     * modnm,
             int        ring,
             int        oc,
             char     * lab,
             int        cpu)
{
   SYMDATA * sdp = &(tp->sd);

   if (tp->ci >= 1)
   {
      arcflush(tp);
   }

   cfinsts += tcnt;

   if (sdp->owning_pid_name == NULL)
   {
      sdp->owning_pid_name = A2nGetProcessName(tp->pid);
   }

   fprintf( gd.arc,
            farcd,
            cpu,
            ku[ring],
            oc,
            tcnt,
            dir,
            offset,
            mnm,
            modnm,
            tp->pid,
            tp->tid,
            sdp->owning_pid_name,
            tp->thread_name,
            "-",
            getlineno(sdp) );
}

/**********************************/
void pt_flush(int pid, int tid)
{
   thread_t * tp;

   tp = insert_thread(tid, pid);

   if ( (aout) && (tp->ci >= 1) )
   {
      fprintf(gd.arc, " ## arcflush %x %x\n", pid, tid);
      fflush(gd.arc);

      arcflush(tp);

      fprintf(gd.arc, " ## arcflush %x %x\n", pid, tid);
      fflush(gd.arc);
   }
}

/**********************************/
void allpt_flush(void)
{
   thread_t * tp;
   int cpu, pid, tid;

   fprintf(gd.arc, " ## allpt_flush\n");
   fflush(gd.arc);

   for (cpu = 0; cpu < 64; cpu++)
   {
      if (cpu2stt[cpu] == 1)
      {
         pid = cpu2pt[cpu][0];
         tid = cpu2pt[cpu][1];
         tp  = cpu2tp[cpu];

         fprintf(gd.arc, " ## pt_flush @ %x %x\n", pid, tid);
         fflush(gd.arc);

         pt_flush(pid, tid);

         fprintf(gd.arc, " ## retn pt_flush @ %x %x\n", pid, tid);
         fflush(gd.arc);
      }
   }
}

/**********************************/
void cf_pidtid(trace_t * tr)
{
   int cpu, pid, tid, opid, otid;
   thread_t * tp;
   int tcnt = 0;

   //  0x2c  tid c           // lx (1 int)
   //  0x2c  pid tid c       // nt
   //  0x2d  pid tid c [ts]* // nt64

   cpu = tr->cpu_no;

   if (tr->ints == 1)
   {                                    // lx
      tid  = tr->minor;
      pid  = A2nGetOwningPid(tid);

      fprintf(gd.arc, " tid %x pid %x\n", tid, pid);
      fflush(gd.arc);

      tcnt = tr->ival[0];
   }
   else
   {                                    // win
      pid  = tr->minor;
      tid  = tr->ival[0];
      tcnt = tr->ival[1];
   }

   if ( (pid == 0) && (tid == 0) )
   {
      tid = cpu;                        // SMP - Idle separation
   }

   if (cpu2stt[cpu] == 1)
   {
      opid = cpu2pt[cpu][0];
      otid = cpu2pt[cpu][1];
      pt_flush(opid, otid);             // arcflush
   }

   tp = insert_thread(tid, pid);

   cpu2stt[cpu]   = 1;
   cpu2pt[cpu][0] = pid;
   cpu2pt[cpu][1] = tid;
   cpu2tp[cpu]    = tp;

   arcout(tp, tcnt, '?', 1, "callflow", "disp", 0, 0, "-disp-", cpu);
}

/**********************************/
void handle_dispatch(trace_t * tr)
{
   unsigned int pid;
   unsigned int tid;
   int cpu, tcnt;

   //      minor
   // win: pid   ts tid tcnt
   // lx:  tid   ts tcnt

   cpu  = tr->cpu_no;
   pid  = tr->minor;
   tid  = tr->ival[0];
   tcnt = tr->ival[1];                  //

   if ( (pid == 0) && (tid == 0) )
   {
      tid = cpu;
   }
}

/**********************************/
// not a callflow hook
void handle_interrupt(trace_t * tr)
{
   int cpu, mnr, tcnt;

   // pid mnr ts tid tcnt
   cpu  = tr->cpu_no;
   mnr  = tr->minor;

   if (cpu2stt[cpu] == 0) return;

   if (mnr == 0x80000001)
   {
      if (tr->ints == 1)
      {                                 // IA64 only
         tcnt = tr->ival[0];
      }
   }
}

/**********************************/
void handle_19_1_hook(trace_t * tr)
{
   char modnm   [MAX_STR_LEN];
   unsigned int pid;
   int size, rc, csum, tstp;
   uint64 addr;                         //STJ64
   static int a2nerr = 0;

   // int : pid size tstp csum alo [ahi]
   // str : modnm

   pid  = tr->ival[0];
   size = tr->ival[1];
   tstp = tr->ival[2];
   csum = tr->ival[3];

   addr = ((uint64)tr->ival[4]) & 0x0ffffffff;   // low //STJ64
   if (tr->ints == 6)
   {                                    // 64
      addr = ( ( (uint64)tr->ival[5] ) << 32) | addr;   //STJ64
   }

   memcpy(modnm, &tr->sval[0], tr->slen[0]);
   modnm[tr->slen[0]] = 0;              /* null terminate it*/

   rc = A2nAddModule( pid,
                      addr,             // mod stt address //STJ64
                      size,
                      modnm,
                      tstp,             // timestamp(or 0)
                      csum,             // checksum(or 0)
                      NULL);

   if (rc != 0)
   {
      a2nerr++;
      if (a2nerr <= 5)
      {
         fprintf(stderr, " A2nAddModule() ERROR rc %d\n", rc);
      }
      else
      {
         if (0 == a2nerr % 100)
         {
            fprintf(stderr, " %d A2nAddModule() ERRORs rc %d\n", a2nerr,rc);
         }
      }
   }

   return;
}

/**********************************/
void a2ntest(void)
{
   int     pid;
   char    mnm[100];
   uint64  addr;                        //STJ64
   int     size, rc;
   SYMDATA sd;
   SYMDATA * sdp = &sd;
   // char          *owning_pid_name;  // nm "owning" process
   // unsigned int  owning_pid;        // pid of "owning" process
   // char          *mod_name;         // Nm of module
   // char          *mod_addr;         // Mod starting (load) virtual address
   // unsigned int  mod_length;        // Mod length (in bytes)
   // char          *sym_name;         // Sym name
   // char          *sym_addr;         // Sym starting virtual address
   // unsigned int  sym_length;        // Sym length (in bytes)

   pid  = 17;
   addr = 0x40000;                      //STJ64
   size = 0x1000;

   rc = A2nAddModule(pid, 0x30000, 0x1000,   //STJ64
                     "/home/bob/cust/aixtools/post.exe",
                     0, 0, NULL);
   OptVMsg("\n *** A2nAddModule rc %d\n", rc);

   pid  = 17;
   addr = 0x5c000;                      //STJ64
   size = 0x1000;

   OptVMsg("\n *** JIT load Data\n");
   OptVMsg(" JIT load Data\n");
   OptVMsg(" pid      %x\n", pid);
   OptVMsg(" stt_addr %016" _P64 "X\n", addr);   //STJ64
   OptVMsg(" size     %x\n", size);

   strcpy(mnm, "java/sun/io/writeByte(int c)Z;");
   rc = A2nAddJittedMethod(pid, mnm, addr, (int)size, NULL);   //STJ64
   OptVMsg(" A2nAddJittedMethod rc %d\n", rc);

   addr = 0x5c150;                      //STJ64
   rc = getSymbol(pid, addr, sdp);
   OptVMsg("\n *** getSymbol rc %d\n", rc);

   OptVMsg("\n JIT lookup input Data\n");
   OptVMsg(" pid      %x\n", pid);
   OptVMsg(" addr     %016" _P64 "X\n", addr);   //STJ64

   OptVMsg("\n JIT lookup Results\n");
   OptVMsg("     tgid %8x\n",           sdp->owning_pid);
   OptVMsg("    modnm <%s>\n",          sdp->mod_name);
   OptVMsg("      mnm <%s>\n",          sdp->sym_name);
   OptVMsg("  symaddr %016" _P64 "X\n", sdp->sym_addr);   //STJ64
   OptVMsg("   offset %016" _P64 "X\n", addr - sdp->sym_addr);   //STJ64

   exit(-1);
}


#if defined(_LINUX)
   #if defined(_AMD64)
      #include <sys/user.h>
      #include <asm/vsyscall.h>
const uint64_t vsyscall32_base = (uint64_t)0xFFFFE000;
const uint64_t vsyscall32_end  = (uint64_t)0xFFFFE000 + PAGE_SIZE;
const uint64_t vsyscall64_base = (uint64_t)VSYSCALL_START;
const uint64_t vsyscall64_end  = (uint64_t)VSYSCALL_END;
   #elif defined(_X86)
const uint64_t vsyscall32_base = (uint64_t)0xFFFFE000;
const uint64_t vsyscall32_end  = (uint64_t)0xFFFFF000;
const uint64_t vsyscall64_base = 0;
const uint64_t vsyscall64_end  = 0;
   #else
const uint64_t vsyscall32_base = 0;
const uint64_t vsyscall32_end  = 0;
const uint64_t vsyscall64_base = 0;
const uint64_t vsyscall64_end  = 0;
   #endif
#endif


/**********************************/
int initSymUtil()
{
   unsigned long v;
   int rc;

#if defined(_LINUX)
   int64  * pzap;
   FILE   * nrmfh;
   char     trace_header[512];
#endif

   // options on NoSymbolFound
   // def           : NoSymbolFound
   // section name  : e.g. .text
   // null          : NULL
   // SymbolQuality
   // any symbol
   // (def)debug symbols (all)
   // SearchPath (list of dirs)
   // LX : def
   // in image in addMod, *.map, map in other dirs
   // NT :
   // def(env vars), add some


   A2nInit();                           // REQUIRED for _ZOS, harmless for all others

   v = A2nGetVersion();
   fprintf(gd.msg, " A2nVersion: %d.%d.%d\n",
           (int)A2N_VERSION_MAJOR(v), (int)A2N_VERSION_MINOR(v), (int)A2N_VERSION_REV(v));

   // a2n debug
   if (gv.a2ndbg == 1)
   {
      A2nSetDebugMode(DEBUG_MODE_ALL);
   }

   if (a2nanysym == 1)
   {
      A2nSetSymbolQualityMode(MODE_ANY_SYMBOL);
   }

   // a2n rename duplicate symbols
   //  BADDOG. add/stub to LX & AIX
#ifdef _WINDOWS
   if (gv.a2nrdup == 1)
   {
      A2nRenameDuplicateSymbols();
   }
#endif

   // standalone a2n test
   if (a2ntestflag == 1)
   {
      // test a2n w/o nrm2
      a2ntest();
   }

   if (gv.sympath != NULL)
   {                                    // List of paths
      A2nSetSymbolSearchPath(gv.sympath);
   }

   // Tell A2N what to return if NoSymbols or NoSymbolFound:
   //   a2n_nsf = 1: Return NoSymbols or NoSymbolFound
   //   a2n_nsf = 0: Return <<section_name>> or NoSymbols
   if (a2n_nsf)
   {
      A2nReturnNoSymbolFoundOnNSF();
   }
   else
   {
      A2nReturnSectionNameOnNSF();
   }

   // add post option to error control
   A2nSetErrorMessageMode(4);           // E:1 W:2 I:3 A:4
   A2nLazySymbolGather();
   // only if -off option
   A2nGatherCode();                     // A2nDontGatherCode();
   A2nSetSystemPid(0);                  // System Pid

#if defined(_LINUX)
   // Add the VSYSCALL virtual address range
   if (vsyscall32_base)
   {
      int vsyscall32_size = (int)(vsyscall32_end - vsyscall32_base);
      A2nAddModule(0, vsyscall32_base, vsyscall32_size, A2N_VSYSCALL32_STR, TS_VSYSCALL32, 0, NULL);
   }
   if (vsyscall64_base)
   {
      int vsyscall64_size = (int)(vsyscall64_end - vsyscall64_base);
      A2nAddModule(0, vsyscall64_base, vsyscall64_size, A2N_VSYSCALL64_STR, TS_VSYSCALL64, 0, NULL);
   }
#endif


#ifndef NOLINENOS
   #if defined(WIN32) || defined(WINDOWS) || defined(_LINUX)
   if ( (gv.off == 1) || (gv.arc) )
   {
      A2nReturnLineNumbers();
   }
   #endif
#endif

#if defined(_LINUX)
   if (2 == (knew & 2) )
   {
      rc = A2nSetKernelMapName(kmap);
      if (rc != 0)
      {
         fprintf(gd.msg, " post: A2nSetKernelMapName() rc %d\n", rc);
         fprintf(stderr, " ================================\n");
         fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", kmap);
         fprintf(stderr, " ================================\n");
         exit(-1);
      }
   }

   if (1 == (knew & 1) )
   {
      rc = A2nSetKernelImageName(knmptr);
      if (rc != 0)
      {
         fprintf(gd.msg, " post: A2nSetKernelImageName() rc %d\n", rc);
         fprintf(stderr, " ================================\n");
         fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", knmptr);
         fprintf(stderr, " ================================\n");
         exit(-1);
      }
   }

   if (4 == (knew & 4))
   {
      rc = A2nSetKallsymsFileLocation(kas, 0, 0);
      if (rc != 0)
      {
         fprintf(gd.msg, " post: A2nSetKallsymsFileLocation() rc %d (%s)\n", rc, A2nRcToString(rc));
         fprintf(stderr, " ================================\n");
         fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", kas);
         fprintf(stderr, " ================================\n");
         exit(-1);
      }
      rc = A2nSetModulesFileLocation(kmd, 0, 0);
      if (rc != 0)
      {
         fprintf(gd.msg, " post: A2nSetModulesFileLocation() rc %d (%s)\n", rc, A2nRcToString(rc));
         fprintf(stderr, " ================================\n");
         fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", kmd);
         fprintf(stderr, " ================================\n");
         exit(-1);
      }
   }
   else
   {
      nrmfh = xopen(gf_nrm2, "r");
      if (nrmfh)
      {
         fseek(nrmfh, 0, SEEK_SET);
         fread(trace_header, sizeof(char), 512, nrmfh);
         pzap = (int64 *)&trace_header[offsetof(PERFHEADER, kallsyms_offset)];
         kas_off  = *pzap;              // offset to copy of kallsyms
         pzap++;
         kas_size = *pzap;              // size of kallsyms
         pzap++;
         kmd_off  = *pzap;              // offset to copy of modules
         pzap++;
         kmd_size = *pzap;              // size of modules
         OptVMsg("kas_off: 0x%016"_P64x", kas_size: 0x%016"_P64x", kmd_off: 0x%016"_P64x", kmd_size: 0x%016"_P64x"\n",
                 kas_off, kas_size, kmd_off, kmd_size);
         fclose(nrmfh);
      }
      if (kas_size != 0 && kmd_size != 0)
      {
         rc = A2nSetKallsymsFileLocation(gf_nrm2, kas_off, kas_size);
         if (rc != 0)
         {
            fprintf(gd.msg, " post: A2nSetKallsymsFileLocation() rc %d (%s)\n", rc, A2nRcToString(rc));
            fprintf(stderr, " ================================\n");
            fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", gf_nrm2);
            fprintf(stderr, " ================================\n");
            exit(-1);
         }
         rc = A2nSetModulesFileLocation(gf_nrm2, kmd_off, kmd_size);
         if (rc != 0)
         {
            fprintf(gd.msg, " post: A2nSetModulesFileLocation() rc %d (%s)\n", rc, A2nRcToString(rc));
            fprintf(stderr, " ================================\n");
            fprintf(stderr, "\n FATAL ERROR : %s Not Found\n\n", gf_nrm2);
            fprintf(stderr, " ================================\n");
            exit(-1);
         }
      }
   }
#endif

   if (mrgfnm != NULL)
   {
      rc = A2nLoadMsiFile(mrgfnm);
      if (rc != 0)
      {
         fprintf(gd.msg, " post: A2nLoadMsiFile() rc %d\n", rc);
      }
   }

   if (gv.nv == 1)
   {
      A2nDontValidateKernel();
   }

   OptVMsg(" Leaving initSym\n");

   return 0;
}

/**********************************/
uint64 str2uint64(unsigned char * cp)
{
   uint64       d = 0;
   unsigned int r;

#ifdef _ZOS
   sscanf(cp, "%lx", &d);
   return(d);
#endif

   while ((r = *cp) != '\0')
   {
      if (     (r & 0xf0) == 0x30)  r = r & 0xf;   /* 0-9 */
      else if ((r & 0xf0) == 0x60)  r = (r & 0x7) + 9;   /* a-f */
      else if ((r & 0xf0) == 0x40)  r = (r & 0x7) + 9;   /* A-F */
      else                         r = 0;   /* 0xX */
      d = (d << 4) + r;
      cp++;
   }
   return(d);
}


/**********************************/
char * hex2bin(char * cp, char * rec)
{
   int i, d, d1 = 0, d2 = 0, len;
   char h1, h2;

   int mcnt = 0;
   static int cnt = 0;

   cnt++;
   len = (int)strlen(rec) - 1;

   if (cnt <= mcnt)
   {
      OptVMsg(" len %d\n", len);    // FIXME!!!!!  Add check for oversize line
   }

   for (i = 0; i < len; i += 2)
   {
      h1 = *rec;
      rec++;
      h2 = *rec;
      rec++;

      if ( (h1 >= '0') && (h1 <= '9') ) d1 = h1 - '0';
      if ( (h1 >= 'a') && (h1 <= 'f') ) d1 = h1 - 'a' + 10;
      if ( (h2 >= '0') && (h2 <= '9') ) d2 = h2 - '0';
      if ( (h2 >= 'a') && (h2 <= 'f') ) d2 = h2 - 'a' + 10;

      d = 16 * d1 + d2;

      *cp = (unsigned char)d;
      cp++;

      if (mcnt >= cnt)
      {
         OptVMsg(" i %d h1 %c h2 %c\n", i, h1, h2);
         OptVMsg("   - d1 %x d2 %x d %x\n", d1, d2, d);
      }
   }
   return(cp);
}

/**********************************/
void read_jita2n( process_t * ppt )
{
   // Called by jitrec, & 19/41 & 19/42
   // use by time | id

   char fnm[ 128];
   char ftnm[128];
   char hrec[128];                      // hex code records
   char rec[2048];                      // jit or jtnm records
   FILE * fd;
   unsigned int tlo, thi;               // trace time
   unsigned int flo   = 0;              // file time
   unsigned int fhi   = 0;
   unsigned int minlo = 0;              // min file time
   unsigned int minhi = 0;
   char ** tokarr;
   char * mnm;
   char   mnmbuf[2048];

   uint64 addr;
   int size;

   int saw_printable_hex = 0;
   int gave_msg = 0;

   int  tcnt  = 0;
   int  rcnt  = 0;                      // jitted method records
   int  dcnt  = 0;                      // delayed records

   JREC      * pJREC    = NULL;
   JREC      * prevJREC = NULL;
   JTNM      * pJTNM    = NULL;
   JTNM      * prevJTNM = NULL;

   // Jitted Code
   char * cp;
   int    cflag;

   // thread names
   int tid;
   char * tnm;
   int usenow = 0;

   int bytime;
   int byid;
   int tnmcnt = 0;

   int pid = ppt->pid;

#if defined(_PPC64) && !defined(_LINUX) // jit & trace time problems
   // jit  : Read and use all. breaks if jit space reused
   // jtnm : Skip. Nothing better than wrong.
   usenow = 1;
#endif

   tlo = tr->tlo;                       // Current trace time
   thi = tr->thi;

   ppt->read   = 1;
   ppt->trcid  = gv.trcid;              // for trace On/Off

   // create file names (chk if files exit)
   sprintf(ftnm, "%s-jtnm_%u",   jitpre, pid);
   sprintf(fnm,  "%s-jita2n_%u", jitpre, pid);

   // THREAD NAME file
   if ( !usenow )
   {
      if ((fd = wopen(ftnm, "rb")) != NULL)
      {
         int tnmid = 0;
         thread_t * tp;

         prevJTNM = NULL;

         while (fgets(rec, 2048, fd) != NULL)
         {
            tokarr = (char **)tokenize(rec, &tcnt);

            if (tcnt == 5)
            {                           // normal rec
               tnmcnt++;
               fhi = strtoul(tokarr[0], NULL, 16);   // file time
               flo = strtoul(tokarr[1], NULL, 16);
               tid = strtoul(tokarr[2], NULL, 16);
               tnm = tokarr[3];

               OptVMsg(" tnmFile %8x %8x %4x %s %s\n",
                       fhi, flo, tid, tnm, tokarr[4]);

               // early jtnm's -> thread hash table
               bytime = (fhi < thi) || ( (fhi == thi) && (flo < tlo) );
               byid   = tnmid < gv.trcid;

               if ( byid || ( (gv.trcid == 0) && bytime ) )
               {
                  ppt->jtnmUsed++;

                  tp  = insert_thread(tid, pid);

                  if (strcmp(tokarr[4], "START") == 0)
                  {
                     tp->thread_name = strdup(tnm);
                  }
                  else
                  {
                     tp->thread_name = "";
                  }
               }

               // save for time matching
               else
               {
                  ppt->jtnmSaved++;

                  pJTNM = (JTNM *)calloc(1, sizeof(JTNM));

                  if ( NULL == pJTNM )
                  {
                     fprintf(stderr, " ** Out of memory in read_jita2n(int) ***\n");
                     exit(1);
                  }

                  if ( prevJTNM  )
                  {
                     prevJTNM->next = pJTNM;   // add to end
                  }
                  else
                  {
                     ppt->tnm_ll = pJTNM;   // 1st tnm entry on pid
                  }
                  prevJTNM = pJTNM;

                  pJTNM->next  = NULL;
                  pJTNM->tlo   = flo;
                  pJTNM->thi   = fhi;
                  pJTNM->tid   = tid;

                  if (strcmp(tokarr[4], "END") == 0)
                  {
                     pJTNM->tnm = "";
                  }
                  else
                  {
                     pJTNM->tnm = strdup(tnm);
                  }
               }
            }
            else if (tcnt == 2)
            {                           // SeqID
               if (strcmp(tokarr[0], "SeqNo") == 0)
               {
                  tnmid = atoi(tokarr[1]);
               }
            }
         }
         OptVMsgLog(" %4d log-jtnm records for PID %d (0x%x). Used %d, Saved %d.\n",
                    tnmcnt, pid, pid, ppt->jtnmUsed, ppt->jtnmSaved);

         fflush(gd.msg);
         fclose(fd);
      }
   }

   // JITA2N File

   if ((fd = wopen(fnm, "rb")) != NULL)
   {
      int seqNo;                        // jit record SeqID

      OptVMsg(" pid %x TraceTime %8x %8x\n", pid, thi, tlo);

      seqNo    = 0;
      prevJREC = NULL;

      while (fgets(rec, 2048, fd) != NULL)
      {
         int use = 0;

         tokarr = (char **)tokenize(rec, &tcnt);

         if ( tcnt > 0 && '(' == tokarr[0][0] )   // Part of header detected
         {
            if ( 0 == strcmp( tokarr[0], "(Platform" ) )
            {
               ppt->flags |= ( strstr( tokarr[2], "64" ) )
                             ? SDFLAG_MODULE_64BIT
                             : SDFLAG_MODULE_32BIT;
            }
         }

         else if (tcnt == 6)
         {                              // normal rec
            int len1, len2;

            // check format : 8 characters
            len1 = (int)strlen(tokarr[0]);
            len2 = (int)strlen(tokarr[1]);

            if ( (len1 == 8) && (len2 == 8) )
            {
               fhi = strtoul(tokarr[0], NULL, 16);   // file time
               flo = strtoul(tokarr[1], NULL, 16);

               rcnt++;
               if (rcnt == 1)
               {
                  minhi = fhi;
                  minlo = flo;
               }

               addr = str2uint64((uchar *)tokarr[2]);
               size = (int)str2uint64((uchar *)tokarr[3]);

               cflag = atoi(tokarr[4]);

               if ( cflag & jrec_cflag_Code )
               {                        // HEX code in file
                  if (!saw_printable_hex)
                  {
                     cflag = 0;         // Make it look like there's no code
                     if (!gave_msg)
                     {
                        ErrVMsgLog(" %s missing \"BINARY DATA IN PRINTABLE HEX\" line. Ignoring jitted code.\n", fnm);
                        gave_msg = 1;
                     }
                  }
               }

               // mnm + hexAddr (for rejitted resolution)
               mnm = tokarr[5];

               if ( (gv.off == 1) && (gv.undup == 0) )
               {                        // gv.off never set to 2 now
                  sprintf(mnmbuf, "%s_%"_P64"x", mnm, addr);

                  mnm = mnmbuf;
               }

               if ( cflag & jrec_cflag_Code )   // HEX code in file
               {
                  int ii;
                  char * cpStart;

                  cpStart = cp = calloc(1, size + 32);   // code space

                  if ( NULL == cp )
                  {
                     fprintf(stderr, " ** Out of memory in read_jita2n(int) ***\n");
                     exit(1);
                  }

                  for ( ii = 0; ii < size; ii += 32 )   // Up to 32 bytes (64 Hex digits) per line
                  {
                     if ( fgets( hrec, 128, fd ) )
                     {
                        cp = hex2bin(cp, hrec);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if ( size != (int)(cp - cpStart) )
                  {
                     ErrVMsgLog(" Incomplete record from jita2n ignored: size = %d, expected size = %d\n",
                             (int)(cp - cpStart), size);

                     free( cpStart );

                     cpStart = NULL; // Error reading binary data, ignore all of it

                     cflag &= ~jrec_cflag_Code;
                  }
                  cp = cpStart;         // restore cp to code ptr
               }
               else
               {
                  cp = NULL;
               }

               // time or id based
               bytime = (fhi < thi) || ( (fhi == thi) && (flo < tlo) );
               byid   = seqNo < gv.trcid;

               if ( gv.nv || usenow || byid || ( (gv.trcid == 0) && bytime ) )
               {
                  use = 1;
                  ppt->jitUsed++;
                  OptVMsg( " A2nAddJitted %08x %08x %s\n",
                           fhi, flo, mnm);

                  if (gd.show)
                  {
                     fprintf( gd.show, " A2nAddJitted %08x %08x %s\n",
                              fhi, flo, mnm);
                  }
                  switch (cflag)        // Use jrec_cflag_????
                  {
                  case 0:               // JIT wo Code
                     A2nAddJittedMethod(pid, mnm, addr, size, NULL);
                     break;

                  case 1:               // JIT w Code
                     A2nAddJittedMethod(pid, mnm, addr, size, cp);
                     break;

                  case 2:               // MMI wo Code
                     A2nAddMMIMethod(pid, mnm, addr, size, NULL);
                     break;

                  case 3:               // MMI w Code
                     A2nAddMMIMethod(pid, mnm, addr, size, cp);
                     break;
                  }
               }

               else
               {                        // save
                  ppt->jitSaved++;
                  dcnt++;

                  pJREC = (JREC *)calloc(1, sizeof(JREC));

                  if ( NULL == pJREC )
                  {
                     fprintf(stderr, " ** Out of memory in read_jita2n(int) ***\n");
                     exit(1);
                  }

                  if (prevJREC == NULL)
                  {
                     ppt->jit_ll    = pJREC;   // Start the new linked list
                  }
                  else
                  {
                     prevJREC->next = pJREC;   // add at end to keep in time order
                  }

                  //OptVMsg(" savedCnt %4d size %x\n", dcnt, size);

                  // fillin struct
                  pJREC->next  = NULL;
                  pJREC->tlo   = flo;
                  pJREC->thi   = fhi;
                  pJREC->addr  = addr;
                  pJREC->size  = size;
                  pJREC->mnm   = strdup(mnm);
                  pJREC->cflag = cflag;
                  pJREC->code  = cp;
                  pJREC->seqNo = seqNo;

                  prevJREC = pJREC;     // keep last ptr to link next @ end
               }
            }
         }

         // default SeqNo = 0, only > 0 for IA32-IA64 machines

         else if ( tcnt == 2 )
         {                              // jitfile SeqNo record
            if (strcmp(tokarr[0], "SeqNo") == 0)
            {
               seqNo = atoi(tokarr[1]);
            }
         }

         // Look for marker. If we don't see it then we'll
         // skip reading jitted code. Most likely the file was
         // generated using and old JProf and the code is in binary.

         else if ( tcnt == 5 )
         {
            if ( (strcmp(tokarr[0], "BINARY"   ) == 0) &&
                 (strcmp(tokarr[1], "DATA"     ) == 0) &&
                 (strcmp(tokarr[2], "IN"       ) == 0) &&
                 (strcmp(tokarr[3], "PRINTABLE") == 0) &&
                 (strcmp(tokarr[4], "HEX"      ) == 0) )
            {
               saw_printable_hex = 1;
            }
         }
      }
      OptVMsgLog(" %4d log-jita2n records for PID %d (0x%x). Used %d, Saved %d.\n",
                 rcnt, pid, pid, ppt->jitUsed, ppt->jitSaved);
      OptVMsg("  mintm %8x %8x\n", minhi, minlo);
      OptVMsg(" lasttm %8x %8x\n", fhi, flo);

      fflush(gd.msg);
      fclose(fd);
   }
}

/**********************************/
void panic(char * why)
{
   fprintf(stderr, "%s\n", why);
   fprintf(gd.msg, "%s\n", why);
   exit(-1);
}

/**********************************/
void cfstats()
{
   int      i;
   unsigned int j;
   double   rat, totb, tote, tot;
   int64    readbytes = getReadBytes();
   int64    fsize;

   if (cfbbs >= 1)
   {
      fprintf(gd.msg, "\n Call Flow Stats\n");
      fprintf(gd.msg, " %10d  BBs\n",           cfbbs);
      fprintf(gd.msg, " %10d  ExecutedBytes\n", cfbytes);
      fprintf(gd.msg, " %10d  Inst\n",          cfinsts);
      rat = cfbytes / (double)cfbbs;
      fprintf(gd.msg, " %10.1f  ExecutedBytes / BB\n",   rat);
      rat = cfinsts / (double)cfbbs;
      fprintf(gd.msg, " %10.1f  Insts / BB\n",           rat);
      rat = cfbytes / (double)cfinsts;
      fprintf(gd.msg, " %10.1f  ExecutedBytes / Inst\n", rat);

      fsize = nrm2_size();
      fprintf(gd.msg, "\n %016" _P64 "X  FileSize\n", fsize);
      fprintf(gd.msg, "\n %016" _P64 "X  BytesRead\n", readbytes);
      rat   = fsize / (double)cfinsts;
      fprintf(gd.msg, " %10.1f  FileBytes / Inst\n", rat);
      rat   = readbytes / (double)cfinsts;
      fprintf(gd.msg, " %10.1f  BytesRead / Inst\n", rat);

      fprintf(gd.msg, "\n BytesExecuted Histogram\n");
      fprintf(gd.msg, "\n       Freq   Bits      Range   %s", "Cum%");
      fprintf(gd.msg, "\n       ====   ====      =====   ====\n" );
      tote = totb = tot = 0;

      for (i = 0; i <= 32; i++)
      {
         tot += hbys[i];
      }
      for (i = 0; i <= 32; i++)
      {
         if (hbys[i] >= 1)
         {
            j = 1 << i;
            totb += hbys[i] * i;
            tote += hbys[i];
            rat = 100.0 * tote / tot;

            fprintf(gd.msg, " %10d %6d %10d %6.1f\n",
                    hbys[i], i, j, rat);
         }
      }
      fprintf(gd.msg, "       ====   ====      =====   ====\n" );
      fprintf(gd.msg, " %10.0f\n", tote);

      tote = totb = tot = 0;

      fprintf(gd.msg, "\n BytesToNextBB Histogram\n");
      fprintf(gd.msg, "\n       Freq   Bits      Range   %s", "Cum%");
      fprintf(gd.msg, "\n       ====   ====      =====   ====\n" );

      for (i = 0; i <= 32; i++)
      {
         tot += hto[i];
      }
      for (i = 0; i <= 32; i++)
      {
         if (hto[i] >= 1)
         {
            j = 1 << i;
            totb += hto[i] * i;
            tote += hto[i];
            rat = 100.0 * tote / tot;
            fprintf(gd.msg, " %10d %6d %10u %6.1f\n",
                    hto[i], i, j, rat);
         }
      }
      fprintf(gd.msg, "       ====   ====      =====   ====\n" );
      fprintf(gd.msg, " %10.0f\n", tote);
   }
}

/**********************************/
void thisto( uint64 fr, uint64 to)
{                                       //STJ64
   int i = 1;
   size_t del;

   // absolute
   if (to > fr) del = (size_t)( to - fr );   //STJ64
   else         del = (size_t)( fr - to );   //STJ64

   if (del >= 1)
   {
      while (del > 1)
      {
         del = del >> 1;
         i++;
      }
      hto[i]++;
   }
}

/**********************************/
void bhisto(int bytes)
{
   int i = 1;
   int tmp = bytes;

   if (tmp >= 1)
   {
      while (tmp > 1)
      {
         tmp = tmp >> 1;
         i++;
      }
      hbys[i]++;
   }
}

/**********************************/
int count_inst(uint64 prevTo, int bytes, SYMDATA * sdp)
{
   int  b       = 0;
   int  size    = 0;
   int  icnt    = 0;
   int  inval   = 0;
   int  istrlen = 0;

   char * cp;
   char * iptr;
   char   istr[10] = {""};
   char * opc = "-";
   char   buf[2048];

   uint64 addr;                         //STJ64

   if (bytes >= 0)
   {
      opc  = "????";
      strcpy(buf, "????");

      // getSym for TO address of last callflow hook

      if (sdp->code != NULL)
      {
         iptr = sdp->code;              // code at beginning of symbol
         iptr += (int)(prevTo - sdp->sym_addr);   // adjust code ptr to "prevTo"
         addr = prevTo;                 // virtual addr of this code

         while (b <= bytes)
         {
            icnt++;

            opc = (char *)i386_instruction((char *)iptr,
                                           (uint64)addr, 
                                           &size, 
                                           buf,
                                           sdp->flags & ( SDFLAG_MODULE_32BIT | SDFLAG_MODULE_64BIT ));   //STJ64

            inval = (int)(strcmp(opc, "INVALID") == 0);

            if ( inval )
            {
               OptVMsgLog("\n INVALID_INSTRUCTION\n");
               OptVMsgLog("   prevTo %016" _P64 "X  err@addr %016" _P64 "X ", prevTo, addr);   //STJ64
               OptVMsgLog(" %s:%s\n", sdp->sym_name, sdp->mod_name);

               cp = sdp->code;
               cp += (int)(prevTo - sdp->sym_addr);   // prevTo

               while (cp < iptr + 8)
               {
                  unsigned char c;

                  if (cp == iptr)
                  {
                     OptVMsgLog("-");
                  }

                  c = (*cp) & 0xff;
                  cp++;
                  OptVMsgLog("%02x", c);
               }
               OptVMsgLog("\n");

               if (gv.showss >= 1)
               {
                  fprintf(gd.arc,
                          " ss: %016" _P64 "X %2d %s\n",   //STJ64
                          addr, size, "INVALID");
               }
            }

            if (gv.showss >= 1)
            {
               if (gv.showss == 2)
               {                        // hex instrs ??
                  int i, ilen;
                  char * ip;
                  unsigned char c;

                  istrlen = 8;
                  cp = (char *)iptr;    // start at to
                  ip = &istr[0];

                  ilen = size;
                  if (ilen > 4) ilen = 4;

                  for (i = 0; i < ilen; i++)
                  {
                     c = (*cp) & 0xff;
                     cp++;
                     sprintf(ip, "%02x", c);
                     ip += 2;
                  }
                  istr[2 * ilen] = 0;
               }

               if ( gv.verbose )
               {
                  fprintf(gd.msg,
                          " ss: %016" _P64 "X %2d %*s %s\n",   //STJ64
                          addr, size, istrlen, istr, buf);
                  fflush(gd.msg);
               }

               fprintf(gd.arc,
                       " ss: %016" _P64 "X %2d %*s %s\n",   //STJ64
                       addr, size, istrlen, istr, buf);
            }

            b    += size;
            iptr += size;
            addr += size;

            if ( (size == 0) || inval )
            {
               break;
            }
         }
      }
   }

   if (icnt == 0)
   {
      if (gv.showss == 1)
      {
         fprintf(gd.arc,
                 " ss: %016" _P64 "X  :: 0 icnt\n",
                 prevTo);
      }
   }
   fflush(gd.arc);

   return(icnt);
}

/**********************************/
char gettype(int oc, int diff, int offset, int rc)
{
   char gtype = '@';

   if (diff != 0) diff = 1;

   if (oc == CALL)
   {
      if (diff == 1)
      {                                 // direct recur NoSym NoMod
         gtype  = '>';
      }
      // call & same => @
   }

   else if ( (oc == RETURN) || (oc == JUMP) )
   {
      if (offset == 0)
      {
         if (diff == 1)
         {
            gtype  = '>';               // runtime_resolve & others
         }
      }
   }

   return(gtype);
}

/**********************************/
void show(char * s)
{
   if (gd.show)
   {
      fprintf(gd.show, " --- %s\n", s);
      fflush(gd.show);
   }
}

/**********************************/
void handle_13(trace_t * tr)
{                                       // jstk hook
   int  pid, tid;
   // int  ty;
   int  stk, min, cpu;
   int del;
   char * fpnm, * pnm;

   static int fpid[32] = {0};           // by cpu. not reqd for
   static int ftid[32] = {0};
   static int fstk[32] = {0};

   // good-FROM => save by CPU
   // good-TO   => post.where

   // Both 13/1 p t s p t s // now both are Good
   // From 13/2 p t s
   // To   13/3 p t s

   min = tr->minor;
   cpu = tr->cpu_no;

   if ( (min == 1) || (min == 2) )
   {
      fpid[cpu] = tr->ival[0];
      ftid[cpu] = tr->ival[1];
      fstk[cpu] = tr->ival[2];
   }

   if ( (min == 1) || (min == 3) )
   {                                    // TO
      if (min == 1) del = 3;
      else         del = 0;

      pid  = tr->ival[0 + del];
      tid  = tr->ival[1 + del];
      stk  = tr->ival[2 + del];

      // diff thrd & TO => out
      // need pnm on F or T ??
      if ( (pid != fpid[cpu]) || (tid != ftid[cpu]) )
      {
         fpnm = A2nGetProcessName(fpid[cpu]);
         pnm  = A2nGetProcessName(pid);

         if ( 0 == gd.wi )
         {
            gd.wi = xopen("post.where", "w");
         }

         fprintf( gd.wi, " %4x %4x %x %s %4x %4x %x %s\n",
                  fpid[cpu], ftid[cpu], fstk[cpu], fpnm,
                  pid, tid, stk, pnm);

         fpid[cpu] = pid;
         ftid[cpu] = tid;
         fstk[cpu] = stk;
      }
   }
}

/**********************************/
void handle_19_41(trace_t * tr)
{                                       // jita2n method load
   int  trcid, pid;
   int  tlo, thi, addr;
   int  rc;

   JREC      * pJREC    = NULL;
   JREC      * prevJREC = NULL;
   process_t * ppt;
   unsigned int jlo, jhi;

   int cnt;
   static int no_jit_record_cnt = 0;

   // debug. hooks written but pretend not
   // -oldjit option
   if ( (gv.oldjit == 1) || (gv.nv == 1) ) return;

   trcid = tr->ival[0];
   pid   = tr->ival[1];                 // pid
   tlo   = tr->ival[2];                 // trace time
   thi   = tr->ival[3];
   addr  = tr->ival[4];                 // jit starting addr (low 32bits)

   ppt = insert_process(pid);

   if (ppt->read == 0)
   {
      WriteFlush(gd.msg, " Reading jita2n %x from 19/41 hook\n", pid);
      read_jita2n(ppt);
   }

   pJREC = (JREC *)ppt->jit_ll;

   //WriteFlush(gd.msg, "\n 19/41 p %p\n", p);
   //WriteFlush(gd.msg, " trcid %4d pid %4x t %8x %8x addr %8x\n",
   //   trcid, pid, thi, tlo, addr);

   // Need to locate record for HexCode & 64 bit address
   // Match on trcid & time. Remove Matched rec and rewire LL
   // Dont issue skipped records, but keep

   // breaks if hook not found.
   cnt = 0;

   while ( pJREC )
   {
      jhi = pJREC->thi;
      jlo = pJREC->tlo;

      if ( cnt == 2 )
      {
         WriteFlush( gd.msg, "Possible problem: Symbol not found in first two JITA2N records\n" );
         WriteFlush( gd.msg, "19/41 = seq:%d time:%x:%x addr:%x\n", trcid, thi, tlo, addr );
         WriteFlush( gd.msg, "JREC  = seq:%d time:%x:%x addr:%x\n", pJREC->seqNo, jhi, jlo, pJREC->addr );
      }
      cnt++;

      if ( ( pJREC->seqNo == trcid) && ((int)jhi == thi) && ((int)jlo == tlo) )
      {
         // post.msg grows very rapidly if there are a lot of jitted methods.
         // print these messages only if -verbmsg option given.

         OptVMsg(" MATCHed\n A2nAddJitted %08x %08x %s cflag %d\n",
                 jhi, jlo, pJREC->mnm, pJREC->cflag);

         if ( pJREC->cflag & jrec_cflag_MMI )   // MMI
         {
            rc = A2nAddMMIMethod(pid, pJREC->mnm, pJREC->addr, (int)pJREC->size, pJREC->code);
         }
         else                           // JIT
         {
            rc = A2nAddJittedMethod(pid, pJREC->mnm, pJREC->addr, (int)pJREC->size, pJREC->code);
         }

         if ( rc  )
         {
            WriteFlush(gd.msg, " ERROR: A2nAddJittedMethod() rc %d method %s\n", rc, pJREC->mnm);
            WriteFlush(gd.msg, "  tm %08x %08x jaddr %016"_P64"x sz %x\n",
                       thi, tlo, pJREC->addr, pJREC->size);
         }

         if ( prevJREC )
         {
            if ( pJREC->next )
            {
               prevJREC->next = pJREC->next;   // Remove JREC from middle of list
            }
            else
            {
               prevJREC->next = NULL;   // Remove last JREC from list
            }
         }
         else
         {
            if ( pJREC->next )
            {
               ppt->jit_ll = pJREC->next;   // Remove JREC from head of list
            }
            else
            {
               ppt->jit_ll = NULL;      // Remove only JREC from list
            }
         }

         if ( pJREC->code )
         {
            free( pJREC->code );
         }
         free( pJREC->mnm );
         free( pJREC );                 // Free the JREC we just removed

         return;                        // Return when match found
      }
      prevJREC = pJREC;
      pJREC    = pJREC->next;
   }
   // No match was found

   no_jit_record_cnt++;                 // Don't print the "No JIT record" message forever unless -verbmsg

   if ( gv_verbose_logmsg & gv.verbose )
   {
      WriteFlush(gd.msg, " 19/41 Hook. No JIT record? ");
      WriteFlush(gd.msg, " trcid %4d pid %4d t %8x %8x addr %8x\n",
                 trcid, pid, thi, tlo, addr);
   }
   else if ( no_jit_record_cnt <= 10 )
   {
      WriteFlush(gd.msg, " 19/41 Hook. No JIT record?  pid %d\n", pid);
   }

   if ( no_jit_record_cnt == 1 )
   {
      ErrVMsgLog(" *****\n");
      ErrVMsgLog(" ***** POST is having trouble resolving jitted method(s).\n");
      ErrVMsgLog(" ***** - It is looking for '%s-jita2n_%u'\n", jitpre, pid);
      ErrVMsgLog(" ***** - Does that look right?\n");
      ErrVMsgLog(" ***** - Are there any log-jita2n files?\n");
      ErrVMsgLog(" ***** - Is the -jdir option set to point to them?\n");
#ifdef _WINDOWS
      ErrVMsgLog(" ***** - Are environment variables IBMPERF_JPROF_LOG_FILES_PATH and\n");
      ErrVMsgLog(" *****   IBMPERF_JPROF_LOG_FILES_NAME_PREFIX (in toolsenv.cmd) set correctly?\n");
#endif
      ErrVMsgLog(" *****\n");
   }
}

/**********************************/
void jitrec(int pid, trace_t * tr, int all)
{
   // all = 1 => read all jit records, forget time
   // Before all getSymbol Calls

   JREC       * pJREC;
   JTNM       * pJTNM;
   int          rc;
   int          db = 0;                 // jita2n extra data
   unsigned int jlo, jhi, tlo, thi;
   process_t  * ppt;
   thread_t   * tp;
   int          cut = 0;                // catchup mode

   // by pid:  jit & jtnm catchup
   ppt = insert_process(pid);

   // jita2n & jtnm files read ?
   if (ppt->read == 0)
   {
      read_jita2n(ppt);
   }

   // catchup mode
   if (gv.trcid > ppt->trcid)
   {
      cut = 2;                          // trcid
      ppt->trcid = gv.trcid;
   }
   else if (gv.trcid == 0)
   {
      cut = 1;                          // time
   }
   else
   {
      return;                           // no catchup needed
   }

   thi = tr->thi;                       // trace time
   tlo = tr->tlo;

   // JITA2N catchup
   while ( ppt->jit_ll )                // List of delayed JITA2N records
   {
      pJREC = ppt->jit_ll;

      jhi   = pJREC->thi;               // jit time
      jlo   = pJREC->tlo;

      // jittime < tracetime
      if ( ( (cut == 1) && ( (jhi < thi) || ((jhi == thi) && (jlo < tlo)) ) ) ||
           ( (cut == 2) && (pJREC->seqNo < gv.trcid) ) )
      {
         if (gd.show)
         {
            fprintf(gd.show, " A2nAddJittedMethod() %08x %08x %s cflag %d\n",
                    jhi, jlo, pJREC->mnm, pJREC->cflag);
         }

         if ( pJREC->cflag & jrec_cflag_MMI )   // MMI
         {
            rc = A2nAddMMIMethod(pid, pJREC->mnm, pJREC->addr, (int)pJREC->size, pJREC->code);

            if ( rc )
            {
               WriteFlush(gd.msg, "A2nAddMMIMethod(): rc %d tm %08x %08x sz %x jaddr %16" _P64 "x\n",
                          rc, thi, tlo, pJREC->size, pJREC->addr);
            }
         }
         else                           // JIT
         {
            rc = A2nAddJittedMethod(pid, pJREC->mnm, pJREC->addr, (int)pJREC->size, pJREC->code);

            if ( rc )
            {
               WriteFlush(gd.msg, "A2nAddJittedMethod(): rc %d tm %08x %08x sz %x jaddr %16" _P64 "x\n",
                          rc, thi, tlo, pJREC->size, pJREC->addr);
            }
         }
         ppt->jitSaved++;

         if (db == 1)
         {
            fprintf(gd.msg, " pid %x jit2 %d A2nAddJittedMethod()\n",
                    pid, ppt->jitSaved);
            fprintf(gd.msg,
                    " hktm %08x %08x jtm %08x %08x jaddr %016" _P64 "X sz %x\n",
                    thi, tlo, jhi, jlo, pJREC->addr, pJREC->size);
            fflush(gd.msg);
         }

         ppt->jit_ll = pJREC->next;     // Eliminate JREC from head of list

         if ( pJREC->code )
         {
            free( pJREC->code );
         }
         free( pJREC->mnm );
         free( pJREC );
      }
      else
      {                                 // catchup completed
         break;
      }
   }
   // End JITA2N CatchUp

   // Beginning JTNM Catchup

   while ( ppt->tnm_ll )
   {
      pJTNM = ppt->tnm_ll;

      jlo = pJTNM->tlo;
      jhi = pJTNM->thi;

      if (db)
      {
         OptVMsg(" JTNM %08x %08x tid %4x pid %4x %s\n",
                 jhi, jlo, pJTNM->tid, pid, pJTNM->tnm);
      }

      if ( ( (cut == 1) && ( (jhi < thi) || ((jhi == thi) && (jlo < tlo)) ) ) ||
           ( (cut == 2) && (pJTNM->tnmid < gv.trcid) ) )
      {
         tp  = insert_thread( pJTNM->tid, pid);

         if ( *pJTNM->tnm )
         {
            tp->thread_name = pJTNM->tnm;   // Use name string allocated for JTNM record
         }
         else
         {
            tp->thread_name = "";
         }
         ppt->tnm_ll = pJTNM->next;     // Eliminate JTNM from head of list

         free(pJTNM);
      }
      else
      {                                 // catchup completed
         break;
      }
   }
   // End JTNM Catchup
}

/**********************************/
int oneoc( uint64 addr, SYMDATA * sdp ) //STJ64
{
   char   buf[2048];
   char * iptr;
   char * opc;
   int    size;
   int    oc = OTHER;

   if ( sdp && sdp->code )
   {
      iptr = sdp->code + (int)(addr - sdp->sym_addr);

#if 0 // defined(CONFIG_PPC64) || defined(_AIX)

      oc  = decode_inst_type( iptr );   // 1-5 call return jmp iret other

      if ( gv.verbose )
      {
         opc  = (char *)i386_instruction(iptr, 
                                         (uint64)addr, 
                                         &size, 
                                         buf, 
                                         sdp->flags & ( SDFLAG_MODULE_32BIT | SDFLAG_MODULE_64BIT ));   //STJ64

         fprintf( gd.msg, "ONEOC: %016" _P64 "X: %08X = %d, %s\n",
                  addr,
                  *( (unsigned long *)iptr ),
                  oc,
                  opc );
         fflush( gd.msg );
      }
#else
      opc = (char *)i386_instruction(iptr, 
                                     (uint64)addr, 
                                     &size, 
                                     buf, 
                                     sdp->flags & ( SDFLAG_MODULE_32BIT | SDFLAG_MODULE_64BIT ));   //STJ64

      oc  = inst_type(opc);             // 1-5 call return jmp iret other
#endif
   }
   return(oc);
}

/**********************************/
void hstats(int bytes, uint64 fr, uint64 to)
{                                       //STJ64
   bhisto(bytes);                       // bytes   histo
   thisto(fr, to);                      // from/to histo
   cfbytes += bytes;
   cfbbs++;
}

/**********************************/
//STJ64 - Begin
uint64 makeptr( uint64 lo, uint64 hi )
{
   uint64 p = ( hi << 32 ) + ( lo & 0x0FFFFFFFF );   //ull suffix needed???
   return(p);
}
//STJ64 - End

/**********************************/
/// Process one CallFlow hook
///
/// \verbatim
///
/// PERCS (MM only on 64 cases(20/21)
/// if first c always instructions
/// 20/21                    ffttc    64 // MM 1
/// 20/21                    ffttcc   64 // MM 2
/// 20/21                    ffttccc  64 // MM 3
///
/// 20/21  ft[c] 32 ft[c] 64 fftt[c] 64  // cf
/// 22/23  fft[c] 64 & fh = th           // cf
/// 24/25  fb 32 ffb 64 (b=BBCC)         // cf
/// 28/29  vcf vcff (ia64)               // vec
///
/// 2c     dispatch xxxccccc
/// \endverbatim
void call_flow(trace_t * tr)
{
   int  offset = -1;
   int  oc  = 0;                        // Opcode type of last instruction in block
   int  oc1 = 0;                        // Opcode type of first instruction in block
//                                      // ia64/lx to fix lack of calls in places??
   int  maj,   pid,  tid,  cpu, ring, rc;
   int  loc;
   int  dcnt,  tcnt,  ucnt, rcnt, vec;
   int  i64,   cx, tx;
   int  bytes;
   int  ints;
   char vbuf[16];

   uint64 fr     = 0;                   // FROM address of callflow hook
   uint64 to     = 0;                   //  TO  address of callflow hook
   uint64 prevTo = 0;                   //  TO  address of PREVIOUS callflow hook

   thread_t  * tp;

   SYMDATA sd    = {0};

   // debug vars
   int mx = 0;

   /*** code ***/

   pid  = tid  = ring = bytes = loc = 0;
   tcnt = dcnt = ucnt = rcnt  = vec  = 0;
   rc   = tx   = cx   = 0;

   cpu  = tr->cpu_no;
   ints = tr->ints;

   // also return if -int3 & no f9 yet
   // actually dont get here on above cases
   if (cpu2stt[cpu] == 0) return;

   i64 = (int)(sizeof(char *) == 8);

   pid  = cpu2pt[cpu][0];
   tid  = cpu2pt[cpu][1];
   tp   = cpu2tp[cpu];                  // NO INSERT_THREAD !!!!

   maj  = tr->major;
   if (maj == 0x20) mx = 1;             // add MM itrace

   ring = maj & 0x1;
   maj  = maj & 0xfe;                   // Separate ring from maj

   if ( (maj == 0x20) || (maj == 0x22) )
   {
      if (maj == 0x20)                  // 0x20/0x21
      {
         if (ints <= 2)
         {
            fr = makeptr( tr->minor,   0 );
            to = makeptr( tr->ival[0], 0 );

            if (ints == 2)
            {
               tcnt = tr->ival[1];
            }
         }
         else
         {                              // 64
            fr = makeptr( tr->minor,   tr->ival[0] );
            to = makeptr( tr->ival[1], tr->ival[2] );

            if (ints == 4)
            {
               tcnt = tr->ival[3];
            }
            else if (ints > 4)
            {
               tcnt = tr->ival[3 + gv.itsel];
            }
         }
      }
      if (maj == 0x22)                  // 0x22/0x23
      {
         fr = makeptr(tr->minor,   tr->ival[0]);
         to = makeptr(tr->ival[1], tr->ival[0]);

         tcnt = tr->ival[2];
      }

      if ((fr == 0) || (to == 0))
      {
         OptVMsg(" fr or to = 0. skip\n");
         return;
      }

      // callflow hook path
      jitrec(pid, tr, 0);               // jita2n catchup
      rc = getSymbol(pid, fr, &sd);     // stack sd

      if (rc <= 2)
      {                                 // sym nsf ns BADDOG-SYMBOL
         int bad = 0;

         prevTo = tp->to;               // Starting address is target from last hook

         if ( (fr >= prevTo) && (prevTo >= sd.sym_addr)
              && (fr < prevTo + 2048) )
         {

            bytes = (int)(fr - prevTo);
            hstats(bytes, fr, to);

            // below for
            // 1) !tcnt  & !ss
            // 2) !tcnt  &  ss
            // 3)  tcnt  & !ss
            // 4)  tcnt  &  ss

            // NB: Below are several special cases
            // => needs revisiting for each port & nuance
            // instCnt in/notin trace
            // platform anomalies (LX below)
            // disasm availability
            // MM PERCS (could provide CALL/RETURN directly)
            // left over debug ??
            // other
            // NB: END

            // tcnt == 0 => calc tcnt from disasm
            if ( (gv.showss >= 1) || (tcnt == 0) )
            {
               if (sd.code)
               {
                  // decode 1st inst : IA64/LX/Java131- issue
                  // convert "alloc" to a previous CALL
                  oc1 = oneoc(prevTo, &sd);
               }
               // disasm the whole block of code
               dcnt = count_inst(prevTo, bytes, &sd);

               // trace count in trace AND trace count != disasm count
               if ( (tcnt > 0) && (tcnt != dcnt) )
               {
                  if ( gv.showss )
                  {
                     if ( (cpu2hk[cpu] >= 0x20) && (cpu2hk[cpu] <= 0x25) )
                     {
                        necnt1++;       // prev a call_flow
                        fprintf(gd.arc,
                                " # UNEX trace count != disasm count # tcnt %4d dcnt %4d\n",
                                tcnt, dcnt);
                     }
                     else
                     {
                        necnt2++;
                        fprintf(gd.arc,
                                " # EX   trace count != disasm count # tcnt %4d dcnt %4d\n",
                                tcnt, dcnt);
                     }
                  }

                  if (dcnt <= 1)
                  {                     // 0 or 1
                     // show rc fr prevTo sym_addr
                     if (gv.db == 4)
                     {
                        fprintf(gd.arc,
                                " ## code %p rc %d symaddr %016" _P64 "X fr %016" _P64 "X prevTo %016" _P64 "X\n",
                                sd.code,
                                rc,
                                sd.sym_addr,
                                fr,
                                prevTo);
                     }
                  }
                  tcnt = dcnt;          // pick trace cnt instead of disasm cont
               }
               else
               {                        // no cnt in trace
                  tcnt = dcnt;
               }
            }
            else
            {                           // cnt in trace & no ss => disasm fr only
                                        // to extrace CALL RETURN etc. to improve tree quality
               if (sd.code)
               {
                  // decode 1st inst : IA64/LX/Java131- issue
                  // convert "alloc" to a previous CALL
                  oc1 = oneoc(prevTo, &sd);
               }
            }

            gv.cfcnt++;
            cfinsts += tcnt;
         }
         else
         {                              // fr, prevTo break
            tcnt = 1;
            bad = 1;
            offset = (int)(fr - sd.sym_addr);

            if (gv.db == 4)
            {
               fprintf(gd.arc, " ### brk_%d: fr %016" _P64 "X prevTo %016" _P64 "X ofrom %016" _P64 "X\n",   //STJ64
                       cpu, fr, prevTo, tp->from);
            }
         }

         if (bad == 0)
         {
            if ( (rc == 0) && (prevTo != 0) )
            {                           // Sym
               // the really really good case
               offset = (int)(prevTo - sd.sym_addr);
            }
            else
            {
               // the not so good cases
               // offset=1 is flag for less than perfect
               offset = 1;              // noS & noSF
            }
         }
      }

      else
      {                                 // NoMod
         if (tcnt == 0)
         {
            tcnt = 1;                   // non-zero
         }
         offset = 1;
      }

      // add the LX alloc fixup here ????

      oc = oneoc(fr, &sd);              // decode last opcode

      arcbuf(tp, &sd, tcnt, offset, ring, oc, cpu, oc1);

      tp->from  = fr;                   // from
      tp->to    = to;                   // to
      tp->ring  = ring;
      tp->sd    = sd;                   // symbol by thread. read every time
      tp->sd_rc = rc;
   }

   if (maj == 0x28)
   {                                    // ia64 stuff
      int mnr;

      mnr  = tr->minor;
      vec  = 0xff & mnr;
      tcnt = tr->ival[0];

      // 28/29  ( enter : vcf vcff ) ( exit  : vc )
      if ((0x80000000 & mnr) == 0)
      {                                 // entry
         fr = makeptr(tr->ival[1], tr->ival[2]);   //STJ64
      }

      if (gv.arc)
      {
         sprintf(vbuf, "Int_0x%02X", vec);

         arcout(tp, tcnt, '?', 1, vbuf, "VECTOR",
                ring, 0, "-vec-", cpu);
      }
   }
}

/**********************************/
void shownrm2(int64 rec, trace_t * tr, FILE * fd)
{
   int        x;
   char       temp_str[MAX_STR_LEN];
   static int first = 0;
   //int mx;

   // enhance with pnm
   // 19.1 for pid => pnm
   // 12 (dispatch) => add pnm (from pid)

   if (first == 0)
   {
      first = 1;
      fprintf( fd, "        OFFSET CPU MAJOR MINOR TimeHigh:TimeLow    [INT]*  [STR]*\n" );
   }

   // integrated show (MM ??)
   //for(mx = 0; mx < tr->mets; mx++) {
   //   fprintf(fd, " %8x", tr->met_lo[mx]);
   //}

   if (gv.showx == 2)
   {
      // show raw binary data
      int i, len;

      len = tr->hook_len;
      if (len > 2048) len = 2048;
      for (i = 0; i < len; i += 4)
      {
         fprintf(fd, " %02x", tr->raw[i]);
         fprintf(fd, "%02x", tr->raw[i + 1]);
         fprintf(fd, "%02x", tr->raw[i + 2]);
         fprintf(fd, "%02x", tr->raw[i + 3]);
      }
      fprintf(fd, "\n");
   }

   fprintf(fd, " **** %016"_P64"x %3d %5x %5x %08x:%08x",
           tr->file_offset,
           tr->cpu_no, tr->major, tr->minor, tr->thi, tr->tlo);

   for (x = 0; x < tr->ints; x++)
   {
      fprintf(fd, " %8x", tr->ival[x]);
   }

   for (x = 0; x < tr->strings; x++)
   {
      memset(temp_str, 0, MAX_STR_LEN);
      strncpy(temp_str, (char *)&tr->sval[x][0], tr->slen[x]);
      fprintf(fd, " [%d]%s", tr->slen[x], temp_str);
   }

   // add pnm to dispatch(0x12)
   // Win Only Debug Hack for WhereIdle.
#ifdef _WINDOWS
   if (!gv.showx)
   {
      if (tr->major == 0x12)
      {
         char * pnm;

         pnm = A2nGetProcessName(tr->minor);
         fprintf(fd, " <%s>", pnm);
      }
   }
#endif

   fprintf(fd, "\n");
   fflush(fd);
}


/**********************************/
void handle_f9(trace_t * tr)
{
   // ints(x x cpu aacnt)
   // reduce between 1st & 2nd int3
   // also AA cnt per cpu is in int[3]

   if (aout)
   {
      if (gv.arc) fprintf(gd.arc, " ## int3\n");
   }
   int3c++;

   if ((aout == 0) && (int3c == 1))
   {
      // int3 On
      int3x = 1;
   }
   aout = int3f || (int3c == 1);

   gv.aatot += tr->ival[3];
}

/**********************************/
// common between NT LX & AIX
void proc_tprof(int pid, int tid, uint64 addr, int cpu)
{
   int rcq;
   SYMDATA sd = {0};
   thread_t * tp;

   char    * mnm;
   char    * mod;
   char    * pnm;
   static char modx[256];               // mod 32/64 decoration

   char     pstr[MAX_STR_LEN];
   char     tstr[MAX_STR_LEN];
   char     symstr[MAX_STR_LEN];
   char     cpustr[10];

   size_t   offset;
   pTNode   p;

   static int bdcnt = 0;                // baddog quality count

   char obuf[16];

   static int acnt       = 0;           // show sym_stt_addr for first 100
   int    acntlim        = 0;
   uint64 xaddr;                        // mod or sym addr
   uint64 zero64         = 0;           // 64 bit 0
   static int64 nosymcnt = 0;

   ///////////////////////
   /// proc_tprof CODE ///
   ///////////////////////

   jitrec(pid, tr, 0);                  // jita2n catchup

   // A2N return codes:
   //  (0) A2N_SUCCESS:           Everything worked. Have symbols.
   //  (1) A2N_NO_SYMBOL_FOUND:   Have symbols but no symbol found for addr
   //  (2) A2N_NO_SYMBOLS:        Module has no symbols
   //  (3) A2N_NO_MODULE:         No loaded module found
   rcq = getSymbol(pid, addr, &sd);

   mod  = (char *)sd.mod_name;
   mnm  = (char *)sd.sym_name;

   if (rcq == A2N_SUCCESS)
   {
      // Have symbols
      xaddr = sd.sym_addr;
   }
   else if (rcq == A2N_NO_SYMBOL_FOUND || rcq == A2N_NO_SYMBOLS)
   {
      // Can't find symbol for addr or there are no symbols.
      // A2N returns either "NoSymbols", "NoSymbolFound" or a section name.
      if (mnm[0] == '<' && mnm[1] == '<')
      {
         // Section name. Consider that a valid symbol
         xaddr = sd.sym_addr;
      }
      else
      {
         // "NoSymbols" or "NoSymbolFound". Not much we can do
         xaddr = sd.mod_addr;
      }
      nosymcnt++;

      if (gv_verbose_logmsg & gv.verbose)
      {
         WriteFlush(gd.msg, " *NOSYM* cnt=%"_L64"d rcq=%d symaddr 0x%8"_P64"X xaddr 0x%8"_P64"X mod %s sym %s\n",
                    nosymcnt, rcq, addr, xaddr, mod, mnm);
      }
   }
   else
   {
      // No module or something else
      xaddr = sd.mod_addr;
   }
   offset = (size_t)(addr - xaddr);

   // for Kean (raw data for debug)
   if (gv.db == 6)
   {
      static int once = 0;

      if (once == 0)
      {
         once = 1;
         OptVMsg(" ADDR: %3s %6s %6s %16s %16s %8s %s %s\n",
                 "rcq", "pid", "tid", "addr", "xaddr", "offset", "mod", "mnm");
      }
      OptVMsg(" ADDR %08x %08x %3d %6x %6x %016" _P64 "X %016" _P64 "X %8x %s %s\n",
              tr->thi, tr->tlo, rcq, pid, tid, addr, xaddr, offset, mod, mnm);
   }

   // error report for low quality A2N
   if (rcq >= 1)
   {
      bdcnt++;
      if (bdcnt <= 5)
      {
         ErrVMsgLog(" Low Quality Symbol %d: a2n_rc=%d  addr=0x%"_PZP"  module=%s\n",
                    bdcnt, rcq, Uint64ToPtr(addr), mod);
      }
      if (bdcnt == 5)
      {
         ErrVMsgLog("\n *WARNING* At least 5 Low Quality Symbols (a2n_rc >= 1).\n");
         WriteFlush(gd.msg, " - a2n_rc = 1: Symbol Not Found\n");
         WriteFlush(gd.msg, " - a2n_rc = 2: No Symbols for Module\n");
         WriteFlush(gd.msg, " - a2n_rc = 3: No Module\n");
      }
   }

   // === BEGIN PNM ===
   if (aix == 0)
   {                                    // Normal A2N
      pnm = (char *)sd.owning_pid_name;
   }
   else
   {                                    // AIX : pnm on hook
      pnm = gv_tppnm;
   }

   if (pnm != NULL)
   {
      if (gv.pnmonly == 0)
      {
         sprintf(pstr, "%s_%04x", pnm, pid);
      }
      else
      {
         sprintf(pstr, "%s", pnm);
      }
   }
   else
   {
      sprintf(pstr, "pid_%04x", pid);
   }
   // === END PNM ===

   if (gv.cpu == 1)
   {
      sprintf(cpustr, "_%d", cpu);
      strcat(pstr, cpustr);
   }

   // === BEGIN TNM ===
   // thread name creation
   tp = insert_thread(tid, pid);

   // 3 formats
   // tid_%04x      - noName
   // tid_tnm_%04x  - Name
   // tid_tnm       _ Name & tnmonly option

   if ( *tp->thread_name )              // Thread name is not ""
   {
      if (gv.tnmonly == 0)
      {
         sprintf(tstr, "tid_%s_%04x", tp->thread_name, tid);
      }
      else
      {
         sprintf(tstr, "tid_%s", tp->thread_name);
      }
   }
   else
   {
      sprintf(tstr, "tid_%04x", tid);
   }
   // === END TNM ===

   // tag 64 bit modules with _64
   if (gv.tmod)
   {
      if ( sd.flags & SDFLAG_MODULE_64BIT )
      {
         sprintf(modx, "%s_64", mod);
         mod = modx;
      }
   }

   // === BEGIN Sym Addr ===
   // append addr to symbols
   if (gv.undup)
   {
      if (rcq == 0)
      {
         int addr = (int)(xaddr & 0xffffffff);

         sprintf(symstr, "%s_%x", mnm, addr);
         mnm = &symstr[0];
      }
   }
   // === END Sym Addr ===


   // push tick into tree (4 or 5 levels)
   // - PID
   // -  TID
   // -   Module
   // -    Symbol
   // -     Offset
   p = pusha(root, pstr,        zero64);   // pid node
   p = pusha(p,    tstr,        zero64);   // tid node
   p = pusha(p,    (char *)mod, zero64);   // mod node
   p = pusha(p,    (char *)mnm, xaddr); // sym node

   // debug peek at sttaddr
   if (acnt < acntlim)
   {
      if (p->rc == 0)
      {
         acnt++;
         OptVMsg(" %3d %16" _P64 "x\n p->nm <%s>\n",
                 acnt, p->sttaddr, p->nm);
      }
   }

   // decorating the SYMBOL Node Only
   p->rc      = rcq;                    // symbols Quality
   p->clen    = sd.sym_length;
   p->cptr    = sd.code;                // code ptr ( or NULL )
   p->module  = sd.module;              // late a2n resolution
   p->symbol  = sd.symbol;              // late a2n resolution
   p->sttaddr = sd.sym_addr;            // ??
   p->flags   = sd.flags;               // Save the symbol flags //STJ64

   // offset for microprofiling
   if (gv.off)
   {
      sprintf(obuf, "0x%x", offset);

      if (rcq == 0)
      {                                 // good symbol found
         p->cptr = sd.code;             // code ptr ( or NULL )
      }
      else p->cptr = NULL;

      p->snm = "-";                     // default bef linenos used

      // if LINENOS & (WIN32 or LINUX)
#ifndef NOLINENOS
   #if defined(WIN32) || defined(WINDOWS) || defined(_LINUX)
      p->snm = sd.src_file;
   #endif
#endif

      // Pushing the OFFSET node
      p = pusha(p, obuf, zero64);

      p->rc = -2;                       // default bef linenos used

#ifndef NOLINENOS
   #if defined(WIN32) || defined(WINDOWS) || defined(_LINUX)
      p->rc  = sd.src_lineno;
   #else
      p->rc = -3;                       // if linenos but not on this platform
   #endif
#endif
   }
   p->base++;                           // tick on lowest level (symbol or offset)

   if ( (gv.db == 4) && gd.arc)
   {
      fprintf(gd.arc, " tprof: addr %016" _P64 "X pid %x < %s %s %s %s >\n",   //STJ64
              addr, pid, mnm, mod, tstr, pstr);
      fflush(gd.arc);
   }
}

/**********************************/
void handle_tprof_hook(trace_t * tr)
{
   // NT & LX : nrm2 format

   uint64    addr = 0;                  //STJ64
   //size_t    addrhi;
   uint64    addrhi;
   int       pid = 0;
   int       tid = 0;

   int mnr, cpu, ring, flg;
   int nx = 0;
   int maj;
   int sub;

   static int64 tpcnt = 0;              // filter tprof range
   static int   first = 0;              // first tprof hook time

   // 0x10 : tprof for LX & NT
   //   mnr 0x10 | 0x11 | 0x13    tid f[f]      // lx
   //       0x20 | 0x21 | 0x23    pid tid f[f]  // nt
   //   mm tprof => use 2nd byte of minor (i.e 0423 => 4-th metric
   //       0xbbxx => bb for selection of which metric to use
   //                 bb = 0 => use all tprof together
   //                 bb != 0 => use selected via tprofmm n

   // tprof windowing
   tpcnt++;
   if (gv.tstt > tpcnt)      return;
   else if (gv.tstp < tpcnt) return;

   cpu = tr->cpu_no;

   // get 1st & last tprof time
   if (first == 0)
   {
      first = 1;
      trstt = getTime(cpu);             // 1st tprof hook time
   }
   trstp = getTime(cpu);                // last tprof hook

   maj = tr->major;
   mnr = tr->minor;
   sub = (0xff00 & mnr) >> 8;           // which tprof metric
   mnr = 0xff & mnr;

   if (gv.tpsel == sub)
   {                                    // normally both 0
      // using this tprof hook

      if ( (mnr & 0x30) != 0)
      {
         flg  = mnr & 0x30;             // NT 0x20 vs LX 0x10
         ring = mnr & 0x1;              // 0 1 3

         if (flg == 0x20)
         {                              // nt
            pid = tr->ival[0];
            tid = tr->ival[1];
            nx = 2;
         }
         else if (flg == 0x10)
         {                              // lx
            tid = tr->ival[0];
            pid = A2nGetOwningPid(tid);
            if (pid == -1)
            {
               fprintf(stderr, " A2nGetOwningPid Err. tid %x\n", tid);
               //return;
               // continue with pid = -1
            }
            nx = 1;
         }

         addr = (uint64)(uint)tr->ival[nx];   //STJ64
         nx++;
         if (tr->ints > nx)
         {
            addrhi = (uint64)((uint)tr->ival[nx]);
            addr   = (addrhi << 32) | addr;   //STJ64
         }
      }
      else
      {
         fprintf(stderr, " Invalid TPROF minor : 0x%x\n", mnr);
         exit(-1);
      }

      cpu = tr->cpu_no;

      proc_tprof(pid, tid, addr, cpu);
   }
   else
   {
      // OptVMsg(" not reducing this guy sub = %d\n", sub);
   }
}

/**********************************/
void handle_19_34_hook(trace_t * tr)
{
   char pnmbuf[MAX_STR_LEN];
   int  pid;

   // ints(pid) [32/64] strs(pnm)
   // ints = 1 strs = 1
   pid = tr->ival[0];

   // AMD & LX 32(x86)/64(x64) problem.
   // running 32b emulation on a x64
   //if(ints == 2) {
   // 0x32 => 1 => pnm_pid(x86)
   // 0x64 => 0
   // decorate pnm for tprof
   //gv.p32 = tr->ival[1]; // 0x32 0x64
   //}

   memcpy(pnmbuf, &tr->sval[0], tr->slen[0]);   // mnm
   pnmbuf[tr->slen[0]] = 0;

   A2nSetProcessName(pid, pnmbuf);      // nop for WIN64
}

/**********************************/
void handle_newpid(trace_t * tr)
{                                       // 19/64
   int rc = 0;

   rc = A2nCreateProcess(tr->ival[0]);
   // err cases ??
}

/**********************************/
void handle_19_42(trace_t * tr)
{                                       // Thread Start / Stop
   int        pid;
   int        tid;
   thread_t * tp;

   if (gv.oldjit == 1) return;

   // All needed data in hook
   // => dont need to Match w saved records.

// trcid = tr->ival[0];
   pid   = tr->ival[1];                 // pid when hook written
// tlo   = tr->ival[2];                 // trace time low
// thi   = tr->ival[3];                 // trace time high
   tid   = tr->ival[4];                 // java thread ID
// ss    = tr->ival[5];                 // 0:Start, 1:Stop

   tp  = insert_thread(tid, pid);

   if ( tr->ival[5] )                   // 0:Start, 1:Stop
   {
      tp->thread_name = "";             // Stop
   }
   else
   {
      tp->thread_name = malloc( 1 + tr->slen[0] );   // Start

      if ( NULL == tp->thread_name )
      {
         fprintf(stderr, " ** Out of memory in handle_19_42(trace_t *) ***\n");
         exit(1);
      }
      memcpy( tp->thread_name, &tr->sval[0], tr->slen[0]);   // thread_name
      tp->thread_name[ tr->slen[0] ] = 0;   // NULL terminate
   }

   //OptVMsg(" ADDR %08x %08x 19/42 JTNM <%s>\n", thi, tlo, jtnm);
}

/**********************************/
// pid parpid ownpid
// action  pid ppid   opid
// ====== ==== ==== ======
//   fork    A    B      A
//   exec    A    B own(B)
void handle_pidmap(trace_t * tr)
{                                       // 19/63
   int  pid, ppid, clone;
   int  rc;
   char fpid[16];
   char tpid[16];
   static int hcnt = 0;

   pid   = tr->ival[0];                 // new  pid
   ppid  = tr->ival[1];                 // prev pid
   clone = tr->ival[2];                 // 0-fork, 0x100-clone

   sprintf(fpid, "%04x", ppid);
   sprintf(tpid, "%04x", pid);

   hcnt++;

   if (clone == 0)
   {                                    // fork
      rc = A2nForkProcess(ppid, pid);   // pid owns itself

      if (rc != 0)
      {
         fprintf(stderr, " a2nERR: A2nForkProcess(%x %x) rc %d\n",
                 ppid, pid, rc);
      }
   }
   else
   {                                    // clone  0x100
      rc = A2nCloneProcess(ppid, pid);

      if (rc != 0)
      {
         fprintf(stderr, " a2nERR: A2nCloneProcess(%x %x) rc %d\n",
                 ppid, pid, rc);
      }
   }

   return;
}

/*****************************/
void handle_startup_hook(trace_t * tr)
{
   int speed;

   switch (tr->minor)
   {
   case 0xb2:                           // trace Sequence No.
      if (gv.oldjit == 0)
      {
         gv.trcid = tr->ival[0];        // traceID for NUMA
      }
      OptVMsg(" 11/b2 trcid %d\n", gv.trcid);
      break;

   case 0xb3:                           // end of start up
      break;

   case 0xb5:
      speed = tr->ival[0];              // MHz
      gv.clock_rate = (double)(speed * 1000000.0);
      fprintf(gd.msg, " ClockRate %8.0f\n", gv.clock_rate);
      break;

   case 0xb7:
      gv.timebase = tr->ival[0];
      fprintf(gd.msg, " ClockRate %d 0x%x\n", gv.timebase, gv.timebase);
      break;

   case 0xb8:
      gv.tprof_mode  = tr->ival[0];
      gv.tprof_count = tr->ival[1]; 
      fprintf(gd.msg, " Tprof mode %d, count = %d\n", gv.tprof_mode, gv.tprof_count);
      break;

   case 0xb9:
      gv.tprof_event = tr->ival[1];
      fprintf(gd.msg, " Tprof event %d\n", gv.tprof_event);
      break;


   case 0xbc:                           // NRM2 Architectural Meeting REQUIRED
      // ind = int[0] - 1
      // metric[int[ind] = int[1]
      // mcnt = int[0]
      //gv.timebase = tr->ival[0];

      //fprintf(gd.msg, " \n");

      break;      
   }
}

/**********************************/
void major_histo(void)
{
   int i;

   fprintf(stdout, "\n    Major     Freq  Function\n    =====     ====  ========\n");

   for (i = 0; i < 64 * 1024; i++)
   {
      if (hmaj[i] >= 1)
      {
         fprintf(stdout, " %8x %8d  %s\n", i, hmaj[i], hmajlab[i]);
      }
   }
   fprintf(stdout, "\n");

   fprintf(gd.msg, "\n %8d  CallFlow Hooks from int3\n", gv.aatot );

   if (ssring[3] >= 1)
   {
      fprintf(gd.msg, "\n  CallFlowInstructions by Ring\n\n     Ring     Freq\n     ====     ====\n");

      for (i = 0; i < 4; i++)
      {
         if (ssring[i] >= 1)
         {
            fprintf(gd.msg, " %8d %8d\n", i, ssring[i]);
         }
      }
   }
}

#ifdef _AIX

/**********************************/
void exec_aix(HookData * p)
{
   int    rc;

   rc = A2nAddModule(p->pid, 0, 0, p->cmd, 0, 0, NULL);
}

/**********************************/
void fork_aix(HookData * p)
{
   int rc;

   rc = A2nForkProcess(p->opid, p->pid);
}

/**********************************/
void tprof_aix(HookData * ph)
{
   int    rc;

   int pid, tid, cpu;
   uint64 addr   = 0;
   static first  = 0;

   addr     = ph->iar;
   if (gv.db == 7)
   {
      addr     = ph->lr;
   }
   cpu      = ph->cpu;
   pid      = ph->pid;
   tid      = ph->tid;

   // pnm on the tprof hook,
   // uses dec value if not found yet => 17 => -17-
   gv_tppnm = ph->pnm;

   // zero based time in nanoseconds
   // coordinated with jita2n time field
   tr->thi = ph->thi;
   tr->tlo = ph->tlo;

   currtt = getTraceTime();

   if (first == 0)
   {
      first = 1;
      trstt = currtt;                   // 1st tprof hook
   }

   trstp = currtt;                      // last tprof hook

   if (gv.db == 6)
   {
      OptVMsg(" time %8x %8x\n", tr->thi, tr->tlo);
   }

   proc_tprof(pid, tid, addr, cpu);
}

/**********************************/
void readAixTrc(void)
{
   HookData   h;
   HookData * hp = &h;
   int hid = 1;

   hid = nextHook(hp);

   while (hid)
   {
      // start with tprof
      switch (hid)
      {
      case 0x234:                       // tprof
         if (gv.tpsel == 0)
         {
            tprof_aix(hp);
         }
         break;

      case 0x238:                       // tprof using eprof
         if (gv.tpsel == 1)
         {
            tprof_aix(hp);
         }
         break;

      case 0x134:                       // exec
         exec_aix(hp);
         break;

      case 0x139:                       // exec
         fork_aix(hp);
         break;

      default:
         break;
      }
      hid = nextHook(hp);
   }
}
#endif // _AIX

/**********************************/
void readnrm2(int64 nstt, int64 nstp)
{
   int64  rec2        = 0;
   int64  rec         = 0;

   OptVMsg(" A2N Version%x\n", A2nGetVersion());

   if (gv.oprof == 1)                   // i.e., Linux
   {
      // pass:1 - mte hooks only, no need if showx

      if (gv.showx == 0)
      {
         init_list_ptr_mte();           // init list of pointers to file sections
         
         tr = get_next_rec_mte();
         while (tr != NULL)
         {
            switch (tr->major)
            {
            case MTE_MAJOR:             // mte & jitted methods
               switch (tr->minor)
               {
               case MTE_MINOR:          // mte mod
                  handle_19_1_hook(tr);
                  break;

               case PID_NAME_MINOR:     // pnm from fork/exec
                  handle_19_34_hook(tr);   // pid pnm
                  break;

               case FORK_MINOR:
                  handle_pidmap(tr);    // pid ppid cflg
                  break;

               case NEWPID_MINOR:
                  handle_newpid(tr);
                  break;
               }
               break;

            case SYS_INFO_MAJOR:        // end of init
               handle_startup_hook(tr);
               break;

            default:
               break;
            }                           // switch

            tr = get_next_rec_mte();
         }                              // while

         // reset trace to beginning (for 2nd pass)
         reset_raw_file();
      }                                 // showx
      // If normal tracing mode, we need to reset values 
      // only for the final MTE section
      if (COLLECTION_MODE_NORMAL == gv.trace_type) 
      {
         init_last_mte();         
      }
      else
      {
         init_list_ptr_mte();
      }         
   }                                    // oprof

   // 1st pass OR oprof 2nd pass
   init_list_ptr_trace();
   tr = get_next_rec();

   while ( (rec <= nstp) && (tr != NULL))
   {
      int  miss = 0;

      hmaj[tr->major]++;
      currtt = getTraceTime();

      if (rec >= nstt)
      {
         if (gd.show)
         {
            shownrm2(rec, tr, gd.show);
         }
         if ( (gv.db == 4) && gd.arc )
         {
            shownrm2(rec, tr, gd.arc);
         }
#ifdef CONFIG_PPC64
         if (gd.qtrace) write_qtrace(tr, cpu2pt[tr->cpu_no][0], gd.qtrace);
         if (gd.ls) write_ls(tr, cpu2pt[tr->cpu_no][0], gd.ls);
#endif

         if (gv.showx == 0)
         {
            switch (tr->major)
            {
            case 0x13:
               handle_13(tr);           // pid-tid-jstk
               break;

               // MM 64 cf : fftt[c]+ (4+)
            case 0x20:                  // cf: ft[c] (1|2) fftt[c] (3|4)
            case 0x21:
            case 0x22:                  // cf: fft[c] : ia64 & fh = th
            case 0x23:
            case 0x24:                  // cf: fb ffb (b=BBCC)
            case 0x25:
            case 0x28:                  // Interrupt / Exception : vc  vcl
            case 0x29:
               if (gd.arc)
               {
                  call_flow(tr);
               }
               break;

               // added for MM ITrace
            case 0x1e:                  // MMdisp LX  : tid [c]+
            case 0x1f:                  // MMdisp Win : pid tid [c]+

            case 0x2c:                  // disp LX : tid c
               //      //LX : tid [c]+
               //      Win : pid tid c
            case 0x2d:                  //      //Win : pid tid c ts*
               if (gd.arc)
               {
                  cf_pidtid(tr);
               }
               break;

            case 0x2e:                  // itracestt cpid ctid sttpid stttid mode
               // equivalent to 0x2c & start now
               break;

            case 0x2f:                  // itracestp cpu cfhkcnt (per processor)
               break;

               // old, retain
            case 0x04:                  // interrupt entry/exit
               handle_interrupt(tr);
               break;

            case 0x10:
               // -arc option NOT used
               if (gv.mtrace == 0)
               {
                  handle_tprof_hook(tr);
               }
               break;

            case 0x11:                  // end of init
               handle_startup_hook(tr);
               break;

            case 0x12:                  // dispatch . tprof only
               handle_dispatch(tr);
               break;

            case 0x19:                  // mte & jitted methods
               if (gv.oprof == 0)
               {
                  switch (tr->minor)
                  {
                  case 0x01:            // mte mod
                     handle_19_1_hook(tr);
                     break;

                  case 0x34:            // pnm from fork/exec
                     handle_19_34_hook(tr);   // pid pnm
                     break;

                  case 0x63:
                     handle_pidmap(tr); // pid ppid cflg
                     break;

                  case 0x64:
                     handle_newpid(tr);
                     break;

                  }
               }
               switch (tr->minor)
               {
               case 0x41:
                  handle_19_41(tr);     // jitload
                  break;

               case 0x42:
                  handle_19_42(tr);     // jtnm
                  break;
               }
               break;

               // DEPRECATED
            case 0xf9:                  // int3 ( logical start stop )
               //handle_f9(tr);
               break;


            case 0x8:                   // low timer overflow
            case 0xa8:                  // low timer overflow
               // handled in bputil
               break;

            case 0x9:                   // MM hook [08]*[09][xx]
               // should be handled in bputil to update hi metric
               break;

            default:
               miss = 1;
               break;
            }

            if (miss & aout)
            {
               // EnEx
               if (gv.arc)
               {
                  //fprintf(gd.arc, " %6d ? hook_%x_%x\n",
                  //      0, tr->major, tr->minor);
                  // ptflush & arcout() ??
               }
            }
         }
      }

      cpu2hk[tr->cpu_no] = tr->major;

      tr = get_next_rec();

      if (rec2 == 10000)
      {
         fprintf(stdout, " Records  %10" _P64 "d\n", rec);
         rec2 = 0;
      }
      rec2++;
      rec++;
   }

   // fflush active ITrace threads
   if (gd.arc) allpt_flush();

   if (cfinsts >= 1)
   {
      fprintf(gd.msg, " Total CallFlow Insts %d\n", cfinsts);
   }
   fprintf(gd.msg, " Total Records %10" _P64 "d\n", rec);

   major_histo();
   cfstats();
   close_raw_file();
}

/**********************************/
void initGV(void)
{
   int i;

   struct _maj
   {
      int    ind;
      char * lab;
   };

   // labels for major histogram
   static struct _maj xmaj[64] =
   {
      { 0x03, "EXCEPTION"},
      { 0x04, "INTERRUPT"},
      { 0x05, "SYSCALL"},
      { 0x06, "Vec-WinIA64"},
      { 0x07, "Inter-Ent-WinIA64"},
      { 0x08, "TIMER_WRAP"},
      { 0x10, "TPROF"},
      { 0x11, "STARTUP"},
      { 0x12, "DISPATCH"},
      { 0x19, "MTE JMNM"},
      { 0x1e, "MMIT-Disp-LX"},
      { 0x1f, "MMIT-Disp-Win"},

      { 0x20, "CALL_FLOW"},
      { 0x21, "CALL_FLOW"},
      { 0x22, "CALL_FLOW"},
      { 0x23, "CALL_FLOW"},
      { 0x24, "CALL_FLOW"},
      { 0x25, "CALL_FLOW"},
      { 0x28, "Inter/Exc"},
      { 0x29, "Inter/Exc"},
      { 0x2c, "CF_PIDTID"},
      { 0x2d, "CF_PIDTID"},
      { 0x2e, "Itrace_stt"},
      { 0x2f, "Itrace_stp"},
      { 0x30, "MEMORY"},

      { 0xa8, "TIMER_WRAP"},
      { 0xaa, "CALL_FLOW"},
      { 0xab, "CALL_FLOW"},
      { 0xef, "SYSCALLS"},
      { 0xf0, "EXCEPTION"},
      { 0xf1, "DISPATCH"},
      { 0xf2, "FORK"},
      { 0xf3, "TASKLET"},
      { 0xf5, "BOTHALF"},
      { 0xf7, "INTERRUPT"},
      { 0xf9, "INT3"},
      { 0xfb, "SINGLESTEP"},
      { 0x00, ""},                      // sentinel (stops linear search)
   };

   // init 64k structs here (avoid exe space)
   for (i = 0; i < 64 * 1024; i++)
   {
      hmaj[i] = 0;
      hmajlab[i] = NULL;
   }

   // dont stop on disasm error
   noStopOnInvalid();

   // linux 64 bit ( initializer on GV didnt work ??? )
   memset(&gv, 0, sizeof(GV));

   gv.sympath  = NULL;
// gv.ptree    = 1;                     // default ITrace output is ptree
   gv.tottks   = 0;                     // redundant
   gv.tstt     = 0;                     // redundant
   gv.tstp = 10000000;

   for (i = 0; i < 64 * 1024; i++)
   {
      hmaj[i] = 0;
      hmajlab[i] = "";
   }

   for (i = 0; i < 64; i++)
   {
      int j;

      j = xmaj[i].ind;
      if (j == 0) break;
      hmajlab[j] = xmaj[i].lab;
   }

   thread_table_init();
   process_table_init();
}

/**********************************/
void Ver_Bld_Plat(FILE * fd)
{
   fprintf(fd, "\n");
   fprintf(fd, " POST - Version 7.2\n");
   fprintf(fd, "  Itrace & Tprof Post Processor\n\n");
   fprintf(fd, "   (Built    : %s)\n\n", BUILDDATE);
   fprintf(fd, "   (Source   : %s)\n\n", SOURCEDATE);
   fprintf(fd, "   (Platform : %s)\n\n", _PLATFORM_STR);
}

/**********************************/
void usage(void)
{
   Ver_Bld_Plat(stderr);

   fprintf(stderr, "   Common\n\n");
   fprintf(stderr, "     -verbose    : (display all optional messages)\n");
   fprintf(stderr, "     -verbmsg    : (display all optional messages to stderr)\n");
   fprintf(stderr, "     -mask=xxxx  : (debugging mask, implies -verbmsg)\n");
   fprintf(stderr, "     -r nrm2     : nrm2 file (def: swtrace.nrm2)\n");
   fprintf(stderr, "     -show       : ascii dump of nrm2 data (post.show)\n");
   fprintf(stderr, "     -showx      : -show, but no other processing\n");
   fprintf(stderr, "     -p  stp     : process records [1, stp]\n");
   fprintf(stderr, "     -pp stt stp : process records [stt, stp]\n");
   fprintf(stderr, "     -?          : Help\n");
   fprintf(stderr, "\n");
   fprintf(stderr, "   A2N\n\n");
   fprintf(stderr, "    See A2N documentation for further details\n");
#if defined(_LINUX)
   fprintf(stderr, "     -k kpath    : kernel file (for symbols)\n");
   fprintf(stderr, "                  (def: /usr/src/linux/vmlinux)\n");
   fprintf(stderr, "     -kmap kpath : kernel map (for symbols)\n");
   fprintf(stderr, "     -kas        : use /proc/kallsyms (for kernel symbols)\n");
   fprintf(stderr, "                   and /proc/modules (for kernel modules symbols)\n");
#endif
   fprintf(stderr, "     -nv         : Don't validate Symbols \n");
   fprintf(stderr, "     -sympath p  : Alternate symbols path list\n");
   fprintf(stderr, "     -lmrg fnm   : Use merge file\n");
   fprintf(stderr, "                   (create merge file with 'genmsi')\n");
   fprintf(stderr, "     -jdir [path/]pref : Location & prefix of jprof jita2n & jtnm files\n");
   fprintf(stderr, "                   (def: log)\n");
   fprintf(stderr, "              o jprof 'jita2n' option creates jita2n files\n");
   fprintf(stderr, "              o jprof 'threadinfo' option creates jtnm files\n");
   fprintf(stderr, "              o jprof 'fnm=xxx' option relocates jita2n & jtnm files \n");
   fprintf(stderr, "     -nsf        : Return \"NoSymbols\" or \"NoSymbolFound\", instead of\n");
   fprintf(stderr, "                   a section name, when there are no symbols (NoSymbols) or\n");
   fprintf(stderr, "                   when there are symbols but an address can't be resolved\n");
   fprintf(stderr, "                   to a symbol (NoSymbolFound).\n");
   fprintf(stderr, "     -anysym     : harvest any available symbols (e.g. exports)\n");
   fprintf(stderr, "     -a2ndbg     : a2n debug mode on \n");
   fprintf(stderr, "     -a2nrdup    : a2n rename duplicate symbols \n");
   fprintf(stderr, "     -a2nmmi NM  : collapse Interpreter & rename as NM\n");
   fprintf(stderr, "     -a2nrng Modnm Stt Stp Rnm\n");
   fprintf(stderr, "              o Collapse symbols between 'Stt' & 'Stp' in module Modnm.\n");
   fprintf(stderr, "              o Rename the range to 'Rnm'\n");
   fprintf(stderr, "     -a2nrfn FNM : Use the file FNM to collapse symbol ranges\n");
   fprintf(stderr, "              o Each file line has 4 parms as for -a2nrng option\n");
   fprintf(stderr, "\n");
   fprintf(stderr, "   tprof\n\n");
   fprintf(stderr, "     -aixtrc fnm : aix ascii trace file\n");
   //fprintf(stderr, "     -clip pc   : clip %% (def .1%%)\n");
   fprintf(stderr, "     -clip pc   : clip %% (def 0)\n");
   fprintf(stderr, "     -pnm       : report by process name (pnm)\n");
   fprintf(stderr, "                  (def: pid_pnm)\n");
   fprintf(stderr, "     -tnm       : report by thread name (tnm)\n");
   fprintf(stderr, "                  (def: tid_tnm)\n");
   fprintf(stderr, "     -cpu       : add _cpu to process name\n");
   fprintf(stderr, "     -rsel v    : select subset of def reports\n");
   fprintf(stderr, "                  (P PM PMS PT PTM PTMS M MS)\n");
   fprintf(stderr, "                  v = xxxxxxxx, where x is 0 or 1\n");
   fprintf(stderr, "                  e.g. -rsel 10100111\n");
   fprintf(stderr, "     -x xf      : 1 extra report per line in xf\n");
   fprintf(stderr, "                  e.g. 31 => MOD_PID report\n");
   fprintf(stderr, "     -off       : Add Mod_Sym_Offset report\n");
   fprintf(stderr, "     -disasm    : Add Mod_Sym_Offset report with disassembly\n");
   fprintf(stderr, "     -twin b e  : tprof windowing (b->e)\n");
   fprintf(stderr, "     -tpsel n   : select which tprof hooks to use\n");
   fprintf(stderr, "     -itsel n   : select a metric-field from itrace hook\n");
   fprintf(stderr, "     -tmod      : tag MODULE with 32b/64b id\n");
   fprintf(stderr, "     -d n       : Additional debug output\n");
   fprintf(stderr, "                  2) show major-minor per hook\n");
   fprintf(stderr, "                  3) cf details in gd.arc\n");
   fprintf(stderr, "                  4) embed nrm2 in gd.arc\n");
   fprintf(stderr, "                  5) +debug to post.show\n");
   fprintf(stderr, "                  6) show tprof addresses => stderr\n");
   fprintf(stderr, "                  7) use lr for tprof\n");
#if defined(_LINUX)
   fprintf(stderr, "     -compat    : Do not print the tprof mode line in tprof.out\n");
#endif   
   fprintf(stderr, "\n");
   fprintf(stderr, "   ITrace\n\n");
   fprintf(stderr, "     Use jprof jinsts option. Includes jitted code in jita2n file.\n");
   fprintf(stderr, "     -arc        : Create arc file (subroutine flow)\n");
   fprintf(stderr, "         reduction:\n");
   fprintf(stderr, "            arcf -F arc -t itrace -r 1\n");
   fprintf(stderr, "     -c          : compact arc (1 line / subroutine)\n");
   fprintf(stderr, "     -int3       : process between 1st & 2nd int3 hks\n");
   fprintf(stderr, "     -ss         : Show individual instructions\n");
   fprintf(stderr, "     -ssi        : Show individual instructions w opcode\n");
   fprintf(stderr, "     -m          : ITrace only (ignore tprof hooks)\n");
   fprintf(stderr, "                   (def: tprof)\n");
#ifdef CONFIG_PPC64
   fprintf(stderr, "     -qt      : Create single step qtrace output\n");
   fprintf(stderr, "     -ls      : Create a simple ascii output of load/store info \n");
#endif
   fprintf(stderr, "\n");

   exit(-1);
}

/**********************************/
void arc_hdr(void)
{
   fprintf(gd.arc, "\n");
   fprintf(gd.arc, "# arc Field Definition:\n");
   fprintf(gd.arc, "#   1: cpu no.\n");
   fprintf(gd.arc, "#   2: K(kernel) or U(user)\n");
   fprintf(gd.arc, "#   3: last instruction type\n");
   fprintf(gd.arc, "#     0=INTRPT, 1=CALL,  2=RETURN, 3=JUMP,\n");
   fprintf(gd.arc, "#     4=IRET,   5=OTHER, 6=UNDEFD, 7=ALLOC\n");
   fprintf(gd.arc, "#   4: no. of instructions\n");
   fprintf(gd.arc, "#   5: @|? = Call_Flow | pid_tid\n");
   fprintf(gd.arc, "#   6: offset (from symbol start)\n");
   fprintf(gd.arc, "#   7: symbol:module\n");
   fprintf(gd.arc, "#   8: pid_tid_pidname_[threadname]\n");
   fprintf(gd.arc, "#   9: last instruction type (string)\n");
   fprintf(gd.arc, "#  10: line number (if available)\n");
   fprintf(gd.arc, "\n");
}

/**********************************/
void parmerr(char * s)
{
   fprintf(stderr, " Missing Parm for %s\n\n", s);
   usage();
}

/**********************************/
void parmerr2(char * s1, char * s2)
{
   fprintf(stderr, " Bad Option Combo : %s %s\n\n", s1, s2);
   usage();
}

/**********************************/
char * get_current_directory(void)
{
   char *p = NULL;
   char buf[4096];

#if defined(_LINUX) || defined(_ZOS) || defined(_AIX)
   getcwd(buf, 4096);
   strcat(buf, "/");
   p = strdup(buf);
#elif defined(_WINDOWS)
   GetCurrentDirectory(4096, buf);
   strcat(buf, "\\");
   p = strdup(buf);
#endif

   return( p );
}

/**********************************/
int main (int argc, char * argv[])
{
   char * nrmnm, * arcnm, * tpout, * tpaddr;
   char defn[]      = "swtrace.nrm2";
   char defa[]      = "arc";
   char deftp[]     = "tprof.out";
   char deftpaddr[] = "tprof.addr";
   char * a2nrfn = "";
   char * a2nrng[4] = { NULL};

   int64  nstt  = 0;
   int64  nstp  = INT64_MAX;

   char *p;

   initGV();

   gd.msg  = xopen("post.msg", "w");

   Ver_Bld_Plat(stdout);
   Ver_Bld_Plat(gd.msg);

   gv.clip = .001;                      // 0.1% tick filter
   gv.clip = 0;                         // tick filter
#if defined(BUILD_STAP)
   gv.stap = 1;				// stap filtering required
#  endif

   nrmnm   = defn;
   arcnm   = defa;
   tpout   = deftp;
   tpaddr  = deftpaddr;

   /* parse options */
   if (argc >= 2)
   {
      int  i;

      for (i = 1; i < argc; i++)        // Pre-parse debugging options
      {
         if (strcmp(argv[i], "-verbose") == 0)
         {
            gv.verbose |= gv_verbose_stderr;
         }

         else if (strcmp(argv[i], "-verbmsg") == 0)
         {
            gv.verbose |= gv_verbose_logmsg;
         }

         else if (strncmp(argv[i], "-mask=", 6) == 0)
         {
            gv.mask     = strtoul(argv[i]+6, NULL, 16);
            gv.verbose |= gv_verbose_logmsg;

            OptVMsgLog( "Parsed: -mask=0x%08X\n", gv.mask );
         }
      }

      // i=0 -> (argc - 1)
      for (i = 1; i < argc; i++)
      {
         if (argv[i][0] == '-')
         {                              /* option */
            if (strcmp(argv[i], "-verbose") == 0)
            {
               ;                        //pre-parsed
            }

            else if (strcmp(argv[i], "-verbmsg") == 0)
            {
               ;                        //pre-parsed
            }

            else if (strncmp(argv[i], "-mask=", 6) == 0)
            {
               ;                        //pre-parsed
            }

            else if (strcmp(argv[i], "-tpsel") == 0)
            {
               i++;
               if (i >= argc) parmerr("-tpsel");
               gv.tpsel = atoi(argv[i]);   /* which tprof hooks */
            }

            else if (strcmp(argv[i], "-undup") == 0)
            {
               gv.undup = 1;
            }

            else if (strcmp(argv[i], "-oprof") == 0)
            {
               gv.oprof = 1;
            }

            else if (strcmp(argv[i], "-itsel") == 0)
            {
               i++;
               if (i >= argc) parmerr("-itsel");
               gv.itsel = atoi(argv[i]);   /* which itrace MM field */
            }

            else if (strcmp(argv[i], "-tmod") == 0)
            {
               gv.tmod = 1;             // tag module with (32bit) (64bit)
            }


            else if (strcmp(argv[i], "-r") == 0)
            {
               i++;
               if (i >= argc) parmerr("-r");
               nrmnm = argv[i];         /* nrm2 file */
            }

            else if (strcmp(argv[i], "-anysym") == 0)
            {
               a2nanysym = 1;
            }

            // a2n debug
            else if (strcmp(argv[i], "-a2ndbg") == 0)
            {
               gv.a2ndbg = 1;
            }

            // a2n rename duplicate symbols
            else if (strcmp(argv[i], "-a2nrdup") == 0)
            {
               gv.a2nrdup = 1;
            }

            // test jit a2n without nrm2
            else if (strcmp(argv[i], "-a2ntest") == 0)
            {
               a2ntestflag = 1;
            }

            // a2n range collapse file
            else if (strcmp(argv[i], "-a2nrfn") == 0)
            {
               i++;
               if (i >= argc) parmerr("-a2nrfn");
               a2nrfn = strdup(argv[i]);   // a2n range file
               A2nSetRangeDescFilename(a2nrfn);
            }

            else if (strcmp(argv[i], "-a2nmmi") == 0)
            {
               i++;
               if (i >= argc) parmerr("-a2nmmi");
               fprintf(gd.msg, " mmi %s\n", argv[i]);
               A2nSetCollapseMMIRange(argv[i]);
            }

            else if (strcmp(argv[i], "-a2nrng") == 0)
            {
               i++;
               if (i >= argc) parmerr("-a2nrng");
               a2nrng[0] = strdup(argv[i]);
               i++;
               if (i >= argc) parmerr("-a2nrng");
               a2nrng[1] = strdup(argv[i]);
               i++;
               if (i >= argc) parmerr("-a2nrng");
               a2nrng[2] = strdup(argv[i]);
               i++;
               if (i >= argc) parmerr("-a2nrng");
               a2nrng[3] = strdup(argv[i]);

               fprintf(gd.msg, " %s %s:%s %s\n",
                       a2nrng[0], a2nrng[1], a2nrng[2], a2nrng[3]);
               fflush(gd.msg);

               A2nCollapseSymbolRange(
                                     a2nrng[0], a2nrng[1], a2nrng[2], a2nrng[3]);
            }

            else if (strcmp(argv[i], "-noskew") == 0)
            {
               gv.noskew = 1;
            }
            else if (strcmp(argv[i], "-sympath") == 0)
            {
               i++;
               if (i >= argc) parmerr("-sympath");
               gv.sympath = strdup(argv[i]);
            }

#ifdef _AIX
            // aix ascii trace
            else if (strcmp(argv[i], "-aixtrc") == 0)
            {
               gv.aixtrc = 1;
               i++;
               if (i >= argc) parmerr("-aixtrc");
               trcInit(argv[i]);
            }
#endif

            // show ascii dump of nrm2
            else if (strcmp(argv[i], "-show") == 0)
            {
               gd.show = xopen("post.show", "w");
            }

            // force oldjit for comparison
            else if (strcmp(argv[i], "-oldjit") == 0)
            {
               gv.oldjit = 1;
            }

            // show ascii dump of nrm2, no other processing
            else if (strcmp(argv[i], "-showx") == 0)
            {
               gd.show = xopen("post.show", "w");
               gv.showx = 1;
            }

            // show ascii dump of nrm2, no other processing
            // add raw file data
            else if (strcmp(argv[i], "-showxx") == 0)
            {
               gd.show = xopen("post.show", "w");
               gv.showx = 2;
            }

            else if (strcmp(argv[i], "-clip") == 0)
            {
               i++;
               if (i >= argc) parmerr("-clip");
               if (argv[i][0] == '-')
               {
                  parmerr2(" -clip", argv[i]);
               }
               gv.clip = .01 * atof(argv[i]);   /* clip % */
            }

            else if (strcmp(argv[i], "-c") == 0)
            {
               gv.compact = 1;          /* compact arc file */
            }

            else if (strcmp(argv[i], "-k") == 0)
            {
               i++;
               if (i >= argc) parmerr("-k");
               knmptr = strdup(argv[i]);   /* kernel file */
               knew |= 0x1;
            }
            else if (strcmp(argv[i], "-kmap") == 0)
            {
               i++;
               if (i >= argc) parmerr("-kmap");
               kmap = strdup(argv[i]);  /* kernel map file */
               knew |= 0x2;
            }
            else if (strcmp(argv[i], "-kas") == 0)
            {
               knew |= 0x4;             /* use kallsyms and modules files */
               kas_off = kas_size = kmd_off = kmd_size = 0;
            }
            else if (strcmp(argv[i], "-nv") == 0)
            {
               gv.nv = 1;               /* dont validate symbols */
            }
            else if (strcmp(argv[i], "-nsf") == 0)
            {
               a2n_nsf = 1;             /* NoSymbols/NoSymbolFound instead of section */
            }
            else if ( strcmp(argv[i], "-off") == 0
                      || strcmp(argv[i], "-disasm") == 0 )
            {
               if ( 0 == gv.off )
               {
                  gv.off = 1;           /* Mod_Sym_Offset report */
                  gd.micro = xopen("tprof.micro", "w");   /* micro reports */
                  mlegend();
                  gv.micro = 1;         /* micro reports */
               }
               if ( strcmp(argv[i], "-disasm") == 0 )
               {
                  gv.disasm = 1;
               }
            }
            else if (strcmp(argv[i], "-compat") == 0)
            {
               i++;
               gv.tprof_compat = 1;
            }
            else if (strcmp(argv[i], "-x") == 0)
            {
               i++;
               if (i >= argc) parmerr("-x");
               gd.xfd = xopen(argv[i], "r");   /* extra reports */
               fprintf(gd.msg, " x option : %s\n", argv[i]);
            }
            else if (strcmp(argv[i], "-jdir") == 0)
            {
               i++;
               if (i >= argc) parmerr("-jdir");
               jitpre = strdup(argv[i]);   /* jita2n file(s) prefix */
            }
            else if (strcmp(argv[i], "-lmrg") == 0)
            {
               i++;
               if (i >= argc) parmerr("-lmrg");
               mrgfnm = strdup(argv[i]);   /* merge file name */
            }
            else if (strcmp(argv[i], "-ss") == 0)
            {
               gv.showss = 1;
            }
            else if (strcmp(argv[i], "-ssi") == 0)
            {
               gv.showss = 2;
            }
            else if (strcmp(argv[i], "-cpu") == 0)
            {
               gv.cpu = 1;
            }
            else if (strcmp(argv[i], "-nt") == 0)
            {
               gv.ntonly = 1;
            }
            else if (strcmp(argv[i], "-nospace") == 0)
            {
               gv.nospace = 1;
            }
            else if (strcmp(argv[i], "-arc") == 0)
            {
               gv.arc = 1;
            }
            else if (strcmp(argv[i], "-rsel") == 0)
            {
               i++;
               if (i >= argc) parmerr("-rsel");
               gv.rsel = 1;
               gv.rselvec = argv[i];
            }
            else if (strcmp(argv[i], "-pnm") == 0)
            {
               gv.pnmonly = 1;
            }
            else if (strcmp(argv[i], "-tnm") == 0)
            {
               gv.tnmonly = 1;
            }
            else if (strcmp(argv[i], "-int3") == 0)
            {
               int3f = 0;
               aout  = 0;
               fprintf(gd.msg, " Using int3: aout = %d\n", aout);
            }
            else if (strcmp(argv[i], "-d") == 0)
            {
               i++;
               if (i >= argc) parmerr("-d");
               gv.db = atoi(argv[i]);
               fprintf(stderr, " debug %d\n", gv.db);
            }
            else if (strcmp(argv[i], "-p") == 0)
            {
               i++;
               if (i >= argc) parmerr("-p");
               nstp = atoll(argv[i]);   /* 1st n recs of raw */
               fprintf(gd.msg, " nstp = %"_L64"d\n", nstp);
            }
            else if (strcmp(argv[i], "-m") == 0)
            {
               gv.mtrace = 1;           // mtrace
               fprintf(gd.msg, " -m option(mtrace)\n");
            }
            else if (strcmp(argv[i], "-pp") == 0)
            {
               i++;
               if (i >= argc) parmerr("-pp");
               nstt = atoll(argv[i]);   /* from */
               i++;
               if (i >= argc) parmerr("-pp");
               nstp = atoll(argv[i]);   /* to */
               fprintf(gd.msg, " nstt = %"_L64"d\n", nstt);
               fprintf(gd.msg, " nstp = %"_L64"d\n", nstp);
            }
            else if (strcmp(argv[i], "-twin") == 0)
            {
               // tprof reduction between tstt -> tstp
               i++;
               if (i >= argc) parmerr("-twin");
               gv.tstt = atoll(argv[i]);   /* from */
               i++;
               if (i >= argc) parmerr("-twin");
               gv.tstp = atoll(argv[i]);   /* to */
               fprintf(gd.msg, " tstt = %"_L64"d\n", gv.tstt);
               fprintf(gd.msg, " tstp = %"_L64d"\n", gv.tstp);
            }
#ifdef CONFIG_PPC64
            else if (strcmp(argv[i], "-qt") == 0)
            {
               gd.qtrace = xopen("output.qt", "w");
               gv.arc = 1;              /* qtrace requres arc */
               initialize_qtrace();
            }
            else if (strcmp(argv[i], "-ls") == 0)
            {
               gd.ls = xopen("output.ls", "w");
               gv.arc = 1;
               initialize_qtrace();
            }
#endif

            else if (argv[i][1] == '?')
            {
               usage();
            }
            else
            {
               fprintf(stderr, "\n INVALID OPTION: <%s>\n", argv[i]);
               usage();
            }
         }
         else
         {
            fprintf(stderr, "\n INVALID OPTION: <%s>\n", argv[i]);
            usage();
         }
      }
   }

   OptVMsg(" aft options\n");

   if (gv.off)
   {
      gv.clip = 0.0;
   }

   gf_nrm2 = nrmnm;

#if defined(BUILD_STAP)
   if (gv.stap)
   {
      int fd;

      strcpy(tmp_nrm2, "/tmp/stap_cvtXXXXXX");
      fd = mkstemp(tmp_nrm2);
      if (!tmp_nrm2[0])
      {
         fprintf(stderr, " ERROR unable to create temp nrm2 file\n");
         exit(1);
      }
      if (!stap_to_nrm2(nrmnm, tmp_nrm2))
      {
         exit(1);
      }
      gf_nrm2 = tmp_nrm2;
   }
#  endif
   
   if (jitpre == NULL)
   {
#if !defined(_WINDOWS)
      jitpre = default_jita2n_file_prefix;
#else
      char * p;
      char tbuf[MAX_PATH];

      p = getenv("IBMPERF_JPROF_LOG_FILES_PATH");
      if (p != NULL)
      {
         strcpy(tbuf, p);
         strcat(tbuf, "\\");
         p = getenv("IBMPERF_JPROF_LOG_FILES_NAME_PREFIX");
         if (p != NULL)
         {
            strcat(tbuf, p);
         }
         else
         {
            strcat(tbuf, default_jita2n_file_prefix);
         }
         jitpre = strdup(tbuf);
      }
      else
      {
         jitpre = default_jita2n_file_prefix;
      }
#endif
   }

   // open files & init A2N
   if (gv.showx == 0)
   {                                    // showx=1 => no processing
      if (gv.arc)
      {
         gd.arc = xopen(arcnm,   "w");
         arc_hdr();
      }
//    gd.ptree = xopen("ptree", "w");

      if (initSymUtil() != 0)
      {
         fprintf(stderr, "ERROR initing SymUtil\n");
         exit(1);
      }
   }

   // Open Files
   if (gv.aixtrc != 1)
   {
      if (init_raw_file(gf_nrm2) != BP_OK)
      {
         fprintf(stderr, " Error opening %s\n", gf_nrm2);
         exit(1);
      }
   }

   if (gv.noskew == 1)
   {
      noskew();
   }

   // tprof. init tree
   root = initTree();

   // nrm2 structure
   tr = (trace_t *)malloc(sizeof(trace_t));

   if ( NULL == tr )
   {
      fprintf(stderr, " ** Out of memory in main(int,char **) ***\n");
      exit(1);
   }

   // READ TRACE FILE ( *nrm2 or aixtrace )
   if (gv.aixtrc != 1)
   {
      readnrm2(nstt, nstp);
   }
#ifdef _AIX
   else
   {
      readAixTrc();
   }
#endif



   // a2n quality
   if (necnt1 + necnt2 >= 1)
   {
      OptVMsg(" %4d %s\n",
              necnt1, " UnExpected Brk. Prev Hk Callflow");
      OptVMsg(" %4d %s\n",
              necnt2, "   Expected Brk. Prev Hk Other");
   }

   if (a2nrc[0] >= 1)
   {
      char s[6][8] =
      {"Good", "NoSymFd", "NoSyms", "NoMod", "+++", "---"};
      int i;

      OptVMsg("\n %8s %s\n", "", "A2N-Quality");
      for (i = 0; i <= 5; i++)
      {
         if (a2nrc[i] >= 1)
         {
            OptVMsg(" %8d %s\n", a2nrc[i], s[i]);
         }
      }
   }

   OptVMsg(" END OF TRACE. Start of Reports\n");

   if ( ( (gv.mtrace == 0) && (hmaj[0x10] >= 1) )
        || gv.aixtrc == 1 )
   {                                    // tprof
      gd.tp     = xopen("tprof.out", "w");
      output();
   }

   if ( 0 != ( p = get_current_directory() ) )
   {
      WriteFlush(stdout, "\nOutput files placed in directory \"%s\".\n", p);
      free(p);
   }

#ifdef CONFIG_PPC64
   if (gd.qtrace) stop_qtrace();
#endif

   //if(gv.ptree == 1) {
   //   xtreeOut("ptree");
   //}

   return(0);
}

/**********************************/
void setrpt(char * p)
{
   int k, i = 0;

   //fprintf(gd.msg, " XX Report: %s\n", p);

   // parse a report definition (ie 134 into global vars)
   while (*p != '\0')
   {
      i++;
      k = (int)(*p) - (int)'0';

      //fprintf(gd.msg, " YY k %d tlab[k]  %s\n", k, tlab[k]);
      //fprintf(gd.msg, " ZZ k %d tlabx[k] %s\n", k, tlabx[k]);

      gv.xind[i] = k;                   // k-lev in orig => i-th lev in new
      gv.flab[i] = tlab[k];             // assoc level label (i.e. 3 => MOD)
      gv.rlab[i] = tlabx[k];            // assoc title label (i.e _MODULE)
      gv.xlevs = i;                     // total levels in new tree
      //fprintf(gd.msg,
      //      " XX i=%d xind %2d flab %8s rlab %8s\n",
      //      i, k, tlab[k], tlabx[k]);
      p++;
   }
}

/**********************************/
void html_link(int i)
{
   if (0 == 1)
   {
      fprintf(gd.tp, " <A HREF=\"#R_%d\">))</A>\n", i);
   }
}

/**********************************/
void html_tag(int i)
{
   if (0 == 1)
   {
      fprintf(gd.tp, "<A NAME=\"R_%d\"></A>\n", i);
   }
}

/**********************************/
void rtitle(int link)
{
   int i;

   if ( (0 == 1) && (link >= 0))
   {                                    // html link
      fprintf(gd.tp, "   <A HREF=\"#R_%d\">)) %s",
              link, gv.rlab[1]);
   }
   else
   {
      fprintf(gd.tp, "    )) %s", gv.rlab[1]);
   }

   //fprintf(gd.tp, "    )) %s", gv.rlab[1]);

   for (i = 2; i <= gv.xlevs; i++)
   {
      fprintf(gd.tp, "_%s", gv.rlab[i]);
   }

   if ((0 == 1) && link >= 0)
   {                                    // html link
      fprintf(gd.tp, "</A>\n");
   }
   else
   {
      fprintf(gd.tp, "\n");
   }
}

/**********************************/
void xfixuppid(pTNode root)
{
   // scan root children(pids)
   // fixup pnm if possible

   // walk pid only and print
}

/**********************************/
// root  - global tick database
// p     - the report type 4321 => SMTP
// tklev - ticks lev in database (4 or 5(w off))
void xreport(pTNode root, char * p, int tklev)
{
   pTNode xroot;

   setrpt(p);                           // set global vars (select data @ level)

   fprintf(gd.tp, "\n   %s\n", ban);    // report banner
   rtitle(-1);
   fprintf(gd.tp, "   %s\n\n", ban);

   xroot = initTree();                  // new root
   tree_new(root, 0, xroot, tklev);     // tree orig(root) => new(xroot);
   tree_cum(xroot, 0);                  // cum
   tree_sort_out_free(xroot, 0, gd.tp); // sort.out.free

   fflush(gd.tp);
}

/**********************************/
void xinput(pTNode root)
{
   char buf[128];
   int rc = 1;
   char * cp;

   while (rc == 1)
   {
      if (fgets(buf, 128, gd.xfd) != NULL)
      {
         gv.xlevs = 0;
         cp = buf;
         while (*cp == ' ') cp++;       // skip leading blanks
         while ( (*cp != ' ') && (*cp != '\0')  && (*cp != '\n') )
         {
            cp++;
         }
         *cp = '\0';
         xreport(root, buf, 4 + gv.off);   // 4: sym 5: offset
      }
      else rc = 0;
   }
}

/**********************************/
void output(void)
{
   int    nreps = 8;
   int    i;
   int    tmp;
   double tcycs, tt, pc;
   int    tlim;
#ifdef _S390
   double k390cycs2sec = 4096.0 * 1000000.0;
#endif
   //int timebase;

   fprintf(gd.tp, "<pre>\n\n Tprof Reports\n\n");
   fprintf(gd.tp, " Platform : %s\n\n", _PLATFORM_STR);

   // AIX zOS
   if (gv.clock_rate  != 0)
   {

      tcycs = trstp - trstt;

      OptVMsgLog(" trstt %f\n", trstt);
      OptVMsgLog(" trstp %f\n", trstp);
      OptVMsgLog(" tcycs %f\n", tcycs);

#ifdef _S390
      tt = tcycs / k390cycs2sec;
#else
      if (gv.timebase > 0)
      {
         tt = tcycs / gv.timebase;
      }
      else
      {
         tt = tcycs / gv.clock_rate;
      }
      fprintf(gd.tp, " %20s %15.0f\n", "ProcessorSpeed", gv.clock_rate);
#endif

      fprintf(gd.tp, " %20s %15.0f\n", "TraceCycles", tcycs);
      fprintf(gd.tp, " %20s %15.3f(sec)\n\n", "TraceTime", tt);

#if defined(_LINUX) && !defined(BUILD_STAP)
      if (!gv.tprof_compat)
      {
         if (gv.tprof_mode)
         {
            // event-based
            fprintf(gd.tp, " %20s, ticks every %d %s[%d]\n", "Event-based mode", 
                    gv.tprof_count, GetPerfCounterEventNameFromId(gv.tprof_event), gv.tprof_event);
         }
         else
         {
            // time-based
            fprintf(gd.tp, " %20s: ticks every %d microseconds\n", "Time-based mode", gv.tprof_count);
         }
      }
#endif            
   }

   // select reports (via the rsel option)
   if (gv.rsel == 1)
   {
      gv.ntonly = 1;                    // turn off defaults
      for (i = 0; i < 8; i++)
      {
         ropt[i] = 0;
         if (gv.rselvec[i] == '1')
         {
            ropt[i] = 1;
         }
      }
   }

   //mlev = 5;  // ???

   // DATA BASE tree from "root"
   tree_cum(root, 0);
   tree_sort(root, 0);
   gv.tottks = root->cum;

   // clumsy log(10) of tottks
   tmp = gv.tottks;
   for (i = 1; i < 10; i++)
   {
      if (tmp < 10)
      {
         gv.twidth = i;
         break;
      }
      tmp = tmp / 10;
   }

   fprintf(gd.tp, "\n\n TOTAL TICKS %14d\n", gv.tottks);

   tlim = (int)(gv.clip * gv.tottks);
   pc   = 100 * gv.clip;

   fprintf(gd.tp, " (Clipping Level : %5.1f %%  %3d Ticks)\n\n", pc, tlim);
   fprintf(gd.tp, "\n   %s\n", ban);
   fprintf(gd.tp, "      TPROF Report Summary\n\n");

   // TITLES for selected reports
   for (i = 0; i < nreps; i++)
   {
      if ( (gv.ntonly == 0) || (ropt[i] == 1) )
      {
         setrpt(rlev[i]);
         //html_link(i);
         rtitle(i);                     // generate HTML links here
      }
   }
   fprintf(gd.tp, "   %s\n\n", ban);

   for (i = 0; i < nreps; i++)
   {
      if ( (gv.ntonly == 0) || (ropt[i] == 1) )
      {
         // generate HTML tags here
         html_tag(i);
         xreport(root, rlev[i], 4 + gv.off);
      }
   }

   // extra reports
   if (gd.xfd != NULL)
   {
      fprintf(gd.tp, "\n\n   Extra Tree Reports\n");
      //fprintf(gd.tp, "    Module-Process\n");
      xinput(root);
   }

   // tprof.micro report
   if (gv.off)
   {                                    // generate Mod-Sym-Offset report
      fprintf(gd.tp, "\n Offset Report\n");
      xreport(root, "345", 4 + gv.off);
   }
   fprintf(gd.tp, "</pre>\n");
   fflush(gd.tp);
}
