/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2007
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdint.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <bfd.h>

#include "post.h"
#include "bputil.h"
#include "a2n.h"
#include "qtlib.h"

/* from pi_insttrace.h */
#define PI_ITRACE_MINOR_STORE           1
#define PI_ITRACE_MINOR_RELATIVE_IP     2
#define PI_ITRACE_MINOR_RELATIVE_LS     4
#define PI_ITRACE_MINOR_VSID            8 
#define PI_ITRACE_MINOR_REPEAT_VSID     16
#define PI_ITRACE_MINOR_DATA_LENGTH     32
#define PI_ITRACE_MINOR_INST_VSID       64

struct status_struct {
	uint64_t rec_count;
	int trace_on;
	uint64_t prev_ss;
	uint64_t prev_jump;
	uint64_t prev_ls_target;
	uint64_t prev_vsid;
	int header_written;
	qtrace_rec curr_qrecord;
	uint64_t curr_qrecord_addr;
	uint64_t inst_count;
	uint64_t error_inst_count;
	uint64_t mod_error_inst_count;
};

#define PROCLINE_MAX_LEN 640
struct module_entry_struct {
	char name[PROCLINE_MAX_LEN];
	uint64_t addr, size;
	char name_lookup_failed;
	bfd *abfd;
	char fq_filename[PROCLINE_MAX_LEN];
};
struct module_entry_struct *module_cache;
unsigned int num_module_cache_entries;
/* for some reason we SIGSEGV if > 10 bfd_openr's are performed */
unsigned int bfd_cache_slots_left = 10;

/* global status variable */
struct status_struct status;

void write_single_steps(trace_t *, unsigned int, FILE *qtfile);
void update_status(trace_t *);
void cache_modules();
void free_module_cache();
void populate_untaken_branch(qtrace_rec *);

/* ----------------------------------------------------------------------- */
void initialize_qtrace()
{
	status.rec_count = 0;
	status.trace_on = 0;
	status.prev_ss = 0;
	status.prev_jump = 0;
	status.prev_ls_target = 0;
	status.prev_vsid = 0;
	status.header_written = 0;
	status.curr_qrecord_addr = 0;
	status.inst_count = 0;
	status.error_inst_count = 0;
	status.mod_error_inst_count = 0;

	bfd_init();
	cache_modules();
}
/* ----------------------------------------------------------------------- */
void write_qtrace(trace_t *tr, unsigned int pid, FILE *qtfile)
{
	write_single_steps(tr, pid, qtfile);
	update_status(tr);
}
/* ----------------------------------------------------------------------- */
void stop_qtrace()
{
	int i = 0;

	while (i < num_module_cache_entries) {
		if (strlen(module_cache[i].fq_filename))
			printf("INFO: Used module file %s\n", module_cache[i].fq_filename);
		i++;
	}

	free_module_cache();
	if (status.error_inst_count)
			fprintf(stderr, "WARNING: inserted %lld nops\n",
				status.error_inst_count);
	if (status.mod_error_inst_count)
			fprintf(stderr, "WARNING: inserted %lld nops for module\n",
				status.mod_error_inst_count);
	printf("Total of %lld instructions processed\n", status.inst_count);
}
/* ======================================================================= */
uint64_t build64(unsigned int high, unsigned int low)
{
        uint64_t rv;
        rv = high;
        rv = rv << 32;
        rv += low;
        return rv;
}
/* ----------------------------------------------------------------------- */
/* recursively search dirname for filename.  If found, store results in memory
 * pointed to by result.  If not found, result is unmodified. */
void find_file_in_dir(char *result, char *filename, char *dirname)
{
	DIR *dirp;
	struct dirent *entry;
	char *found_file = NULL;
	struct stat sbuf;
	char full_name[512];

	dirp = opendir(dirname);
	entry = readdir(dirp);
	while ((entry) && (!found_file)) {
		if (entry->d_name[0] != '.') {
				strcpy(full_name, dirname);
				strcat(full_name, entry->d_name);
				stat(full_name, &sbuf);
				if (S_ISDIR(sbuf.st_mode)) {
					strcat(full_name, "/");
					find_file_in_dir(result, filename, full_name);
				} else if (!(strcmp(entry->d_name, filename))) {
					strcpy(result, dirname);
					strcat(result, filename);
					found_file = result;
				}
		}
		entry = readdir(dirp);
	}
	closedir(dirp);
}
/* ----------------------------------------------------------------------- */
/* search in a few well known places for a module that has the name in
 * module->name.  If found, bfd_open it and fill in module->abfd.  Return 1 on
 * success, 0 on failure */
int cache_module_bfd(struct module_entry_struct *module)
{
	struct utsname name_info;
	char module_dir_name[80], filename[80], fq_filename[PROCLINE_MAX_LEN];

	/* If named lookup already failed once, don't try again. */
	if (module->name_lookup_failed)
		return 0;

	fq_filename[0] = '\0';
	uname(&name_info);

	/* first search /lib/modules/<version>/kernel */
	strcpy(module_dir_name, "/lib/modules/");
	strcat(module_dir_name, name_info.release);
	strcat(module_dir_name, "/kernel/");
	strcpy(filename, module->name);
	strcat(filename, ".ko");
	find_file_in_dir(fq_filename, filename, module_dir_name);

	/* next try /lib/modules/<version>/kernel */
	if (!strlen(fq_filename)) {
		strcpy(module_dir_name, "/lib/modules/");
		strcat(module_dir_name, name_info.release);
		strcat(module_dir_name, "/extra/");
		find_file_in_dir(fq_filename, filename, module_dir_name);
	}

	/* finally try /usr/src/linux */
	if (!strlen(fq_filename)) {
		find_file_in_dir(fq_filename, filename, "/usr/src/linux/");
	}

	if (!strlen(fq_filename))
		goto fail; /* not found */

	module->abfd = bfd_openr(fq_filename, NULL);
	if (!module->abfd)
		goto fail;

	if (bfd_check_format(module->abfd, bfd_object)) {
		strncpy(module->fq_filename, fq_filename, PROCLINE_MAX_LEN);
		return 1;
	}

	fail:
		/* remember that this lookup has failed */
		module->name_lookup_failed = 1;
		return 0;
}
/* ----------------------------------------------------------------------- */
void cache_modules()
{
	FILE *modules;
	char procline[PROCLINE_MAX_LEN], number[PROCLINE_MAX_LEN];
	char *tok;
	int i = 0;

	modules = fopen("/proc/modules", "r");
	fgets(procline, PROCLINE_MAX_LEN - 1, modules);
	while (!feof(modules)) {
		i++;
		fgets(procline, PROCLINE_MAX_LEN - 1, modules);
	}
	fclose(modules);
	module_cache = malloc(sizeof(struct module_entry_struct) * i);
	num_module_cache_entries = i;

	i = 0;
	modules = fopen("/proc/modules", "r");
	fgets(procline, PROCLINE_MAX_LEN - 1, modules);
	while (!feof(modules)) {
		if (strstr(procline, "Live")) {
			tok = strtok(procline, " ");
			strncpy(module_cache[i].name, tok, 640);
			tok = strtok(NULL, " ");
			module_cache[i].size = strtol(tok, NULL, 10);
			tok = strtok(NULL, " ");
			tok = strtok(NULL, " ");
			tok = strtok(NULL, " ");
			tok = strtok(NULL, " ");
			strncpy(number, tok, 640);
			number[strlen(number) - 1] = '\0';
			module_cache[i].addr = strtol(number + 3, NULL, 16);
			module_cache[i].addr += 0xd000000000000000ULL;
			i++;
		}
		fgets(procline, PROCLINE_MAX_LEN - 1, modules);

	}
	fclose(modules);
}
/* ----------------------------------------------------------------------- */
void free_module_cache()
{
	int i;
	for (i = 0; i < num_module_cache_entries; i++)
		if (module_cache[i].abfd)
			bfd_close(module_cache[i].abfd);
	free(module_cache);
}
/* ----------------------------------------------------------------------- */
/* Returns the instruction address for the trace record tr.  For jump records,
 * the instruction address is the from address. */
uint64_t rec_inst_addr(trace_t *tr)
{
        switch (tr->major) {
        case 0x20:
        case 0x21:
                return build64(tr->ival[0], tr->minor);
                break;
        case 0x31:
		if (tr->minor & PI_ITRACE_MINOR_RELATIVE_IP)
			return status.prev_jump + tr->ival[0];
		else
			return build64(tr->ival[1], tr->ival[0]);
                break;
        default:
                return 0;
        }
}
/* ----------------------------------------------------------------------- */
uint64_t rec_load_store_target_addr(trace_t *tr)
{
	int ival_start = 1;
	/* only makes sense for rec 0x31 major code */
	if (tr->major != 0x31)
		return 0;

	/* if the full instruction address is present, the target address is
	 * moved over one slot. */
	if (!(tr->minor & PI_ITRACE_MINOR_RELATIVE_IP))
		ival_start++;

	if (tr->minor & PI_ITRACE_MINOR_RELATIVE_LS)
		return status.prev_ls_target + (int) tr->ival[ival_start];
	else
		return build64(tr->ival[ival_start + 1], tr->ival[ival_start]);
}
/* ----------------------------------------------------------------------- */
uint64_t rec_vsid(trace_t *tr)
{
	int ival_start = 2;
	/* only makes sense for rec 0x31 major code */
	if (tr->major != 0x31)
		return 0;

	if (tr->minor & PI_ITRACE_MINOR_VSID) {
		/* if the full instruction address or target address is present,
		 * the target address is moved over one slot. */
		if (!(tr->minor & PI_ITRACE_MINOR_RELATIVE_IP))
			ival_start++;
		if (!(tr->minor & PI_ITRACE_MINOR_RELATIVE_LS))
			ival_start++;
		return build64(tr->ival[ival_start + 1], tr->ival[ival_start]);
	} else if (tr->minor & PI_ITRACE_MINOR_REPEAT_VSID)
		return status.prev_vsid;
	else
		return 0; /* should never happen */
}
/* ----------------------------------------------------------------------- */
unsigned int rec_load_store_data_length(trace_t *tr)
{
	int loc = 2;
	/* only makes sense for rec 0x31 major code */
	if (tr->major != 0x31)
		return 0;

	if (tr->minor & PI_ITRACE_MINOR_DATA_LENGTH) {
		/* if the full instruction address, target address, or vsid is
		 * present, the target address is moved over one slot. */
		if (!(tr->minor & PI_ITRACE_MINOR_RELATIVE_IP))
			loc++;
		if (!(tr->minor & PI_ITRACE_MINOR_RELATIVE_LS))
			loc++;
		if (tr->minor & PI_ITRACE_MINOR_VSID)
			loc+= 2;

		return tr->ival[loc];
	} else
		return 0;
}
/* ----------------------------------------------------------------------- */
/* Does a linear search through the module cache array for the module
 * containing the given address.  If no module is found, return NULL */
struct module_entry_struct *get_module(uint64_t addr)
{
	int i = 0;
	struct module_entry_struct *module = NULL;

	while ((i < num_module_cache_entries) && (!module)) {
		if ((addr >= module_cache[i].addr) && 
			(addr <  module_cache[i].addr +  module_cache[i].size))
			module = &(module_cache[i]);
		i++;
	}

	return module;
}
/* ----------------------------------------------------------------------- */
unsigned int get_module_instruction(uint64_t addr)
{
	struct module_entry_struct *module;
	uint64_t offset;
	char bytes[8];
	int rc;
	static char printed_warning = 0;

	module = get_module(addr);
	if (!module)
		return 0;
	offset = addr - module->addr;

	if ((!module->abfd) && (bfd_cache_slots_left)) {
		rc = cache_module_bfd(module);
		if (!rc) {
			if ((!bfd_cache_slots_left) && (!printed_warning)) {
				printed_warning = 1;
				fprintf(stderr, "WARNING: ran out of BFD slots\n");
			}
			return 0;
		}
		else
			bfd_cache_slots_left--;
	}

	rc = bfd_get_section_contents(module->abfd, module->abfd->sections, bytes, 
		offset, 4);
	if (!rc)
		return 0;
	return *((unsigned int *) &bytes);
}
/* ----------------------------------------------------------------------- */
/* Attempt to figure out instructions in the dynamically generated plt section.
 * This is only a concern for 32-bit executables on powerpc.  See
 * sysdeps/powerpc/powerpc32/dl-machine.c from the glibc sources to see how
 * the dynamic loader generates these.

 * For now, we only deal with one of the PLT entry types - a direct jump to the
 * actual routine.  TODO: handle the other PLT entry types. */
unsigned int get_plt_instruction(uint64_t addr, trace_t *tr)
{
	static uint64_t last_seen_plt_addr = 0;
	static uint64_t current_plt_start_addr = 0;

	uint64_t tr_addr = rec_inst_addr(tr);

	if ((addr - 4) != last_seen_plt_addr)
		current_plt_start_addr = addr;
	last_seen_plt_addr = addr;

	/* if this instruction matches the instruction in the trace record, and 
	 * this is a jump instruction ... */
	if ((tr_addr == addr) && ((tr->major == 0x20) || (tr->major == 0x21)))
		/* If this is the first instruction in this section of the plt ... */
		if (addr == current_plt_start_addr) {
			/* This is a single instruction plt section.  Reset static vars */
			current_plt_start_addr = last_seen_plt_addr = 0;

			/* Check for case 1 from dl-machine.c */
			uint64_t jump_target = build64(tr->ival[2], tr->ival[1]);
			int64 delta = jump_target - addr;
			/* will this jump fit in a relative branch inst? */
			if (delta << 14 >> 14 == delta)
				/* relative branch */
				return 0x48000000 | (delta & 0x03fffffc);
			else {
				/* probably absolute branch but verify a couple of things */
				if (jump_target & 0xffffffff00000000ULL)
					/* jump target > 32 bits - punt */
					return 0;
				else if (jump_target <= 0x01fffffc || jump_target >= 0xfe000000)
					/* verified absolute branch */
					return 0x48000002 | (jump_target & 0x03fffffc);
				else
					/* jump target will not fit into branch inst - punt */
					return 0;
			}
		}

	tr_addr = build64(tr->ival[2], tr->ival[1]);
	return 0;

}
/* ----------------------------------------------------------------------- */
/* Returns the 32-bit instruction at the given address.  If the instruction
 * can't be found, returns no-op instruction. */
unsigned int get_instruction(uint64_t addr, unsigned int pid, trace_t *tr)
{
	SYMDATA symbol_data;
	char *inst_ptr;

	/* hack - these ppc64 addresses come from kernel modules which a2n does not
	 * handle right now.  Use get_module_instruction to get these insts. */
	if ((addr >> 56) == 0xd0) {
		unsigned int mod_inst = get_module_instruction(addr);
		if (mod_inst)
			return mod_inst;
		else {
			status.mod_error_inst_count++;
			return 0x60000000; /* noop instruction */
		}
	}

	A2nGetSymbol(pid, addr, &symbol_data);

	if (!strcmp(A2N_SYMBOL_PLT_STR, symbol_data.sym_name)) {
		/* attempt to determine dynamic instructions from plt section */
		unsigned int plt_inst = get_plt_instruction(addr, tr);
		if (plt_inst)
			return plt_inst;
		else {
			status.error_inst_count++;
			return 0x60000000;
		}
	}

	if (!symbol_data.code) {
		status.error_inst_count++;
		return 0x60000000;
	} else
		inst_ptr = symbol_data.code;

	return *((unsigned int *)inst_ptr);
}
/* ----------------------------------------------------------------------- */
void write_qheader(uint64_t addr, FILE *qtfile)
{
	qtlib_header(addr, qtfile);
}
/* ----------------------------------------------------------------------- */
void write_qrecord(qtrace_rec *rec, FILE *qtfile)
{
	status.inst_count++;
	qtlib_record(rec, qtfile);
}
/* ----------------------------------------------------------------------- */
void populate_qrecord(qtrace_rec *rec, unsigned int inst)
{
	memset(rec, 0, sizeof(qtrace_rec));
	rec->instruction = inst;
	populate_untaken_branch(rec);
}
/* ----------------------------------------------------------------------- */
int inst_is_branch(unsigned int inst)
{
	unsigned int majorOpcode = inst >> 26;
	unsigned int minorOpcode;

	switch (majorOpcode) {
	case 16:
	case 17:
	case 18:
		return 1;
	case 19:
		minorOpcode = (inst >> 1) & 0x000003ff;
		if ((minorOpcode == 16) || (minorOpcode == 528))
			return 1;
	}
	return 0;
}
/* ----------------------------------------------------------------------- */
void populate_untaken_branch(qtrace_rec *rec)
{
	if (inst_is_branch(rec->instruction)) 
		rec->flags |= (QTRACE_REC_BRANCH | QTRACE_REC_UNTAKEN_BRANCH);
}
/* ----------------------------------------------------------------------- */
int inst_is_fake_branch(unsigned int pid, uint64_t addr)
{
	SYMDATA prev_inst_sym;
	unsigned int prev_inst;

	A2nGetSymbol(pid, addr - 4, &prev_inst_sym);
	if ((unsigned long) prev_inst_sym.code > 0x1000) {
			prev_inst = *((unsigned int *) prev_inst_sym.code);
			if ((prev_inst == 0x7c000164) || (prev_inst == 0x7fa10164))
					return 1;
	}
	return 0;
}
/* ----------------------------------------------------------------------- */
/* Insert special data into the given qrecord from the given trace record. */
void special_qrecord(qtrace_rec *rec, trace_t *tr, unsigned int pid)
{
	uint64_t rec_target;
	uint64_t vsid;
	uint64_t data_length;

	switch (tr->major) {
	case 0x20:
	case 0x21:
		/* some cases where the instruction might not be a branch:
			- interrupt
		    - avoiding MSR_RI-clear section (kernel driver should not flag as
				0x20 or 0x21)
		   A no-op on an instruction that is marked as a branch comes from a
		   failure in the address -> decoded instruction translation.  We 
		   should still mark this as a branch in the qtrace output.
		*/
		if ((inst_is_branch(rec->instruction)) || 
			(rec->instruction == 0x60000000)) {
			rec->flags |= QTRACE_REC_BRANCH;
			rec->flags &= ~QTRACE_REC_UNTAKEN_BRANCH;

			rec_target = build64(tr->ival[2], tr->ival[1]);
			rec->instruction_ea = rec_target;
		}
		break;
	case 0x31:
		rec->flags |= QTRACE_REC_LOAD_STORE;
		rec_target = rec_load_store_target_addr(tr);
		rec->data_ea = rec_target;

		if ((tr->minor & PI_ITRACE_MINOR_VSID) ||
			(tr->minor & PI_ITRACE_MINOR_REPEAT_VSID)) {
			rec->flags |= QTRACE_REC_DATA_VSID;
			vsid = rec_vsid(tr);
			rec->data_vsid = vsid;
		}

		if (tr->minor & PI_ITRACE_MINOR_DATA_LENGTH) {
			rec->flags |= QTRACE_REC_LENGTH;
			data_length = rec_load_store_data_length(tr);
			rec->length = data_length;
		}
		break;
	default:
		fprintf(stderr,
			"special_qrecord passed unexpected major code: %d\n",
			tr->major);
	}
}
/* ----------------------------------------------------------------------- */
/* This protects us from writing out potentially millions of qtrace records
 * if there is some kind of discontinuity in the single step addresses. */
int sensible_single_step_range(uint64_t from, uint64_t to)
{
	return from && (to > from) && (to - from <= 0x1000);
}
/* ----------------------------------------------------------------------- */
/* Outputs qtrace records for all instructions bewteen the previous call to
 * write_single_steps and the given trace record tr.  We do not write a trace
 * record for the instruction that tr refers to because it might be incomplete -
 * the next trace record might refer to the same instruction. */
void write_single_steps(trace_t *tr, unsigned int pid, FILE *qtfile)
{
	uint64_t addr, place;
	unsigned int inst;
	qtrace_rec qrecord;

	if (!status.trace_on)
		return;

	addr = rec_inst_addr(tr);
	/* if NULL is returned, this trace record does not cause us to output
	 * single steps. */
	if (!addr)
		return;

	if (!(status.header_written)) {
		write_qheader(addr, qtfile);
		status.header_written = 1;
		status.prev_ss        = addr;		
	} else if (sensible_single_step_range(status.prev_ss, addr)) {
		if (status.curr_qrecord_addr) {

			if (status.curr_qrecord_addr == addr) {
			/* if this trace record refers to the same instruction
			 * as the previous one, just update the remembered
			 * qrecord and return. */
				special_qrecord(&status.curr_qrecord, tr, pid);
				return;
			}
			else
			/* otherwise write the already filled in qrecord to 
			 * disk and continue. */
				write_qrecord(&status.curr_qrecord, qtfile);
		}
		for (place = status.prev_ss + 4; place <= addr-4; place += 4) {
			inst = get_instruction(place, pid, tr);
			populate_qrecord(&qrecord, inst);
			write_qrecord(&qrecord, qtfile);
		}
		/* fill in the qrecord for the instruction that tr refers to.
		 * It will get written out to disk the next time that
		 * write_single_steps is called. */
		inst = get_instruction(addr, pid, tr);
		populate_qrecord(&status.curr_qrecord, inst);
		special_qrecord(&status.curr_qrecord, tr, pid);
		status.curr_qrecord_addr = addr;
		status.prev_ss           = addr;		
	}
}
/* ----------------------------------------------------------------------- */
void update_status(trace_t *tr)
{
	switch (tr->major) {
	case 0x20:
	case 0x21:
		status.prev_jump = build64(tr->ival[2], tr->ival[1]);
		status.prev_ss = status.prev_jump - 4;
		break;
	case 0x2e:
		status.trace_on = 1;
		break;
	case 0x2f:
		status.trace_on = 0;
		break;
	case 0x31:
		status.prev_ls_target = rec_load_store_target_addr(tr);
		status.prev_vsid = rec_vsid(tr);
		break;
	}
}
/* ----------------------------------------------------------------------- */


/* ----------------------------------------------------------------------- */
void write_ls(trace_t *tr, unsigned int pid, FILE *qtfile)
{	
	SYMDATA  sd  = {0};
	int rc;
	uint64_t inst_addr;	
	
	if (tr->major == 0x31) {
		if (!(tr->minor & PI_ITRACE_MINOR_INST_VSID)) {
			
			fprintf(qtfile, "%d ", tr->minor & PI_ITRACE_MINOR_STORE); // load/store
			
			inst_addr = rec_inst_addr(tr);
			jitrec(pid, tr, 0);               // jita2n catchup
			rc = getSymbol(pid, inst_addr, &sd);			
			fprintf(qtfile, "%x ", pid);                               // pid
			fprintf(qtfile, "%llx ", inst_addr);               // instruction address
			fprintf(qtfile, "%llx  ", rec_load_store_target_addr(tr)); // data address
			fprintf(qtfile, "%llx  ", rec_vsid(tr));                   // data address vsid 
			fprintf(qtfile, "%x ", rec_load_store_data_length(tr));   // data length
			fprintf(qtfile, "%s ", sd.owning_pid_name);
			fprintf(qtfile, "%s ", sd.mod_name);
			fprintf(qtfile, "%s \n", sd.sym_name);
		}
	}
	update_status(tr);	  
		
}

