/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2008
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#define MAX_NR_CPUS 32
#define PERFHEADER_BUFFER_SIZE 512
#define HDR_CPU_BIAS 1000

typedef struct hdr {
	int cpu;
	int len;
} hdr_t;

struct data_rec;
typedef struct data_rec {
	int len;
	unsigned char* datap;
	struct data_rec *next;
} data_rec_t;

struct per_cpu_list {
	data_rec_t *last;
	data_rec_t *list;
} per_cpu_records[MAX_NR_CPUS];

void insert_record(int cpu, data_rec_t *prec)
{
	prec->next = per_cpu_records[cpu].list;
	per_cpu_records[cpu].list = prec;

	if (!per_cpu_records[cpu].last) {
		per_cpu_records[cpu].last = prec;
	}
}


static void append_record(int cpu, data_rec_t *prec)
{
	if (!per_cpu_records[cpu].last) {
		per_cpu_records[cpu].list = prec;
		per_cpu_records[cpu].last = prec;
		return;
	}

	per_cpu_records[cpu].last->next = prec;
	per_cpu_records[cpu].last = prec;
}

static void dump_bytes (unsigned char *rawData, int rawDataLen)
{
    int i = rawDataLen;
	unsigned char *p = rawData;
	int count = 0;

	while (i--) {
		printf("%02x", *p);
		count++;
		p++;
		if (count % 32 == 0) 
			printf("\n");
		else if (count % 4 == 0)
			printf(" ");
	}
	if (count % 32 != 0) printf("\n");
}

static unsigned char *next_record(unsigned char *current)
{
	data_rec_t *prec = (data_rec_t *)malloc(sizeof(data_rec_t));
	hdr_t *overlay = (hdr_t *)current;
	int cpu = overlay->cpu;
	int len = overlay->len;


	prec->len = len;
	current += 2*sizeof(int);
	prec->datap = current;
	current += len;
	prec->next = 0;

	if (cpu >= HDR_CPU_BIAS) {
		/* headers written at the end of the trace but must be at the 
		** beginning of the NRM2 file, so that are specially tagged
		*/
		cpu -= HDR_CPU_BIAS;
		insert_record(cpu, prec);
	} else {
		append_record(cpu, prec);
	}
	return current;
}

static void parse_data(unsigned char *data_start, unsigned char *data_end)
{
	unsigned char *current = data_start;

	while (current != data_end) {
		current = next_record(current);
	}
}

static int write_nrm2(char *nrm2_name)
{
	FILE *nrm2 = NULL;
	data_rec_t *p;
	int i;
	int rc = 1;
	
	nrm2 = fopen(nrm2_name, "w");
	for (i = 0; i < MAX_NR_CPUS && per_cpu_records[i].list; i++) {
		p = per_cpu_records[i].list;
		while (p) {
			if (!fwrite(p->datap, 1, p->len, nrm2)) {
				fprintf(stderr, "cannot write to temp nrm2 file: %s\n", nrm2_name);
				rc = 0;
				break;
			}
			p = p->next;
		}
	}
	fclose(nrm2);
	return rc;
}


int stap_to_nrm2(char * data_name, char *nrm2_name)
{
	unsigned char *data_blk;
	unsigned char *data_end;
	int data_fd;

	struct stat st;
		

	data_fd = open(data_name, O_RDONLY);
	if (data_fd == -1) {
		fprintf(stderr, "cannot open input stap data file %s\n", data_name);
		return 0;
	}

	if (fstat(data_fd, &st)) {
		fprintf(stderr, "cannot stat %s\n", data_name);
		return 0;
	}

	data_blk = malloc(st.st_size);
	data_end = data_blk + st.st_size;
	if (read(data_fd, data_blk, st.st_size) != st.st_size) {
		fprintf(stderr, "cannot read stap data file: %s\n", data_name);
		return 0;
	}

	/* initialize per_cpu_records */
	memset(per_cpu_records, 0,
		(MAX_NR_CPUS-1)*sizeof(struct per_cpu_list));

	parse_data(data_blk, data_end);
	return write_nrm2(nrm2_name);
}
