/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2008
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/*
 * itrace is a tool to do an instruction trace by generating a
 * systemtap script and running it to collect instruction trace data
 */
enum options {
	FUNC, PROG, PID, OUTPUT, VERBOSE, KEEP, QTRACE,
	DEBUGINFO, HELP, UNKNOWN};

char *option_names[] = {
	"--func", "--prog", "--pid", "--output", "--verbose", "--keep",
	"--qtrace", "--debuginfo", "--help"
};


#define MAX_ARGS 100
int num_funcs = 0;
int num_stap_args = 0;
int verbose = 0;
int keep = 0;
int debuginfo = 0;
int qtrace = 0;
char *prog = NULL;
char *trace_pid = NULL;
char *output_name = "itrace.out";
char *step_mode = "single_step";
char *stap_args[MAX_ARGS];
char *funcs[MAX_ARGS];
char generated_stap_file[PATH_MAX];
char *line_buffer;
pid_t child_pid;

char str_heap[5000];  /* space for allocating strings for func name/func addr */
int str_heap_index = 0;
char template_dir[PATH_MAX];
char *tmp_dir;

int gen_stapfile(void);

/* print help message */
void do_help(void)
{ 
	printf("%s%s\n",
		"usage: itrace --func=<func> --prog=<prog> --pid=<pid>",
		" [--output=<file>]");
	printf("%s\n", "   --help          this message");
	printf("%s%s\n", "   --func=<function>   function to be traced ",
		" [required]");
	printf("%s%s\n", "   --prog=<program>   program to be executed and traced",
		" [required]");
	printf("%s\n\t\t\t%s\n", "   --pid=<pid>     pid of process to be traced",
	       "[if not given, the program specified in --prog option will be executed]");
	printf("%s%s\n", "   --output=<file> name output trace file",
		" [default: itrace.out]");
	printf("%s\n", "   --verbose       produce verbose output");
	printf("%s\n", "   --keep          keep generated tapscript");
	printf("%s\n", "   --debuginfo     use kernel debuginfo to terminate systemtap session");
	printf("%s\n", "   --qtrace        generate qtrace output");
	printf("%s\n", "   Arguments to stap");
}

void error_if_empty(char *arg, char *value)
{
	if (!value || strlen(value) == 0) {
		printf("No value given for option %s\n", arg);
		do_help(); 
		exit(1);
	}
}

enum options lookup_option(char *arg)
{
	enum options i;

	for (i = 0; i < UNKNOWN; i++) 
		if (strcmp(arg, option_names[i]) == 0)
			return i;
	return UNKNOWN;
	
}



int check_options (int argc, char * argv[])
{
	int i;

	if (argc == 2) {
		do_help();
		exit(0);
	}

	for (i=1; i < argc; i++) {
		char *arg = argv[i];
		char * equal = index(arg, '=');
		char *val = NULL;

		if (equal) {
			*equal = 0;	/* terminate arg string */
			val = equal + 1;
		}

		switch(lookup_option(arg)) {
		case FUNC:
			error_if_empty(arg, val);
			funcs[num_funcs++] = val;
			break;

		case PROG:
			if (prog)
				printf("duplicate prog option ignored: %s\n", argv[i]);
			else {
				error_if_empty(arg, val);
				prog = val;
			}
			break;
		case PID:
			if (trace_pid)
				printf("duplicate pid option ignored: %s\n", argv[i]);
			else {
				error_if_empty(arg, val);
				trace_pid = val;
			}
			break;


		case OUTPUT:
			error_if_empty(arg, val);
			output_name = val;
			break;

		case VERBOSE:
			verbose=1;
			break;

		case KEEP:
			keep=1;
			break;

		case QTRACE:
			qtrace=1;
			break;

		case DEBUGINFO:
			debuginfo=1;
			break;

		case HELP:
			do_help();
			exit(0);
			break;

		default:
			stap_args[num_stap_args++]  = argv[i];
			break;
		}
	}

}

void dump_options(void)
{
	int i;

	if (trace_pid)
		printf("PID:\t\t%s\n", trace_pid);
	printf("OUTPUT_NAME:\t%s\n", output_name);
	printf("VERBOSE:\t%d\n", verbose);
	printf("KEEP:\t\t%d\n", keep);
	printf("DEBUGINFO:\t\t%d\n", debuginfo);
	printf("QTRACE:\t\t%d\n", qtrace);
	if (prog) {
		printf("PROG:\t\t%s\n", prog);
		for (i = 0; i < num_funcs; i++)
			printf("FUNC:\t\t%s\n",
				funcs[i]);
	}
}

char *strinit(char *str)
{
	char *new_str = str_heap + str_heap_index;

	str_heap_index += strlen(str);
	str_heap[str_heap_index++] = 0;
	strcpy(new_str, str);
	return new_str;
}


int find_func(char *func)
{
	int i;

	for (i=0; i < num_funcs; i++)
		if (strcmp(func, funcs[i]) == 0)
			return i;
	return -1;
}


void cleanup(void)
{
	char cmd[PATH_MAX];

	sprintf(cmd, "rm -fr %s", tmp_dir);
	system(cmd);
	fflush(stderr);
}

void sigint_handler(int signum)
{
	kill(child_pid, SIGINT);
	if (verbose)
		printf("parent killed %d\n", child_pid);
}

void run_stap(void)
{
	int i;
	int status = 0;
	int signum;
	sigset_t sigset;
	int arg_number = 0;
	char *argv[100];
	struct stat statbuf;
	char argument[PATH_MAX + 3];

	signal(SIGINT, sigint_handler);
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGINT);
	sigprocmask(SIG_BLOCK, &sigset, NULL);

	if ((child_pid = fork()) < 0) {
		perror ("fork");
		return;
	} else if (child_pid == 0) {	/* child */
		if (trace_pid)
			snprintf(argument, PATH_MAX + 3, "-x %s", trace_pid);
		else
			snprintf(argument, PATH_MAX + 3, "-c %s", prog);

		printf("stap -g %s %s -o %s",
		       argument, generated_stap_file, output_name);
		argv[arg_number++] = "stap";
		argv[arg_number++] = "-g";
		if (trace_pid) {
			argv[arg_number++] = "-x";
			argv[arg_number++] = trace_pid;
		} else {
			argv[arg_number++] = "-c";
			argv[arg_number++] = prog;
		}
		argv[arg_number++] = generated_stap_file;
		argv[arg_number++] = "-o";
		argv[arg_number++] = output_name;
		for (i=0; i < num_stap_args; i++) {
			argv[arg_number + i] = stap_args[i];
			printf(" %s", stap_args[i]);
			arg_number++;
		}
		putchar('\n');
		argv[arg_number] = NULL;
		unlink(output_name);

		signal(SIGINT, SIG_DFL);
		printf("NOTE: If stap produces errors re-run with -vvvv option\n");
		printf("If using with --pid option:\n");
		printf("- check dmesg to wait for SystemTap informational message");
		printf("- continue program to be traced (if necessary)\n");
		printf("- hit ^C to terminate systemtap session when done tracing\n");
		printf("\t(^C is not necessary if using --debuginfo option)\n");
		fflush(stdout);
		if (execvp("stap", argv) < 0) {
			perror("stap");
			exit(-1);
		}
	}


	/* wait here until signaled */
	if (verbose)
		printf("stap child pid=%d\n", child_pid);
	waitpid(child_pid, &status, 0);

	if (stat(output_name, &statbuf))
		return;

	argv[0] = "post";
	argv[1] = "-arc";
	argv[2] = "-ss";
	argv[3] = "-show";
	argv[4] = "-r";
	argv[5] = output_name;
	i = 6;
	if (qtrace)
		argv[i++] = "-qt";
	argv[i] = 0;
	(void)freopen("/dev/null", "w", stdout);
	(void)freopen("/dev/null", "w", stderr);
	if (execvp("post", argv) < 0)
		perror("post");

}

int main (int argc, char * argv[])
{
	int arg_error = 0;

	strcpy(template_dir, "/tmp/itraceXXXXXX");
	if ((tmp_dir = mkdtemp(template_dir)) < 0) {
		perror("trace");
		fprintf(stderr, "itrace: cannot make directory %s\n", tmp_dir);
		exit(1);
	}

	check_options(argc, argv);

	if (!num_funcs || !prog) {
		fprintf(stderr,
			"itrace: --prog and --func options required\n");
		arg_error = 1;
	}


	if (keep)
		printf("temp file directory= %s\n", tmp_dir);

	if (!arg_error && gen_stapfile()) {
		if (verbose)
			dump_options();
		run_stap();
	}

	if (!keep)
		cleanup();
}

int gen_stapfile(void)
{
	int i;
	FILE *stap_fd;

	sprintf(generated_stap_file, "%s/itrace_it.stp", tmp_dir);
	if (!(stap_fd = fopen(generated_stap_file, "w"))) {
	      fprintf(stderr,
	      	"itrace: ERROR cannot create %s\n", generated_stap_file);
	      return 0;
	}

#define FPF(str) fprintf(stap_fd, str);
#define FPF1(str, arg) fprintf(stap_fd, str, arg);
#define FPF2(str, arg1, arg2) fprintf(stap_fd, str, arg1, arg2);
FPF("# AUTOMATICALLY GENERATED SCRIPT FOR USER INSTRUCTION TRACING\n");
FPF("# Copyright (C) 2005, 2006, 2007, 2009 IBM Corp.\n");
FPF("\n");
FPF("# You can redistribute it and/or modify it under the terms of the GNU General\n");
FPF("# Public License (GPL); either version 2, or (at your option) any\n");
FPF("# later version.\n");
FPF("\n");
FPF("\n");
FPF("global trace_tid\n");
FPF("global itrace_on = 0\n");
FPF("probe begin\n");
FPF("{\n");
FPF("    if (!init_pi_itrace()) {\n");
FPF("        error(\"init_pi_itrace failed\\n\")\n");
FPF("		exit()\n");
FPF("    }\n");
FPF("}\n");
FPF("\n");
FPF("probe end\n");
FPF("{\n");
FPF("	cleanup_pi_itrace()\n");
FPF("}\n");
FPF("\n");
FPF("\n");

		for (i=0; i < num_funcs; i++) {
FPF1("probe process(\"%s\").insn if (itrace_on) ", prog);
FPF("{\n");
FPF("    pi_trace_it()\n");
FPF("}\n");

FPF2("probe process(\"%s\").function(\"%s\") ", prog, funcs[i]);
FPF("{\n");
FPF("    trace_tid = tid()\n");
FPF("    pi_itrace_create_task(task_current())\n");
FPF("    pi_itrace_on(trace_tid)\n");
FPF("    itrace_on = 1\n");
FPF("}\n");
FPF("\n");
FPF2("probe process(\"%s\").function(\"%s\").return ", prog, funcs[i]);
FPF("{\n");
FPF("    itrace_on = 0\n");
FPF("    pi_itrace_off()\n");
FPF("}\n");
FPF("\n");
			}

if (debuginfo) {
FPF("probe kernel.function(\"do_exit\")\n");
FPF("{\n");
FPF("	if (tid() == trace_tid)\n");
FPF("		exit()\n");
FPF("}\n");
}
	fclose(stap_fd);
	return 1;
}
